
#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

__doc__ = """
The part of the ADLER code responsible for
the handling of the files and processing the data.
"""

import math
import numpy as np
import os
import time
import sys
import gzip
from os.path import expanduser

has_voigt = True
try:
    from scipy.special import voigt_profile
except:
    from scipy.special import wofz
    has_voigt = False
from scipy.optimize import leastsq
from scipy.interpolate import interp1d
from scipy.fftpack import rfft, irfft, fftfreq
from astropy.io import fits as fits_module

# import ctypes

from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot, QMutex
from ExtendGUI import CustomThreadpool
# this is a Windows thing
# ctypes.windll.kernel32.SetDllDirectoryW('.')

# simple mathematical functions are defined here

rand_mt = np.random.Generator(np.random.MT19937(np.random.SeedSequence(31337)))

gauss_denum = 2.0 * (2.0 * math.log(2.0))**0.5
gauss_const = 1/((2.0*math.pi)**0.5)

oldval = 0.0

# has_voigt = False
# from scipy.special import wofz

if has_voigt:
    def my_voigt(xarr, A, xc, wG, wL):
        x_span = xarr.max()-xarr.min()
        temp_x = np.linspace(-2*x_span, 2*x_span, 8*len(xarr))
        temp_y = voigt_profile(temp_x, wG, wL) *abs(A)
        temp_func = interp1d(temp_x + xc, temp_y, bounds_error = False, fill_value = 0.0)
        return temp_func(xarr)
else:
    def my_voigt(xarr, A_in, xc_in, wG_in, wL_in):
        A = np.abs(A_in)
        xc = xc_in
        wG = np.abs(wG_in)
        wL = np.abs(wL_in)
        x_span = xarr.max()-xarr.min()
        temp_x = np.linspace(-2*x_span, 2*x_span, 8*len(xarr))
        # temp_y = voigt_profile(temp_x, wG, wL) *A
        temp_y = A * np.real(wofz((temp_x + 1j*wL)/(wG*np.sqrt(2))) ) / (wG * np.sqrt(2*np.pi))
        temp_func = interp1d(temp_x + xc, temp_y, bounds_error = False, fill_value = 0.0)
        return temp_func(xarr)

def loadtext_wrapper(fname):
    result = []
    try:
        source = open(fname, 'rb')
    except:
        return None
    else:
        for n, line in enumerate(source):
            try:
                textline = line.decode('utf-8')
            except:
                try:
                    textline = line.decode('cp1252')
                except:
                    continue
                else:
                    result.append(textline.strip('\n'))
            else:
                result.append(textline.strip('\n'))
        source.close()
        return result

def compress_array(inp_array):
    data = inp_array.copy()
    shape = data.shape
    dtype = data.dtype.str
    bdata = data.tobytes()
    compdata = gzip.compress(bdata)
    return compdata, shape, dtype

def decompress_array(compdata, shape, dtype):
    bdata = gzip.decompress(compdata)
    linedata = np.frombuffer(bdata, dtype=dtype)
    truedata = linedata.reshape(shape)
    return truedata

def store_array(fname, compdata, shape, dtype, maxdim=4, dimlen=5, dtypelen=4):
    ndim = len(shape)
    header = (maxdim*dimlen+dtypelen)*'0'
    for n in np.arange(ndim):
        tnum = shape[n]
        text = str(tnum).zfill(5)
        header[n*dimlen:(n+1)*dimlen] = text
    header[maxdim*dimlen:maxdim*dimlen+dtypelen] = dtype.zfill(dtypelen)
    dump = open(fname, 'wb')
    dump.write(header.encode('ascii'))
    dump.write(compdata)
    dump.close()

def restore_array(fname, maxdim=4, dimlen=5, dtypelen=4,  buffsize = 8192):
    source = open(fname, 'rb')
    header = source.read(maxdim*dimlen+dtypelen)
    compdata = b''
    buffer = b'999'
    while buffer:
        buffer = source.read(buffsize)
        compdata += buffer
    source.close()
    shape = []
    for x in np.arange(maxdim):
        val = int(header[x*dimlen:(x+1)*dimlen].decode('ascii'))
        if val>0:
            shape.append(val)
    dtype= header[maxdim*dimlen:maxdim*dimlen+dtypelen].decode('ascii').strip('0')
    the_array = decompress_array(compdata, shape, dtype)
    return the_array


# print(sys._MEIPASS)

def resource_path(relative_path):
    """ Get absolute path to resource, works for dev and for PyInstaller """
    try:
        # PyInstaller creates a temp folder and stores path in _MEIPASS
        base_path = sys._MEIPASS
    except Exception:
        base_path = os.path.abspath(".")

    return os.path.join(base_path, relative_path)

class Worker(QObject):
    finished = pyqtSignal()
    intReady = pyqtSignal(int)
    runPeriodically = None
    keeprunnig = True
    waittime = 1
    def set_end(self):
        self.keeprunning = False
    def set_waittime(self, wtime):
        self.waittime = wtime
    def assignFunction(self, some_function):
        self.runPeriodically = some_function
    def assignArgs(self, args):
        self.func_args = args
    @pyqtSlot()
    def updater(self): # A slot takes no params
        while self.keeprunnig:
            time.sleep(self.waittime)
            if self.runPeriodically is not None:
                self.runPeriodically()
        self.finished.emit()
    @pyqtSlot()
    def run_once(self): # A slot takes no params
        try:
            nargs = len(self.func_args)
        except:
            nargs = 0
        if self.runPeriodically is not None:
            if nargs == 0:
                self.runPeriodically()
            elif nargs == 1:
                self.runPeriodically(self.func_args[0])
            elif nargs == 2:
                self.runPeriodically(self.func_args[0],  self.func_args[1])
        self.finished.emit()
    @pyqtSlot()
    def run_repeated(self): # A slot takes no params
        try:
            nargs = len(self.func_args)
        except:
            nargs = 0
        if self.runPeriodically is not None:
            for n in range(nargs):
                self.runPeriodically(self.func_args[n])
                self.intReady.emit(n)
        self.finished.emit()

def gaussian(a, FWHM, centre, array):
    """
    gaussian(a, FWHM, centre, arguments_array)
    Takes an array of x values and returns a
    gaussian peak with heighth a with FWHM and centre
    defined by input parameters.
    """
    c = FWHM / gauss_denum
    return  a * np.exp( - (array - centre)**2 / (2 * c**2))

def polynomial(args, array):
    """
    An n-th order polynomial.
    """
    ys = np.zeros(array.shape)
    for n, x in enumerate(args):
        ys += x*array**n
    return ys

def fit_polynomial(args, array):
    newfit = polynomial(args, array[:,0])
    return (newfit - array[:,1])/(array[:,2]+np.abs(0.00001*array[:,1].mean()))

def fit_gaussian(args, array):
    a, FWHM, centre, zeroline = args[0], args[1], args[2], args[3]
    newvals = gaussian(a, FWHM, centre, array[:,0]) + zeroline
    return newvals - array[:,1]

def fit_gaussian_fixedbase(args, array, base):
    a, FWHM, centre = args[0], args[1], args[2]
    newvals = gaussian(a, FWHM, centre, array[:,0]) + base
    return newvals - array[:,1]

def elastic_line(profile, background, fixwidth = None, olimits = (None, None)):
    """
    As an initial guess, this function will assume that the 
    elastic line is at the point with the highest intensity.
    Otherwise, an override for the peak centre can be provided.
    """
    chi2 = -1.0
    if len(background) > 1:
        zline, zerr = background[:,1].mean(), background[:,1].std()
    else:
        zline = profile[:, 1].min()
        zerr = abs(profile[:, 1].max() - profile[:, 1].min())**0.5
    temp = profile.copy()
    temp[:,1] -= zline
    maxval = temp[:,1].max()
    nlimits = [0, -1]
    if olimits[0] is not None and olimits[1] is not None:
        boolarr1 = temp[:, 0] > olimits[0]
        if boolarr1.any() and not boolarr1.all():
            nlimits[0] = max(np.argmax(boolarr1) -1, 0)
        else:
            nlimits[0] = 0
        boolarr2 = temp[:, 0] > olimits[1]
        if boolarr2.any() and not boolarr2.all():
            nlimits[1] = min(np.argmax(boolarr2), len(temp)-1)
        else:
            nlimits[1] = len(temp)-1
        if abs(nlimits[1]-nlimits[0]) < 5:
            nlimits[1] += 1
        if abs(nlimits[1]-nlimits[0]) < 5:
            nlimits[0] -= 1
        if abs(nlimits[1]-nlimits[0]) < 5:
            nlimits[1] += 1
        if abs(nlimits[1]-nlimits[0]) < 5:
            nlimits[0] -= 1
        # centre = (temp[nlimits[0],0] + temp[nlimits[1],0])/2
        nmaxval = temp[nlimits[0]:nlimits[1],1].max()
        centre = temp[nlimits[0]:nlimits[1],0][np.where(temp[nlimits[0]:nlimits[1],1] == nmaxval)][-1]
    else:
        centre = temp[:,0][np.where(temp[:,1] == maxval)][-1]
    tempcrit = np.abs(temp[:,0]-centre)
    crit = tempcrit.min()
    cindex, = np.where(tempcrit == crit)
    if len(cindex) == 1:
        cindex = int(cindex)
    else:
        cindex = int(len(temp)/2)
    table = np.zeros((90,2))
    table[:,0] = np.arange(len(table)) + 1
    for x in np.arange(len(table)):
        try:
            table[x,1] = temp[cindex - table[x,0],1] + temp[cindex + table[x,0],1]
        except IndexError:
            break
    criterion = np.abs(table[:,1] - maxval)
    fwhm = int(table[:,0][np.where(criterion == criterion.min())].mean())
    step = np.array([fwhm])[0]
    if olimits[0] is not None and olimits[1] is not None:
        subset = profile[nlimits[0]:nlimits[1]]
    else:
        subset = profile[-step + cindex: step + cindex]
        counter, running = 0, 1
        while (len(subset) < 350) and running:
            step +=1
            subset = profile[-step + cindex: step + cindex]
            if ((subset[:3,1].mean() + subset[-3:,1].mean() - 2*zline)/zerr)**2 < 2.0:
                counter += 1
            if counter > 8:
                running = 0
            # print(((subset[:3,1].mean() + subset[-3:,1].mean() - 2*zline)/zerr)**2)
    if fixwidth is not None:
        subset = profile[-int(fixwidth) + cindex: int(fixwidth) + cindex]
        maxval = subset[:,1].max()
        zline = subset[:,1].min()
        fwhm = subset[:,0].std()/2.0
        centre = subset[:,0].mean()
        if len(subset) < 5:
            print("Not enough points in the curve.")
            return None, None, None
        pfit, pcov, infodict, errmsg, success = leastsq(fit_gaussian, np.array([maxval, fwhm, centre, zline]), args=(subset,), full_output=1)
        if pcov is not None:
            s_sq = (fit_gaussian(pfit, subset)**2).sum()/(len(subset)-len(pfit))
            a, FWHM, centre, zeroline = pfit
            tpeak = gaussian(a, FWHM, centre, subset[:,0]) + zeroline
            chi2 = chisquared(tpeak, subset,  zeroline)
            pcov = pcov * s_sq
    else:
        if len(subset) < 5:
            print("Not enough points in the curve.")
            return None, None,  None
        pfit, pcov, infodict, errmsg, success = leastsq(fit_gaussian_fixedbase, np.array([maxval, fwhm, centre]), args=(subset, zline), full_output=1)
        if pcov is not None:
            s_sq = (fit_gaussian_fixedbase(pfit, subset, zline)**2).sum()/(len(subset)-len(pfit))
            a, FWHM, centre = pfit
            tpeak = gaussian(a, FWHM, centre, subset[:,0]) + zline
            chi2 = chisquared(tpeak, subset, zline)
            pcov = pcov * s_sq
    error = [] 
    for i in range(len(pfit)):
        try:
            error.append(np.absolute(pcov[i][i])**0.5)
        except:
            error.append( 0.00 )
    optimised = [pfit, np.array(error)]
    a,b,c = optimised[0][0], optimised[0][1], optimised[0][2]
    newx =  np.linspace(subset[0,0], subset[-1,0], len(subset)*10)
    if fixwidth is not None:
        peak = gaussian(a,b,c, newx) + optimised[0][3]
    else:
        peak = gaussian(a,b,c, newx) + zline
    return optimised, np.column_stack([newx, peak]),  chi2

def elastic_line_anyx(profile, background, fixwidth = None, olimits = (None, None), init_fwhm = 0.05):
    """
    As an initial guess, this function will assume that the 
    elastic line is at the point with the highest intensity.
    Otherwise, an override for the peak centre can be provided.
    """
    chi2 = -1.0
    if len(background) > 1:
        zline, zerr = background[:,1].mean(), background[:,1].std()
    else:
        zline = profile[:, 1].min()
        zerr = abs(profile[:, 1].max() - profile[:, 1].min())**0.5
    temp = profile.copy()
    temp[:,1] -= zline
    maxval = temp[:,1].max()
    if olimits[0] is not None and olimits[1] is not None:
        # centre = ((temp[:, 0][np.where(crit1==min1)] + temp[:, 0][np.where(crit2==min2)])/2)[0]
        if olimits[0] <= temp[:, 0].min():
            if olimits[1] >= temp[:, 0].max():
                subset = profile
            else:
                subset = profile[:np.argmax(temp[:, 0] >= olimits[1])]
        else:
            if olimits[1] >= temp[:, 0].max():
                subset = profile[np.argmax(temp[:, 0] >= olimits[0]):]
            else:
                subset = profile[np.argmax(temp[:, 0] >= olimits[0]):np.argmax(temp[:, 0] >= olimits[1])]
        if len(subset) < 2:
            return None, None,  None
        centre = subset[:,0][np.where(subset[:,1] == subset[:, 1].max())][-1]
    else:
        subset = None
        centre = temp[:,0][np.where(temp[:,1] == maxval)][-1]
    tempcrit = np.abs(temp[:,0]-centre)
    crit = tempcrit.min()
    cindex, = np.where(tempcrit == crit)
    if len(cindex) == 1:
        cindex = int(cindex)
    else:
        cindex = int(cindex.mean())
    table = np.zeros((90,2))
    table[:,0] = np.arange(len(table)) + 1
    for x in np.arange(len(table)):
        try:
            table[x,1] = temp[cindex - table[x,0],1] + temp[cindex + table[x,0],1]
        except IndexError:
            break
    criterion = np.abs(table[:,1] - maxval)
    fwhm = int(table[:,0][np.where(criterion == criterion.min())].mean())
    step = np.array([fwhm])[0]
    if subset is None:
        subset = profile[-step + cindex: step + cindex]
        counter, running = 0, 1
        while (len(subset) < 350) and running:
            step +=1
            subset = profile[-step + cindex: step + cindex]
            if ((subset[:3,1].mean() + subset[-3:,1].mean() - 2*zline)/zerr)**2 < 2.0:
                counter += 1
            if counter > 8:
                running = 0
            # print(((subset[:3,1].mean() + subset[-3:,1].mean() - 2*zline)/zerr)**2)
    if fixwidth is not None:
        subset = profile[-int(fixwidth) + cindex: int(fixwidth) + cindex]
        maxval = subset[:,1].max()
        zline = subset[:,1].min()
        fwhm = subset[:,0].std()/2.0
        centre = subset[:,0].mean()
        if len(subset) < 5:
            print("Not enough points in the curve.")
            return None, None,  None
        pfit, pcov, infodict, errmsg, success = leastsq(fit_gaussian, np.array([maxval, fwhm, centre, zline]), args=(subset,), full_output=1)
        if pcov is not None:
            s_sq = (fit_gaussian(pfit, subset)**2).sum()/(len(subset)-len(pfit))
            a, FWHM, centre, zeroline = pfit
            tpeak = gaussian(a, FWHM, centre, subset[:,0]) + zeroline
            chi2 = chisquared(tpeak, subset,  zeroline)
            pcov = pcov * s_sq
    else:
        if len(subset) < 5:
            print("Not enough points in the curve.")
            return None, None,  None
        pfit, pcov, infodict, errmsg, success = leastsq(fit_gaussian_fixedbase, np.array([maxval, init_fwhm, centre]), args=(subset, zline), full_output=1)
        if pcov is not None:
            s_sq = (fit_gaussian_fixedbase(pfit, subset, zline)**2).sum()/(len(subset)-len(pfit))
            a, FWHM, centre = pfit
            tpeak = gaussian(a, FWHM, centre, subset[:,0]) + zline
            chi2 = chisquared(tpeak, subset, zline)
            pcov = pcov * s_sq
    error = [] 
    for i in range(len(pfit)):
        try:
            error.append(np.absolute(pcov[i][i])**0.5)
        except:
            error.append( 0.00 )
    optimised = [pfit, np.array(error)]
    a,b,c = optimised[0][0], optimised[0][1], optimised[0][2]
    newx =  np.linspace(subset[0,0], subset[-1,0], len(subset)*10)
    if fixwidth is not None:
        peak = gaussian(a,b,c, newx) + optimised[0][3]
    else:
        peak = gaussian(a,b,c, newx) + zline
    return optimised, np.column_stack([newx, peak]),  chi2

# def discrete_rebin(data, offset = 0.0, padding = 6, nsubdiv = 100):
    # startlen = len(data) + 2*padding
    # newdata = np.zeros(nsubdiv*startlen)
    # for n, i in enumerate(data):
        # newdata[int((padding+offset+n)*nsubdiv):int((padding+offset+n+1)*nsubdiv)] = i/nsubdiv
    # target = np.zeros(len(data))
    # for n in range(len(data)):
        # target[n] = newdata[int((padding+n)*nsubdiv):int((padding+n+1)*nsubdiv)].sum()
    # return target
    
def simplify_number_range(flist):
    numbs = []
    prefs = []
    for generic, fname in enumerate(flist):
        fpath, shortname = os.path.split(fname)
        shortname = ".".join(shortname.split('.')[:-1])
        #for bad_start in ['Merged_', 'BestMerge_']:
        #    shortname = shortname.lstrip(bad_start)
        #for bad_end in ['_1D_deltaE', '_1D']:
        #    shortname = shortname.rstrip(bad_end)
        if 'Merged_' in shortname[:7]:
            shortname = shortname[7:]
        elif 'BestMerge_'in shortname[:10]:
            shortname = shortname[10:]
        if '_1D' in shortname[-3:]:
            shortname = shortname[:-3]
        elif '_1D_deltaE' in shortname[-10:]:
            shortname = shortname[:-10]
        for n in range(len(shortname)):
            try:
                fnumber = int(shortname[n:])
            except:
                continue
            else:
                prefix = shortname[:n]
                break
        try:
            fnumber
        except:
            fnumber = generic+1
            prefix = shortname
        numbs.append(fnumber)
        prefs.append(prefix)
    prefs = np.array(prefs)
    numbs = np.array(numbs).astype(np.int)
    unique_names = np.unique(prefs)
    textsegs = []
    for un in unique_names:
        cname = un
        aaa = numbs[np.where(prefs == un)]
        subnums = np.sort(aaa)
        ranges = []
        for n,  val in enumerate(subnums):
            if n == 0:
                beg = val
                end = val
            else:
                if val == (subnums[n-1] + 1):
                    end = val
                else:
                    ranges.append((beg,  end))
                    beg = val
                    end = val
            if n==(len(subnums)-1):
                ranges.append((beg,  end))
        for rr in ranges:
            if rr[0] == rr[1]:
                textsegs.append(cname+str(rr[0]))
            else:
                textsegs.append(cname+str(rr[0])+'-'+str(rr[1]))
    return textsegs
    
def unit_to_int(text):
    if 'hanne' in text:
        return 0
    elif 'ransfe' in text:
        return 1
    elif 'nergy' in text:
        return 2
    else:
        return -1

def int_to_unit(num):
    if num == 0:
        return "Detector channels"
    elif num ==1:
        return "Energy transfer [eV]"
    elif num ==2:
        return "Energy [eV]"
    else:
        return "Unknown"
    
def discrete_rebin(data, offset = 0.0, padding = 6, nsubdiv = 100):
    startlen = len(data) # + 2*padding
    newdata = np.ones((startlen, nsubdiv)) / nsubdiv
    newdata = (newdata.T * data).T
    fulllen = startlen*nsubdiv
    newdata = newdata.reshape(fulllen)
    target = np.zeros(len(data))
    firstone, lastone = int((offset)*nsubdiv), int((offset+startlen)*nsubdiv)
    if firstone < 0:
        startind = 0
    else:
        startind = firstone
    newdata = newdata[startind:lastone+1]
    templen = len(newdata)
    missing = fulllen - templen
    if firstone < 0:
        newdata = np.concatenate([np.zeros(missing), newdata])
    elif lastone >= fulllen:
        newdata = np.concatenate([newdata, np.zeros(missing)])
    newdata = newdata.reshape((startlen, nsubdiv))
    target = newdata.sum(1)
    # for n in range(len(data)):
    #     target[n] = newdata[int((offset+n)*nsubdiv):int((offset+n+1)*nsubdiv)].sum()
    return target

def merge2curves(source, target):
    """
    This function will treat two curves with different binning as histograms,
    and merge the data proportionally to the bin overlap.
    The input curves have to be sorted in ascending order along the x axis.
    """
    xs1, ys1 = source[:,0], source[:,1]
    xs2, ys2 = target[:,0], target[:,1]
    step1 = xs1[1:] - xs1[:-1]
    step1 = np.concatenate([step1[:1], step1, step1[-1:]])
    lowlim1 = xs1 - 0.5*step1[:-1]
    highlim1 = xs1 + 0.5*step1[1:]
    # print('step1: ', step1.min(), step1.max(), step1.mean())
    # now we have low limits, high limits and middle for each bin, plus the bin size
    step2 = xs2[1:] - xs2[:-1]
    step2 = np.concatenate([step2[:1], step2, step2[-1:]])
    lowlim2 = xs2 - 0.5*step2[:-1]
    highlim2 = xs2 + 0.5*step2[1:]
    # print('step2: ', step2.min(), step2.max(), step2.mean())
    # now we can begin
    newy = ys2.copy()
    lowest_n = 0
    highest_n = len(ys2)-1
    lowest_m=0
    highest_m=len(ys1)-1
    # acc_n = []
    for n in np.arange(len(xs2)):
        m1 = np.argmax(highlim1 > lowlim2[n])
        m2 = np.argmax(lowlim1 > highlim2[n])
        if highlim1[0] > lowlim2[n]:
            m1 = 0
        elif highlim1[-1] < lowlim2[n]:
            m1 = len(highlim1) -1
        elif highlim1[-2] < lowlim2[n]:
            m1 = len(highlim1) -2
        if lowlim1[0] > highlim2[n]:
            m2 = 0
        elif lowlim1[-1] < highlim2[n]:
            m2 = len(lowlim1)-1
        elif lowlim1[-2] < highlim2[n]:
            m2 = len(lowlim1)-2
        # now m1 is the first bin in the source curve that overlaps with the current bin
        # and m2 is the first bin after that that doesn't
        for m in np.arange(m1-2,m2+2):
            if (m > highest_m) or (m < lowest_m):
                continue
            if (highlim1[m] < lowlim2[n]) or (lowlim1[m] > highlim2[n]):
                continue
            if lowlim1[m] <= lowlim2[n]:
                if highlim1[m] > highlim2[n]:
                    newy[n] += ys1[m]*(highlim2[n]-lowlim2[n])/step1[m]
                else:
                    newy[n] += ys1[m]*(highlim1[m]-lowlim2[n])/step1[m]
            else:
                if highlim1[m] > highlim2[n]:
                    newy[n] += ys1[m]*(highlim2[n]-lowlim1[m])/step1[m]
                else:
                    newy[n] += ys1[m]
            # factor = (highlim1[m] - max(lowlim2[n],lowlim1[m])/step1[m]
            # altfactor = (min(highlim2[n],highlim1[m]) - lowlim1[m])/step1[m]
    # acc_n = np.array(acc_n)
    # print('acc_n: ', acc_n.min(0), acc_n.max(0), acc_n.mean(0))
    return np.column_stack([xs2,newy])

def profile_offsets(args,  data,  to_match, tpoolin = None, pbarin = None,  maxthreads = 1):
    xshift = args[0]
    yshift = args[1]
    yscale = args[2]
    tempdat = data.copy()
    tempdat[:, 0] -= xshift
    tempdat[:, 1] *= yscale
    tempdat[:, 1] += yshift
    temptarget = to_match.copy()
    temptarget[:, 1] = 0.0
    # rebinned = merge2curves(tempdat,  temptarget)
    # the_object = MergeCurves(tempdat,  temptarget,  tpool = tpoolin,  pbar = pbarin)
    the_object = MergeCurves(tempdat,  temptarget, None,  pbar = pbarin,  mthreads = maxthreads)
    the_object.runit()
    rebinned = the_object.postprocess()
    return (rebinned[10:-10,1] - to_match[10:-10, 1])

def apply_offset_to_2D_data(array, locut, hicut, offset):
    """This is a dummy function for now.
    Later on, it will apply a function to transform the image,
    so that all the lines are straight instead of curved.
    """
    xaxis = np.arange(locut, hicut, 1) + 1 
    result = np.zeros(array.shape)
    for n, i in enumerate(xaxis):
        result[:,n] = discrete_rebin(array[:,n], offset = offset, padding = 6, nsubdiv=100)
    return result

def chisquared(peak, data,  zline):
    temp = peak-zline
    toppoint = temp.max()
    crit = np.where(np.abs(temp)/abs(toppoint) > 0.001)
    peakpoints = len(peak[crit].ravel())
    print(peakpoints,  len(peak))
    toppart = peak-data[:, 1]
    values = (toppart[crit])**2
    bottompart = (peak[crit])# **2
    values = values.sum()/bottompart.sum() /peakpoints
    # toppart = peak-data[:, 1]
    # values = (toppart*toppart) / np.abs(peak)
    return values

def curvature_profile(data2D, blocksize = 16, percentile = 70, override = None, olimits = (None, None)):
    width = data2D.shape[1]
    segnums = int(round(width/blocksize))
    segments = [data2D[:,int(n*blocksize):int((n+1)*blocksize)].sum(1) for n in np.arange(segnums)]
    results = []
    for num, seg in enumerate(segments):
        background = np.where(seg<np.percentile(seg, percentile))
        xaxis = np.arange(len(seg))+1
        prof = np.column_stack([xaxis, seg])
        back = prof[background]
        try:
            a,b, c = elastic_line(prof, back, override, olimits = olimits)
        except:
            continue
        else:
            if olimits[0] is None or olimits[1] is None:
                results.append([(num+0.5)*blocksize, a[0][2], a[1][2]])
            elif a[0][2] > olimits[0] and a[0][2] < olimits[1]:
                results.append([(num+0.5)*blocksize, a[0][2], a[1][2]])
    return np.array(results)

#### data processing part

def read_1D_curve(fname):
    curve = []
    envals = []
    units = 'Detector channels'
    source = loadtext_wrapper(fname)
    if source is None:
        print("could not read file: ", fname)
        return curve, envals, units
    else:
        for line in source:
            if len(line.split()) > 0:
                if '#' in line.split()[0][0]:
                    if 'hoton energy' in line:
                        label = line.split('hoton energy')[-1].strip()
                        for tok in label.split():
                            try:
                                tval = float(tok)
                            except:
                                continue
                            else:
                                envals.append(tok)
                else:
                    xy = [float(z) for z in line.split(',')]
                    curve.append((xy[0], xy[1]))
        curve = np.array(curve)
        sorting = np.argsort(curve[:,0])
        curve = curve[sorting]
        print(curve.shape)
        # source.close()
        xmin,  xmax = curve[:, 0].min(),  curve[:, 0].max()
        if xmin <0:
            units = "Energy Transfer [eV]"
        elif xmax - xmin > 1000.0:
            units = "Detector channels"
        else:
            units = "Energy [eV]"
        if len(envals) == 0:
            envals = [-1.0]
    return curve, envals, units

def ReadFits(fname):
    """Reads the .sif file produced by the Andor iKon-L CCD camera.
    The dimensions of the pixel array can be changed if necessary.
    """
    fitsfile = fits_module.open(fname)
    img = fitsfile[0]
    d_array = img.data
    return d_array.T.astype(np.float64), ""

def ReadAsc(fname):
    """Reads the old 2D .asc file format which for some reason
    was used on PEAXIS in the very beginning.
    Or, if things don't work out, it doesn't read it.
    """
    source = open(fname, 'r')
    data = []
    for n, line in enumerate(source):
        if n >= 2048:
            continue
        else:
            templine = str(line).replace('\t', ' ')
            toks = [float(x) for x in templine.split()[1:]]
            data.append(toks)
    source.close()
    d_array = np.array(data)# .astype(np.float64)
    print('ASC dimensions:',  d_array.shape)
    return d_array.T, ""

def ReadAndor(fname, dimensions = (2048,2048), byte_size=4, data_type='c'):
    """Reads the .sif file produced by the Andor iKon-L CCD camera.
    The dimensions of the pixel array can be changed if necessary.
    """
    size = os.path.getsize(fname)
    source = open(fname, 'rb')
    nrows = dimensions[-1]
    header, data = [],[]
    bindata = None    
    total_length = 0
    header = np.fromfile(source, np.byte, size - 4*dimensions[0]*dimensions[1] -2*4, '') # 2772 is close
    print("Headersize",  size - 4*dimensions[0]*dimensions[1] -2*4)
    data = np.fromfile(source, np.float32, dimensions[0]*dimensions[1], '')
    header = bytes(header)
    lines = header.split(b'\n')
    header = []
    for n, line in enumerate(lines):
        try:
            header.append(line.decode('ascii'))
        except UnicodeDecodeError:
            print('header lines skipped:', n+1, 'with length:', len(line))
    source.close()
    return np.array(data).reshape(dimensions), header

def SphericalCorrection(array, params = [1.0,1.0,1.0], locut = 0, hicut = 2048, direct_offsets = None):
    """This is a dummy function for now.
    Later on, it will apply a function to transform the image,
    so that all the lines are straight instead of curved.
    """
    xaxis = np.arange(locut, hicut, 1) + 1 
    if params[0] > 5.0 and params[1] > 5.0 and params[2] >5.0:
        if direct_offsets is not None:
            offsets = direct_offsets.copy()
        else:
            return array
    else:
        offsets = polynomial(params, xaxis)
    offsets -= offsets.mean()
    result = array.copy()
    for n, i in enumerate(xaxis):
        result[:,n] = discrete_rebin(array[:,n], offset = offsets[n], padding = 6, nsubdiv=100)
    return result

def RemoveCosmics(array, NStd = 3):
    """This function will filter out the cosmic rays from the signal.
    At the moment it assumes that each column should have uniform counts.
    Then it keeps only the part that is within N standard deviations away from the mean.
    """
    for n, line in enumerate(array):
        lmean = line.mean()
        lstddev = line.std()
        badpix = np.where(np.abs(line-lmean) > (NStd*lstddev))
        goodpix = np.where(np.abs(line-lmean) <= (NStd*lstddev))
        newmean = line[goodpix].mean()
        array[n][badpix] = newmean
    return array

def WriteProfile(fname, profile,
                        header = None, params = [None,  None], 
                        varlog = [None,  None]):
    target = open(fname, 'w')
    if header is not None:
        for ln in header:
            newln = "".join([str(x) for x in ln if x.isprintable()])
            target.write(' '.join(['#',newln.strip().strip('\n'),'\n']))
    if params[0] is not None and params[1] is not None:
        keys, vals = params[0], params[1]
        for kk in keys:
            val = str(vals[kk])
            target.write(' '.join([str(x) for x in ['# ADLER',kk, ':', val.strip().strip('()[]\n'), '\n']]))
    if varlog[0] is not None and varlog[1] is not None:
        keys, vals = list(varlog[0]), varlog[1]
        if len(keys)>0:
            target.write("# VARIABLES TIMELOG\n")
            target.write(" ".join(['#'] + keys + ['\n']))
            for ln in range(len(vals[keys[0]])):
                target.write(" ".join(['#'] + [str(round(vals[k][ln], 5)) for k in keys] + ['\n']))
    for n, val in enumerate(profile):
        target.write(",".join([str(x) for x in [val[0], val[1]]]) + '\n')
    target.close()

def WriteEnergyProfile(fname, array, header, params = None):
    target = open(fname, 'w')
    for ln in header:
        newln = "".join([str(x) for x in ln if x.isprintable()])
        target.write(' '.join(['#',newln.strip().strip('\n'),'\n']))
    if params is not None:
        keys, vals = params[0], params[1]
        for kk in keys:
            try:
                val = str(vals[kk[0]][kk[1]])
            except:
                val = "Missing"
            target.write(' '.join([str(x) for x in ['#',kk[0], kk[1], ':', val.strip().strip('()\n'), '\n']]))
    for n, val in enumerate(array):
        target.write(",".join([str(x) for x in [val[0], val[1]]]) + '\n')
    target.close()

def make_profile(array, reduction = 1.0,  xaxis = None,  tpoolin = None, pbarin = None,  maxthreads = 1):
    profile = array.sum(1)
    initial_x = np.arange(len(profile))+1
    if xaxis is None:
        final_x = np.linspace(1, 2048, int(len(profile)/reduction))
    else:
        final_x = xaxis.copy()
    final_y = np.zeros(len(final_x))
    finalprof = np.column_stack([final_x, final_y])
    profile = np.column_stack([initial_x, profile])
    # pprofile = merge2curves(profile, finalprof)
    # the_object = MergeCurves(profile,  finalprof, tpool = tpoolin, pbar = pbarin)
    the_object = MergeCurves(profile,  finalprof, tpool = None, pbar = pbarin, mthreads = maxthreads)
    the_object.runit()
    pprofile = the_object.postprocess()
    return pprofile

def make_stripe(data,  xlims,  ylims):
    stripe = data[ylims[0]:ylims[1],xlims[0]:xlims[1]].sum(0)
    stripe = np.column_stack([np.arange(xlims[0], xlims[1])+1, stripe])
    return stripe
    
def make_histogram(data,  xlims,  ylims):
    temp = data[ylims[0]:ylims[1],xlims[0]:xlims[1]]
    flat = np.ravel(temp)
    ordered = np.sort(flat)
    cons_limits = (ordered[2:10].mean() -0.5,  ordered[-10:-2].mean()+0.5)
    # nbins = max(int(round(len(flat)**0.5)),  10)
    nbins = max(int(round(cons_limits[1]) - round(cons_limits[0])) + 1,  3)
    hist,  binlims = np.histogram(temp,  bins = nbins,  range = cons_limits)
    return hist,  binlims

def load_lise(fname):
    header,  values,  finalvalues = [],  [],  []
    result,  names,  units = {}, {}, {}
    numbs,  sequence = [], []
    vallength = 0
    payattention = False
    dataspotted = False
    source = loadtext_wrapper(fname)   
    if not (source is None):
        for line in source:
            toks = line.split()
            if len(toks) > 0:
                if '#' in toks[0]:
                    header.append(line.strip('# '))
                else:
                    values.append(toks)
                    vallength = max(vallength,  len(toks))
        # source.close()
    for line in values:
        if len(line) == vallength:
            finalvalues.append([float(x) for x in line])
    for line in header:
        if 'LEGEND' in line:
            payattention = True
        if not payattention:
            continue
        else:
            if ':' in line:
                parts = line.split(':')
                number = int(parts[0])
                name = parts[1].split('>')[0].strip(' ')
                unit = parts[1].split('in')[-1].strip(' ')
                numbs.append(number)
                names[number] = name
                units[number] = unit
            elif 'DATA' in line:
                dataspotted = True
            elif dataspotted:
                toks = line.split()
                if len(toks) == len(numbs):
                    sequence = [int(x) for x in toks]
    finalvalues = np.array(finalvalues)
    for num in numbs:
        result[num] = finalvalues[:, num]
    return result, names, units, numbs

def load_datheader(fname):
    header = []
    resdict = {}
    source = loadtext_wrapper(fname)  
    if not (source is None):
        for line in source:
            if len(line.split()) > 0:
                if '#' in line.split()[0][0]:
                    header.append(line)
        # source.close()
    for ll in header:
        toks = ll.split()
        foundit = False
        for tn,  tok in enumerate(toks):
            if tn > 1:
                try:
                    val = float(tok)
                except:
                    continue
                else:
                    foundit = True
                    keyval = " ".join(toks[:tn])
                    resdict[keyval] = val
                    break
        if not foundit:
            ntoks = ll.split(':')
            if len(ntoks) > 1:
                resdict[ntoks[0]] = ntoks[1]
    return resdict

def load_datlog(fname):
    header,  values,  finalvalues = [],  [],  []
    vallength = 0
    finalheader = None
    source = loadtext_wrapper(fname)   
    if not (source is None):
        for line in source:
            toks = line.split()
            if len(toks) > 0:
                if '#' in toks[0]:
                    header.append(toks)
                else:
                    values.append(toks)
                    vallength = max(vallength,  len(toks))
        # source.close()
    for line in values:
        if len(line) == vallength:
            finalvalues.append([float(x) for x in line])
    for line in header:
        if len(line) == vallength+1:
            finalheader = line[1:]
    finalvalues = np.array(finalvalues)
    result = {}
    if finalheader is not None:
        for n,  k in enumerate(finalheader):
            result[k] = finalvalues[:, n]
    return result

#class WorkerSignals(QObject):
#    '''
#    Defines the signals available from a running worker thread.
#
#    Supported signals are:
#
#    finished
#        No data
#    
#    error
#        `tuple` (exctype, value, traceback.format_exc() )
#    
#    result
#        `object` data returned from processing, anything
#
#    progress
#        `int` indicating % progress 
#
#    '''
#    finished = pyqtSignal()
#    error = pyqtSignal(tuple)
#    result = pyqtSignal(object)
#    progress = pyqtSignal(int)

def load_lise_logs(flist):
    logs, ns,  us, nbs = [], [], [], []
    vallog, unitlog = {}, {}
    for fname in flist:
        result, names, units, numbs = load_lise(fname)
        logs.append(result)
        ns.append(names)
        us.append(units)
        nbs.append(numbs)
    for nn, lg in enumerate(logs):
        names = ns[nn]
        units = us[nn]
        numbs = nbs[nn]
        lk = [names[int(x)] + " (" + units[int(x)].strip(' \r\n\t')  + ")" for x in lg.keys()]
        for nk, k in enumerate(lk):
            if k in vallog.keys():
                vallog[k] = np.concatenate([vallog[k],  lg[nk]])
            else:
                vallog[k] = lg[nk].copy()
    if 'time' in vallog.keys():
        sequence = np.argsort(vallog['time'])
        for k in vallog.keys():
            vallog[k] = vallog[k][sequence]
    return vallog
    
def load_only_logs(flist):
    logs = []
    vallog = {}
    for fname in flist:
        logs.append(load_datlog(fname))
    for lg in logs:
        lk = lg.keys()
        for k in lk:
            if k in vallog.keys():
                vallog[k] = np.concatenate([vallog[k],  lg[k]])
            else:
                vallog[k] = lg[k].copy()
    if 'Time' in vallog.keys():
        sequence = np.argsort(vallog['Time'])
        for k in vallog.keys():
            vallog[k] = vallog[k][sequence]
    return vallog

def load_and_average_logs(flist):
    logs = []
    vallog = {}
    errlog = {}
    for fname in flist:
        logs.append(load_datlog(fname))
    for lg in logs:
        lk = lg.keys()
        for k in lk:
            if k in vallog.keys():
                vallog[k] = np.concatenate([vallog[k],  lg[k]])
            else:
                vallog[k] = lg[k].copy()
    if 'Time' in vallog.keys():
        sequence = np.argsort(vallog['Time'])
        for k in vallog.keys():
            vallog[k] = vallog[k][sequence]
    if 'TARGET' in vallog.keys():
        points = np.unique(vallog['TARGET'])
        full = vallog['TARGET']
        for k in vallog.keys():
            temp = vallog[k]
            newone,  newerr = [],  []
            for pp in points:
                subset = temp[np.where(full == pp)]
                newone.append(subset.mean())
                newerr.append(subset.std())
            vallog[k] = np.array(newone)
            errlog[k] = np.array(newerr)
    return vallog, errlog

class LoadAndorFile(QObject):
    finished = pyqtSignal()
    error = pyqtSignal(tuple)
    result = pyqtSignal(object)
    progress = pyqtSignal(int)
    def __init__(self,  fname,  bpp,  cray, report_to = None):
        super().__init__()
        # self.signals = WorkerSignals()
        fpath, shortname = os.path.split(fname)
        if "." in shortname:
            nameroot = ".".join(shortname.split(".")[:-1])
        else:
            nameroot = shortname
        self.fpath = fpath
        self.fname = fname
        self.bpp = bpp
        self.cray = cray
        self.nameroot = nameroot
        self.reply_here = report_to
        self.finished.connect(self.reply_here.increment)
    @pyqtSlot()
    def run(self):
        # print("Starting a file-loading thread.")
        # self.reply_here.debug_output(15)
        data, header = ReadAndor(self.fname)
        tadded_header = load_datheader(os.path.join(self.fpath, self.nameroot+".dat"))
        tadded_log = load_datlog(os.path.join(self.fpath, self.nameroot+".xas"))
        if self.cray > 0.0:
            data = RemoveCosmics(data, NStd = self.cray)
        if self.bpp is None:
            self.bpp = np.percentile(data, 10).mean()*0.99
        data -= self.bpp
        print("Thread has data with shape",  data.shape)
        # self.result.emit((data, tadded_header, tadded_log,  self.fname))  # Return the result of the processing
        # self.reply_here.accept_values([data, tadded_header, tadded_log,  self.fname])
        # self.reply_here.increment()
        self.results = [data, tadded_header, tadded_log,  self.fname]
        self.finished.emit()  # Done

class MergeCurves(QObject):
    def __init__(self, source, target, tpool = None, pbar = None,  mthreads = 1):
        super().__init__()
        xs1, ys1 = source[:,0], source[:,1]
        xs2, ys2 = target[:,0], target[:,1]
        step1 = xs1[1:] - xs1[:-1]
        step1 = np.concatenate([step1[:1], step1, step1[-1:]])
        lowlim1 = xs1 - 0.5*step1[:-1]
        highlim1 = xs1 + 0.5*step1[1:]
        # print('step1: ', step1.min(), step1.max(), step1.mean())
        # now we have low limits, high limits and middle for each bin, plus the bin size
        step2 = xs2[1:] - xs2[:-1]
        step2 = np.concatenate([step2[:1], step2, step2[-1:]])
        lowlim2 = xs2 - 0.5*step2[:-1]
        highlim2 = xs2 + 0.5*step2[1:]
        # print('step2: ', step2.min(), step2.max(), step2.mean())
        # now we can begin
        newy = np.zeros(len(ys2))
        lowest_m=0
        highest_m=len(ys1)-1
        self.arrays = [lowlim1, lowlim2, highlim1, highlim2, step1, step2, xs2, newy, ys1]
        self.consts = [lowest_m, highest_m]
        self.mutex = QMutex()
        self.datalist = []
#        if tpool is not None:
#            self.threadpool = tpool
#        else:
#            # self.threadpool = QThreadPool(self)
#            self.threadpool = CustomThreadpool(MAX_THREADS = mthreads)
        self.progbar = pbar
        self.counter = 0
        self.total = len(ys1)
        # self.progbar.setRange(self.counter,  self.total)
        self.target = target.copy()
        nthreads = mthreads
        limits = np.linspace(0, len(newy), nthreads+1).astype(int)
        self.runnables = []
        for nt in range(nthreads):
            self.runnables.append(CurveMergeWorker(limits[nt],  limits[nt+1], self.arrays,  self.consts, self))
    @pyqtSlot(int)
    def debug_output(self, somenum):
        print("Received a progress signal: ",  somenum)
    @pyqtSlot()
    def increment(self,  snum = 1):
        self.mutex.lock()
        self.counter += snum
        # self.progbar.setValue(self.counter)
        self.mutex.unlock()
    @pyqtSlot(object)
    def accept_values(self, results):
        self.mutex.lock()
        # print("Received values from thread, ",  results[3])
        self.datalist.append(results)
        self.mutex.unlock()
    def isitover(self):
        self.mutex.lock()
        self.finished = self.counter == self.total
        self.mutex.unlock()
    def runit(self):       
        for rr in self.runnables:
            rr.run()
        for rr in self.runnables:
            self.accept_values(rr.results)
    def postprocess(self):
        for r in self.datalist:
            beg,  end = r[0]
            data = r[1]
            self.target[beg:end, 1] += data[beg:end, 1]
        return self.target

class CurveMergeWorker(QObject):
    finished = pyqtSignal()
    def __init__(self, nbeg, nend, arrays, const, report_to = None):
        super().__init__()
        self.arrays = arrays
        self.consts = const
        self.nbeg = nbeg
        self.nend = nend
        self.reply_here = report_to
    @pyqtSlot()
    def run(self):
        lowlim1, lowlim2, highlim1, highlim2, step1, step2, xs2, newy, ys1 = self.arrays
        lowest_m, highest_m = self.consts
        for n in np.arange(self.nbeg, self.nend):
            m1 = np.argmax(highlim1 > lowlim2[n])
            m2 = np.argmax(lowlim1 > highlim2[n])
            if highlim1[0] > lowlim2[n]:
                m1 = 0
            elif highlim1[-1] < lowlim2[n]:
                m1 = len(highlim1) -1
            elif highlim1[-2] < lowlim2[n]:
                m1 = len(highlim1) -2
            if lowlim1[0] > highlim2[n]:
                m2 = 0
            elif lowlim1[-1] < highlim2[n]:
                m2 = len(lowlim1)-1
            elif lowlim1[-2] < highlim2[n]:
                m2 = len(lowlim1)-2
            # now m1 is the first bin in the source curve that overlaps with the current bin
            # and m2 is the first bin after that that doesn't
            for m in np.arange(m1-2,m2+2):
                if (m > highest_m) or (m < lowest_m):
                    continue
                if (highlim1[m] < lowlim2[n]) or (lowlim1[m] > highlim2[n]):
                    continue
                if lowlim1[m] <= lowlim2[n]:
                    if highlim1[m] > highlim2[n]:
                        newy[n] += ys1[m]*(highlim2[n]-lowlim2[n])/step1[m]
                    else:
                        newy[n] += ys1[m]*(highlim1[m]-lowlim2[n])/step1[m]
                else:
                    if highlim1[m] > highlim2[n]:
                        newy[n] += ys1[m]*(highlim2[n]-lowlim1[m])/step1[m]
                    else:
                        newy[n] += ys1[m]
                # factor = (highlim1[m] - max(lowlim2[n],lowlim1[m])/step1[m]
                # altfactor = (min(highlim2[n],highlim1[m]) - lowlim1[m])/step1[m]
            if not n%16:
                self.reply_here.increment(16)
        # acc_n = np.array(acc_n)
        # print('acc_n: ', acc_n.min(0), acc_n.max(0), acc_n.mean(0))
        # self.reply_here.accept_values([(self.nbeg, self.nend), np.column_stack([xs2,newy])])
        self.results = [(self.nbeg, self.nend), np.column_stack([xs2,newy])]
        self.finished.emit()

class LoadFileList(QObject):
    def __init__(self, flist, bpp,  cray, tpool = None, pbar = None,  mthreads =1):
        super().__init__()
        self.mutex = QMutex()
        if tpool is not None:
            self.threadpool = tpool
        else:
            # self.threadpool = QThreadPool(self)
            self.threadpool = CustomThreadpool(MAX_THREADS = mthreads)
        self.progbar = pbar
        self.bpp = bpp
        self.cray = cray
        self.counter = 0
        self.total = len(flist)
        self.flist = flist
        # self.progbar.setRange(self.counter,  self.total)
        self.names = []
        self.datalist = []
        self.headlist = []
        self.varloglist = []
        self.finished = False
        self.noproblems = False
        self.runnables = []
        # timer = QTimer(self)
        for name in self.flist:
            rr = LoadAndorFile(name, self.bpp,  self.cray,  report_to = self)
            self.runnables.append(rr)
            # rr.setAutoDelete(False)
            # rr.progress.connect(self.debug_output)
            # rr.result.connect(self.accept_values)
            # rr.finished.connect(self.increment)
    @pyqtSlot(int)
    def debug_output(self, somenum):
        print("Received a progress signal: ",  somenum)
    @pyqtSlot()
    def increment(self):
        self.mutex.lock()
        self.counter += 1
        # self.progbar.setValue(self.counter)
        self.mutex.unlock()
    @pyqtSlot(object)
    def accept_values(self, results):
        self.mutex.lock()
        print("Received values from thread, ",  results[3])
        self.datalist.append(results[0])
        self.headlist.append(results[1])
        self.varloglist.append(results[2])
        self.names.append(results[3])
        self.mutex.unlock()
    def isitover(self):
        self.mutex.lock()
        self.finished = self.counter == self.total
        self.mutex.unlock()
    def runit(self):
        for rr in self.runnables:
            self.threadpool.add_task(rr.run)
        self.threadpool.run()
        for rr in self.runnables:
            self.accept_values(rr.results)
    def postprocess(self):
        temps = []
        shortnames = []
        headers,  logs = [],  []
        nfiles,  times = 0,  []
        header,  vallog = {},  {}
        textheader = ""
        for n, tname in enumerate(self.names):
            nfiles += 1
            temps.append(self.datalist[n])
            headers.append(self.headlist[n])
            logs.append(self.varloglist[n])
            aaa, bbb = os.path.split(tname)
            shortnames.append(bbb)
        for hd in headers:
            hk = hd.keys()
            for k in hk:
                if k in header.keys():
                    header[k].append(hd[k])
                else:
                    header[k] = []
        for lg in logs:
            lk = lg.keys()
            for k in lk:
                if k in vallog.keys():
                    vallog[k] = np.concatenate([vallog[k],  lg[k]])
                else:
                    vallog[k] = lg[k].copy()
        if 'Time' in vallog.keys():
            sequence = np.argsort(vallog['Time'])
            for k in vallog.keys():
                vallog[k] = vallog[k][sequence]
        allnames = ",".join(shortnames)
        start,  end = 0, 60
        segment = allnames[start:end]
        while (len(segment) > 0):
            textheader += "# Filenames: " + segment + "\n"
            start += 60
            end += 60
            segment = allnames[start:end]
        textheader += "# Parameters: Minimum Maximum Mean Stddev\n"
        for hk in header.keys():
            tarr = np.array(header[hk])
            tmin,  tmax,  tmean,  tstd = tarr.min(),  tarr.max(),  tarr.mean(),  tarr.std()
            textheader += " ".join(['#', hk] + [str(round(x,4)) for x in [tmin, tmax, tmean, tstd]] + ['\n'])
        data = np.array(temps)# .sum(0)
        # it is better to return the files separately.
        # we can always sum them up outside of this function
        # or do something else to them first.
        return data,  textheader,  vallog,  shortnames

def simple_read(fname, bpp,  cray):
    fpath, shortname = os.path.split(fname)
    if "." in shortname:
        nameroot = ".".join(shortname.split(".")[:-1])
        extension = shortname.split(".")[-1]
    else:
        nameroot = shortname
        extension = ""
    if 'fits' in extension.lower():
        try:
            data, header = ReadFits(fname)
        except:
            print("Thought that " + fname + " was a FITS file, but ReadFits did not work...")
            data, header = ReadAndor(fname)
    if 'asc' in extension.lower():
        try:
            data, header = ReadAsc(fname)
        except:
            print("Thought that " + fname + " was an old 2D ASC file, but ReadAsc did not work...")
            data, header = ReadAndor(fname)
    else:
        data, header = ReadAndor(fname)
    tadded_header = load_datheader(os.path.join(fpath, nameroot+".dat"))
    tadded_log = load_datlog(os.path.join(fpath, nameroot+".xas"))
    if cray > 0.0:
        data = RemoveCosmics(data, NStd = cray)
    if bpp is None:
        bpp = np.percentile(data, 10).mean()*0.99
    data -= bpp
    print("Thread has data with shape",  data.shape)
    # self.result.emit((data, tadded_header, tadded_log,  self.fname))  # Return the result of the processing
    # self.reply_here.accept_values([data, tadded_header, tadded_log,  self.fname])
    # self.reply_here.increment()
    return data, tadded_header, tadded_log,  fname

def header_read(fname):
    fpath, shortname = os.path.split(fname)
    if "." in shortname:
        nameroot = ".".join(shortname.split(".")[:-1])
    else:
        nameroot = shortname
    tadded_header = load_datheader(os.path.join(fpath, nameroot+".dat"))
    tadded_log = load_datlog(os.path.join(fpath, nameroot+".xas"))
    return tadded_header, tadded_log

def postprocess(inputlist):
    datalist = inputlist[0]
    headlist = inputlist[1]
    varloglist = inputlist[2]
    names = inputlist[3]
    nfiles,  times = 0,  len(datalist)*[0.0]
    temps = []
    energy = []
    shortnames = []
    headers,  logs = [],  []
    header,  vallog = {},  {}
    textheader = ""
    for n, tname in enumerate(names):
        nfiles += 1
        temps.append(datalist[n])
        headers.append(headlist[n])
        logs.append(varloglist[n])
        aaa, bbb = os.path.split(tname)
        shortnames.append(bbb)
    for hd in headers:
        hk = hd.keys()
        for k in hk:
            if k in header.keys():
                header[k].append(hd[k])
            else:
                header[k] = [hd[k]]
    for lg in logs:
        lk = lg.keys()
        for k in lk:
            if k in vallog.keys():
                vallog[k] = np.concatenate([vallog[k],  lg[k]])
            else:
                vallog[k] = lg[k].copy()
    if 'Time' in vallog.keys():
        sequence = np.argsort(vallog['Time'])
        for k in vallog.keys():
            vallog[k] = vallog[k][sequence]
    allnames = ",".join(shortnames)
    start,  end = 0, 60
    segment = allnames[start:end]
    while (len(segment) > 0):
        textheader += "# Filenames: " + segment + "\n"
        start += 60
        end += 60
        segment = allnames[start:end]
    textheader += "# Parameters: Mean Minimum Maximum Stddev\n"
    for hk in header.keys():
        tarr = np.array(header[hk])
        if "hoton energy" in hk:
            energy.append(header[hk])
        if "Measuring time" in hk:
            times = tarr
        try:
            tmin,  tmax,  tmean,  tstd = tarr.min(),  tarr.max(),  tarr.mean(),  tarr.std()
        except:
            textheader += " ".join(['#', hk] + list(tarr) + ['\n'])
        else:
            textheader += " ".join(['#', hk] + [str(round(x,4)) for x in [tmean, tmin, tmax, tstd]] + ['\n'])
    energy = np.array(energy)
    data = np.array(temps)# .sum(0)
    # it is better to return the files separately.
    # we can always sum them up outside of this function
    # or do something else to them first.
    return data,  textheader,  vallog,  shortnames, (nfiles, times), energy

def load_file(fname, bpp, cray, poly = None):
    fpath, shortname = os.path.split(fname)
    if "." in shortname:
        nameroot = ".".join(shortname.split(".")[:-1])
    else:
        nameroot = shortname
    data, header = ReadAndor(fname)
    tadded_header = load_datheader(os.path.join(fpath, nameroot+".dat"))
    tadded_log = load_datlog(os.path.join(fpath, nameroot+".xas"))
    # data = data[:,lcut:hcut]
    if cray > 0.0:
        data = RemoveCosmics(data, NStd = cray)
    if bpp is None:
        bpp = np.percentile(data, 10).mean()*0.99
    data -= bpp
    return data, tadded_header,  tadded_log

def load_filelist(flist,  bpp,  cray,  poly):
    if len(flist) == 0:
        return None,  None,  None
    else:
        temps = []
        shortnames = []
        headers,  logs = [],  []
        header,  vallog = {},  {}
        textheader = ""
        for tname in flist:
            tdata, theader, tlog = load_file(tname, bpp, cray, poly = poly)
            temps.append(tdata)
            headers.append(theader)
            logs.append(tlog)
            aaa, bbb = os.path.split(tname)
            shortnames.append(bbb)
        for hd in headers:
            hk = hd.keys()
            for k in hk:
                if k in header.keys():
                    header[k].append(hd[k])
                else:
                    header[k] = []
        for lg in logs:
            lk = lg.keys()
            for k in lk:
                if k in vallog.keys():
                    vallog[k] = np.concatenate([vallog[k],  lg[k]])
                else:
                    vallog[k] = lg[k].copy()
        if 'Time' in vallog.keys():
            sequence = np.argsort(vallog['Time'])
            for k in vallog.keys():
                vallog[k] = vallog[k][sequence]
        allnames = ",".join(shortnames)
        start,  end = 0, 60
        segment = allnames[start:end]
        while (len(segment) > 0):
            textheader += "# Filenames: " + segment + "\n"
            start += 60
            end += 60
            segment = allnames[start:end]
        for hk in header.keys():
            tarr = np.array(header[hk])
            tmin,  tmax,  tmean,  tstd = tarr.min(),  tarr.max(),  tarr.mean(),  tarr.std()
            textheader += " ".join(['#', hk] + [str(round(x,4)) for x in [tmin, tmax, tmean, tstd]] + ['\n'])
        data = np.array(temps)# .sum(0)
        # it is better to return the files separately.
        # we can always sum them up outside of this function
        # or do something else to them first.
        return data,  textheader,  vallog


#### Object-Oriented part

class SimpleCore(QObject):
    cleared = pyqtSignal()
    loaded = pyqtSignal()
    fileparams = pyqtSignal(object)
    logmessage = pyqtSignal(str)
    finished_fitting = pyqtSignal()
    finished_overplot = pyqtSignal()
    finished_rixsmap = pyqtSignal()
    finished_merge = pyqtSignal()
    def __init__(self, master, boxes, logger,  progress_bar = None,  table = None, max_threads = 1, 
                                startpath = expanduser("~")):
        if master is None:
            super().__init__()
        else:
            super().__init__(master)
        self.parlist = []
        self.pardict = {}
        self.maxthreads = max_threads
        self.log = logger
        self.logmessage.connect(self.log.put_logmessage)
        self.progbar = progress_bar
        self.table_obj = table
        self.temp_path = startpath
        self.profiles = []
        self.reduced_profiles = []
        self.lastreduction = 1.0
        self.fullnames = []
        self.shortnames = []
        self.timedata = None
        self.units = []
        self.energies = []
        self.prof_numbers = []
        self.mplot_curves = []
        self.mplot_labels =[]
        self.mplot_override = []
        self.mplot_fits = []
        self.mplot_fitparams = []
        self.map2D = None
        self.overplotworked = False
        self.rixs_worked = False
        self.fitsworked = False
        self.mergeworked = False
        self.map2Dplotax = [(1, 2048,),  (1, 2048)]
        self.prof_count = 0
        self.legpos = 0
        self.retvals = []
    def assign_boxes(self, boxes):
        self.boxes = boxes
        for db in boxes:
            names, values = db.returnValues()
            self.parlist += names
            for nam in names:
                self.pardict[nam] = values[nam]
            db.values_changed.connect(self.make_params_visible)
        self.make_params_visible()
    @pyqtSlot()
    def make_params_visible(self):
        for db in self.boxes:
            names, values = db.returnValues()
            for n in names:
                if not n in self.parlist:
                    self.parlist.append(n)
                self.pardict[n] = values[n]
        for k in self.parlist:
            val = self.pardict[k]
            if len(val) == 1:
                val = val[0]
            if 'cuts' in k:
                self.cuts = val
            elif 'eline' in k:
                self.eline = val
            elif 'bkg_perc' in k:
                self.bkg_perc = val
            elif 'offmax' in k:
                self.offmax = val
            elif 'redfac' in k:
                self.redfac = val
            elif 'legpos' in k:
                self.legpos = val
            elif 'smear' in k:
                self.smear = val
    def reduce_profiles(self):
        if len(self.profiles) == len(self.reduced_profiles):
            if self.redfac == self.lastreduction:
                return False
        self.reduced_profiles = []
        if self.redfac == 1.0:
            for p in self.profiles:
                self.reduced_profiles.append(p)
        else:
            for p in self.profiles:
                steps = len(p)
                newsteps = int(round(steps/self.redfac))
                target = np.zeros([newsteps, 2])
                target[:, 0] = np.linspace(p[:, 0].min(), p[:, 0].max(), newsteps)
                newone = merge2curves(p,  target)
                self.reduced_profiles.append(newone)
        self.lastreduction = self.redfac
        return True
    @pyqtSlot()
    def take_table_values(self):
        self.retvals = self.table_obj.return_values()
    def logger(self,  message):
        now = time.gmtime()
        timestamp = ("ProfileCore "
                     +"-".join([str(tx) for tx in [now.tm_mday, now.tm_mon, now.tm_year]])
                     + ',' + ":".join([str(ty) for ty in [now.tm_hour, now.tm_min, now.tm_sec]]) + '| ')
        self.logmessage.emit(timestamp + message)
    @pyqtSlot(object)
    def load_profiles(self,  flist):
        fnames, snames, profiles,  envals,  units = [], [], [], [], []
        for fnum, fname in enumerate(flist):
            self.temp_path, short_name = os.path.split(fname)
            try:
                p, e, u = read_1D_curve(fname)
            except:
                self.logger("Could not parse file:" + str(fname))
            else:
                self.prof_numbers.append(self.prof_count+fnum)
                fnames.append(fname)
                profiles.append(p)
                envals.append(e[0])
                units.append(u)
                snames.append(short_name)
        self.fullnames += fnames
        self.shortnames += snames
        self.profiles += profiles
        self.energies += envals
        self.units += units
        self.prof_count += len(fnames)
        self.loaded.emit()
        self.fileparams.emit([snames, profiles, envals, units])
        return snames, profiles, envals, units
    @pyqtSlot()
    def clear_profiles(self):
        self.fullnames = []
        self.shortnames = []
        self.profiles = []
        self.reduced_profiles = []
        self.energies = []
        self.units = []
        self.timedata = None
        self.prof_numbers = []
        self.prof_count = 0
        self.cleared.emit()
    def autofit_many(self):
        self.fitsworked = False
        if len(self.retvals) < 1:
            self.finished_fitting.emit()
            return None
        else:
            self.reduce_profiles()
            nums, Ei, xmin, xmax, xunits = [], [], [], [], []
            for nr in range(len(self.retvals)):
                nums.append(self.retvals[nr][0])
                Ei.append(self.retvals[nr][1])
                xmin.append(self.retvals[nr][2])
                xmax.append(self.retvals[nr][3])
                xunits.append(self.retvals[nr][4])
            chan,  en,  entran = 0, 0, 0
            for un in xunits:
                if un == 0:
                    chan+=1
                elif un == 1:
                    entran+=1
                elif un == 2:
                    en +=1
            iunits = np.array(xunits)
            if entran >0:
                self.logger("Plotting energy transfer. Other units will be discarded.")
                crit = np.where(iunits ==1)
                unit = 1
                fwhm_guess = 0.05
            elif en > 0:
                self.logger("Plotting energy. Other units will be discarded.")
                crit = np.where(iunits ==2)
                unit = 2
                fwhm_guess = 0.05
            else:
                self.logger("Plotting raw channels. This is most likely a debugging run.")
                crit = np.where(iunits ==0)
                unit = 0
                fwhm_guess = 4.0
            nums =np.array(nums)[crit]
            xvals =np.array(Ei)[crit]
            xmin =np.array(xmin)[crit]
            xmax =np.array(xmax)[crit]
            curves = []
            bkgs = []
            labels = []
            for rn, num in enumerate(nums):
                temp = self.reduced_profiles[num].copy()
                bkg = temp[np.where(temp[:, 1] < np.percentile(temp[:,1], self.bkg_perc))]
                temp = temp[np.where(temp[:, 0] >= xmin[rn])]
                temp = temp[np.where(temp[:, 0] <= xmax[rn])]
                curves.append(temp)
                bkgs.append(bkg)
                labels.append(str(xvals[rn]) + " eV, " + self.shortnames[num])
                # step = cur[np.where( cur[:,1] < np.percentile(cur[:,1], 60.0) )][:,1].std()
            width,  widtherr = [], []
            area,  areaerr = [], []
            peakcurves = []
            for n in range(len(curves)):
                temp = curves[n]
                bkg = bkgs[n]
                fit, peakshape,  chi2 = elastic_line_anyx(temp,  bkg,  init_fwhm = fwhm_guess)
                if fit is None:
                    width.append(-1.0)
                    widtherr.append(-1.0)
                    area.append(0.0)
                    areaerr.append(-1.0)
                    peakcurves.append(np.column_stack([np.arange(10),  np.zeros(10)]))
                else:
                    peak_area = fit[0][0]*abs(fit[0][1])/gauss_denum*(2*np.pi)**0.5
                    peak_area_error = peak_area*((fit[1][0]/fit[0][0])**2 + (fit[1][1]/fit[0][1])**2)**0.5
                    width.append(abs(round(fit[0][1],3)))
                    widtherr.append(abs(round(fit[1][1],3)))
                    area.append(round(peak_area,3))
                    areaerr.append(abs(round(peak_area_error,3)))
                    peakcurves.append(peakshape)
            self.mplot_curves = curves
            self.mplot_labels = labels
            self.mplot_fits = peakcurves
            self.mplot_fitparams = [nums, width,  widtherr,  area,  areaerr]
            if unit ==1:
                self.mplot_override = ["Energy transfer [eV]", ""]
            elif unit ==2:
                self.mplot_override = ["Energy [eV]", ""]
            else:
                self.mplot_override = ["Channels", ""]
            self.fitsworked = True
            self.finished_fitting.emit()
            return "Done"
    def fit_many(self):
        self.fitsworked = False
        if len(self.retvals) < 1:
            self.finished_fitting.emit()
            return None
        else:
            self.reduce_profiles()
            nums, Ei, xmin, xmax, xunits = [], [], [], [], []
            for nr in range(len(self.retvals)):
                nums.append(self.retvals[nr][0])
                Ei.append(self.retvals[nr][1])
                xmin.append(self.retvals[nr][2])
                xmax.append(self.retvals[nr][3])
                xunits.append(self.retvals[nr][4])
            chan,  en,  entran = 0, 0, 0
            for un in xunits:
                if un == 0:
                    chan+=1
                elif un == 1:
                    entran+=1
                elif un == 2:
                    en +=1
            iunits = np.array(xunits)
            if entran >0:
                self.logger("Plotting energy transfer. Other units will be discarded.")
                crit = np.where(iunits ==1)
                unit = 1
                fwhm_guess = 0.05
            elif en > 0:
                self.logger("Plotting energy. Other units will be discarded.")
                crit = np.where(iunits ==2)
                unit = 2
                fwhm_guess = 0.05
            else:
                self.logger("Plotting raw channels. This is most likely a debugging run.")
                crit = np.where(iunits ==0)
                unit = 0
                fwhm_guess = 4.0
            nums =np.array(nums)[crit]
            xvals =np.array(Ei)[crit]
            xmin =np.array(xmin)[crit]
            xmax =np.array(xmax)[crit]
            curves = []
            bkgs = []
            labels = []
            for rn, num in enumerate(nums):
                temp = self.reduced_profiles[num].copy()
                bkg = temp[np.where(temp[:, 1] < np.percentile(temp[:,1], self.bkg_perc))]
                temp = temp[np.where(temp[:, 0] >= xmin[rn])]
                temp = temp[np.where(temp[:, 0] <= xmax[rn])]
                curves.append(temp)
                bkgs.append(bkg)
                labels.append(str(xvals[rn]) + " eV, " + self.shortnames[num])
                # step = cur[np.where( cur[:,1] < np.percentile(cur[:,1], 60.0) )][:,1].std()
            width,  widtherr = [], []
            area,  areaerr = [], []
            peakcurves = []
            for n in range(len(curves)):
                temp = curves[n]
                bkg = bkgs[n]
                fit, peakshape, chi2 = elastic_line_anyx(temp,  bkg, olimits = self.eline,  init_fwhm = fwhm_guess)
                if fit is None:
                    width.append(-1.0)
                    widtherr.append(-1.0)
                    area.append(0.0)
                    areaerr.append(-1.0)
                    peakcurves.append(np.column_stack([np.arange(10),  np.zeros(10)]))
                else:
                    peak_area = fit[0][0]*abs(fit[0][1])/gauss_denum*(2*np.pi)**0.5
                    peak_area_error = peak_area*((fit[1][0]/fit[0][0])**2 + (fit[1][1]/fit[0][1])**2)**0.5
                    width.append(abs(round(fit[0][1],3)))
                    widtherr.append(abs(round(fit[1][1],3)))
                    area.append(round(peak_area,3))
                    areaerr.append(abs(round(peak_area_error,3)))
                    peakcurves.append(peakshape)
            self.mplot_curves = curves
            self.mplot_labels = labels
            self.mplot_fits = peakcurves
            self.mplot_fitparams = [nums, width,  widtherr,  area,  areaerr]
            if unit ==1:
                self.mplot_override = ["Energy transfer [eV]", ""]
            elif unit ==2:
                self.mplot_override = ["Energy [eV]", ""]
            else:
                self.mplot_override = ["Channels", ""]
            self.fitsworked = True
            self.finished_fitting.emit()
            return "Done"    
    def manual_merge(self):
        self.mergeworked = False
        if len(self.retvals) < 1:
            self.finished_merge.emit()
            return None
        else:
            self.reduce_profiles()
            nums, Ei, xmin, xmax, xunits = [], [], [], [], []
            for nr in range(len(self.retvals)):
                nums.append(self.retvals[nr][0])
                Ei.append(self.retvals[nr][1])
                xmin.append(self.retvals[nr][2])
                xmax.append(self.retvals[nr][3])
                xunits.append(self.retvals[nr][4])
            chan,  en,  entran = 0, 0, 0
            for un in xunits:
                if un == 0:
                    chan+=1
                elif un == 1:
                    entran+=1
                elif un == 2:
                    en +=1
            iunits = np.array(xunits)        
            if entran >0:
                self.logger("Merging energy transfer. Other units will be discarded.")
                crit = np.where(iunits ==1)
                self.merged_units = ["Energy transfer [eV]",  "Intensity [arb. units]"]
                unit = 1
                units = "Energy Transfer [eV]"
                fwhm_guess = 0.05
            elif en > 0:
                self.logger("Merging energy. This result is not likely to be useful.")
                crit = np.where(iunits ==2)
                self.merged_units = ["Energy [eV]",  "Intensity [arb. units]"]
                unit = 2
                units = "Energy [eV]"
                fwhm_guess = 0.05
            else:
                self.logger("Merging raw channels. This is probably a bad idea.")
                crit = np.where(iunits ==0)
                self.merged_units = ["Detector channels [pixel]",  "Intensity [arb. units]"]
                units = "Detector channels"
                unit = 0
                fwhm_guess = 4.0
            nums =np.array(nums)[crit]
            xvals =np.array(Ei)[crit]
            xmin =np.array(xmin)[crit]
            xmax =np.array(xmax)[crit]        
            minx, maxx, stepx = 1e5,-1e5,-1.0    
            curves = []
            labels = []
            for rn, num in enumerate(nums):
                dat = self.reduced_profiles[num].copy()
                dat = dat[np.where(dat[:, 0] > xmin[rn])]
                dat = dat[np.where(dat[:, 0] < xmax[rn])]
                minx = min(minx, dat[:,0].min())
                maxx = max(maxx, dat[:,0].max())
                step = (dat[1:,0] - dat[:-1,0]).mean()
                stepx = max(stepx, step)
                curves.append(dat)
            newx = np.arange(minx, maxx + 0.1*stepx, stepx)
            target = np.zeros((len(newx),2))
            target[:,0] = newx
            for n in range(len(curves)):
                target = merge2curves(curves[n],target)
            self.merged_curve = target
            self.merged_units = units
            self.merged_energy = str(xvals.mean())
            self.mergeworked = True
            # now we try to add the fitting            
            self.fullnames += ["No file"]
            self.shortnames += ["Merged data"]
            self.profiles += [target]
            self.energies += [self.merged_energy]
            self.units += [units]
            self.prof_count += 1
            self.finished_merge.emit()
            # self.finished_fitting.emit()
            return target
    @pyqtSlot(str)
    def save_merged_profile(self, fname):
        if self.merged_curve is None:
            self.logger("There is no merged curve to be saved.")
            return None
        else:
            WriteEnergyProfile(fname, self.merged_curve, [])
            return True
            # outname = os.path.join(self.temp_path, self.temp_name+"_HISTOGRAM.txt")
            # WriteEnergyProfile(outname, target, [])
    def multiplot(self):
        self.overplotworked = False
        if len(self.retvals) < 1:
            self.finished_overplot.emit()
            return None
        else:
            self.reduce_profiles()
            nums, Ei, xmin, xmax, xunits = [], [], [], [], []
            for nr in range(len(self.retvals)):
                nums.append(self.retvals[nr][0])
                Ei.append(self.retvals[nr][1])
                xmin.append(self.retvals[nr][2])
                xmax.append(self.retvals[nr][3])
                xunits.append(self.retvals[nr][4])
            chan,  en,  entran = 0, 0, 0
            for un in xunits:
                if un == 0:
                    chan+=1
                elif un == 1:
                    entran+=1
                elif un == 2:
                    en +=1
            iunits = np.array(xunits)
            if entran >0:
                self.logger("Plotting energy transfer. Other units will be discarded.")
                crit = np.where(iunits ==1)
                unit = 1
            elif en > 0:
                self.logger("Plotting energy. Other units will be discarded.")
                crit = np.where(iunits ==2)
                unit = 2
            else:
                self.logger("Plotting raw channels. This is most likely a debugging run.")
                crit = np.where(iunits ==0)
                unit = 0
            nums =np.array(nums)[crit]
            xvals =np.array(Ei)[crit]
            xmin =np.array(xmin)[crit]
            xmax =np.array(xmax)[crit]
            curves = []
            labels = []
            for rn, num in enumerate(nums):
                temp = self.reduced_profiles[num].copy()
                temp = temp[np.where(temp[:, 0] > xmin[rn])]
                temp = temp[np.where(temp[:, 0] < xmax[rn])]
                curves.append(temp)
                labels.append(str(xvals[rn]) + " eV, " + self.shortnames[num])
                # step = cur[np.where( cur[:,1] < np.percentile(cur[:,1], 60.0) )][:,1].std()
            for n in range(len(curves)):
                xaxis = curves[n][:,0]
                xstep = xaxis[1:] - xaxis[:-1]
                xstep = np.concatenate([xstep[:1], xstep])
                norm = (np.abs(curves[n][:,1]) * xstep).sum()
                # norm = curves[n][:,1].sum()
                curves[n][:,1] /= norm
                shift = np.percentile(curves[n][:,1], self.bkg_perc) - 0.01
                curves[n][:,1] -= shift
                # curves[n] = curves[n][np.where(curves[n][:,1] > 0.0)]
                # curves[n][:,1] *= 100.0
            self.mplot_curves = curves
            self.mplot_labels = labels
            if unit ==1:
                self.mplot_override = ["Energy transfer [eV]", ""]
            elif unit ==2:
                self.mplot_override = ["Energy [eV]", ""]
            else:
                self.mplot_override = ["Channels", ""]
            self.overplotworked = True
            self.finished_overplot.emit()
            return "Done"
    def rixsmap(self):
        self.rixs_worked = False
        if len(self.retvals) <= 1:
            self.finished_rixsmap.emit()
            return None
        else:
            self.reduce_profiles()
            nums, Ei, xmin, xmax, xunits = [], [], [], [], []
            for nr in range(len(self.retvals)):
                nums.append(self.retvals[nr][0])
                Ei.append(self.retvals[nr][1])
                xmin.append(self.retvals[nr][2])
                xmax.append(self.retvals[nr][3])
                xunits.append(self.retvals[nr][4])
            counter = 0
            for e in Ei:
                if e > 0.0:
                    counter += 1
            if not (counter > 0):
                return None
            chan,  en,  entran = 0, 0, 0
            for un in xunits:
                if un == 0:
                    chan+=1
                elif un == 1:
                    entran+=1
                elif un == 2:
                    en +=1
            iunits = np.array(xunits)
            if entran >0:
                self.logger("Plotting energy transfer. Other units will be discarded.")
                crit = np.where(iunits ==1)
            elif en > 0:
                self.logger("Plotting energy. Other units will be discarded.")
                crit = np.where(iunits ==2)
            else:
                self.logger("Plotting raw channels. This is most likely a debugging run.")
                crit = np.where(iunits ==0)
            nums =np.array(nums)[crit]
            xvals =np.array(Ei)[crit]
            xmin =np.array(xmin)[crit]
            xmax =np.array(xmax)[crit]
            ocurves, curves = [], []
            step = 0.0
            ymin, ymax, ystep = 1e5, -1e5, 100.0
            for rn, num in enumerate(nums):
                temp = self.reduced_profiles[num].copy()
                temp = temp[np.where(temp[:, 0] > xmin[rn])]
                temp = temp[np.where(temp[:, 0] < xmax[rn])]
                ocurves.append(temp)
            for n in range(len(ocurves)):
                cxaxis = ocurves[n][:,0]
                cxstep = cxaxis[1:] - cxaxis[:-1]
                cxstep = np.concatenate([cxstep[:1], cxstep])
                ymin = min(ymin, cxaxis.min())
                ymax = max(ymax, cxaxis.max())
                ystep = min(ystep, cxstep.max())
                norm = (ocurves[n][:,1] * cxstep).sum()
                ocurves[n][:,1] /= norm
                shift = np.percentile(ocurves[n][:,1], 75.0) - 0.01
                ocurves[n][:,1] -= shift
                ocurves[n] = ocurves[n][np.where(ocurves[n][:,1] > 0.0)]
                ocurves[n][:,1] *= 100.0
            sequence = np.argsort(xvals)
            # print("RIXSMAP energies: ", xvals)
            # print("RIXSMAP sequence: ", sequence)
            xvals = xvals[sequence]
            crit = np.where(xvals > 0.0)
            # print(crit)
            xvals = xvals[crit]
            for ns, se in enumerate(sequence):
                # print(se)
                if ns in crit[0]:
                    curves.append(ocurves[se])
            # curves = curves[crit]
            # now we need to define a grid for the map
            yaxis = np.arange(ymin, ymax+0.1*ystep, ystep)
            npixy = len(yaxis)
            xmin = xvals.min()
            xmax = xvals.max()
            xstep = max((xvals[1:] - xvals[:-1]).min(), 0.05)
            xaxis = np.arange(xmin - xstep, xmax + xstep*1.01, xstep/5.0)
            xmask = np.zeros(len(xaxis))
            npixx = len(xaxis)
            # define the 2D arrays as needed
            map_array = np.zeros((npixy,npixx)).astype(np.float64)
            # assign correct values
            # mcurves = []
            for n in range(len(curves)):
                xcrit = np.abs(xaxis - xvals[n])
                pos, = np.where(xcrit == xcrit.min())
                target = np.zeros((npixy,2))
                target[:,0] = yaxis.copy()
                print("RIXSMAP curve min/max:",curves[n][:,1].min(),curves[n][:,1].max())
                yvals = merge2curves(curves[n], target)[:,1]
                # mcurves.append(merge2curves(curves[n], target))
                map_array[:,pos] = yvals.reshape(map_array[:,pos].shape)
                print("RIXSMAP pos, posshape: ", pos, map_array[:,pos].shape)
                # map_array[:len(curves[n]),pos] = curves[n][:,1].reshape(map_array[:len(curves[n]),pos].shape)
                xmask[pos] = 1
            # apply smearing
            # plot1D_sliders(mcurves, [], fig = self.single_figure, text = 'Results comparison.\n', label_override = ["Energy transfer [eV]", ""], curve_labels = labels[:len(curves)])
            # return None
            # print("RIXSMAP array min/max:",map_array.min(),map_array.max())
            virt_array = np.zeros(map_array.shape).astype(np.float64)
            weight_array = np.zeros(len(xaxis))
            # smearwidth = 2.0 # meV
            smearwidth = self.smear
            gridstep = xstep/5.0
            width = int(math.ceil(smearwidth/gridstep))
            # print("RIXSMAP step, gridstep, width", xstep, gridstep, width)
            for n in range(len(xaxis)):
                if xmask[n]:
                    w_axis = np.zeros(len(xaxis))
                    w_axis[n] = 1.0
                    neglim, poslim = 0,0
                    for s in range(1,width):
                        if n-s >= 0:
                            neglim = s
                            if xmask[n-s]:
                                break
                    for s in range(1,neglim):
                        if n-s >= 0:
                            w_axis[n-s] = 1.0 - s/float(neglim)
                    for s in range(1,width):
                        if n+s < len(xmask):
                            poslim = s
                            if xmask[n+s]:
                                break
                    for s in range(1,poslim):
                        if n+s < len(xmask):
                            w_axis[n+s] = 1.0 - s/float(poslim)
                    for s in range(len(xaxis)):
                        if w_axis[s] > 0.0:
                            if xmask[s] > 0.0:
                                virt_array[:,s] = map_array[:,s].copy()
                                weight_array[s] = 1.0
                            else:
                                virt_array[:,s] += w_axis[s]*(map_array[:,n].copy())
                                weight_array[s] += w_axis[s]
            print("RIXSMAP weigth array: ", weight_array)
            for n in range(len(xaxis)):
                if weight_array[n] > 0.0:
                    virt_array[:,n] /= weight_array[n]
            # plot!
            # plot2D_sliders([virt_array, map_array], [(yaxis[0],yaxis[-1]), (xaxis[0], xaxis[-1])], fig = self.single_figure, text = "")
            self.map2D = [virt_array, map_array]
            self.map2Dplotax = [(ymin,ymax), (xmin, xmax)]
            self.rixs_worked = True
            self.finished_rixsmap.emit()
            return "Done"

class AdlerCore(QObject):
    fittingresult = pyqtSignal(object)
    energyresult = pyqtSignal(object)
    curvatureresult = pyqtSignal(object)
    segmentresult = pyqtSignal(object)
    historesult = pyqtSignal(object)
    logmessage = pyqtSignal(str)
    params_fitting = pyqtSignal(object)
    params_energy = pyqtSignal(object)
    finished_preprocess = pyqtSignal()
    finished_process = pyqtSignal()
    finished_offsets = pyqtSignal()
    finished_2D = pyqtSignal()
    finished_poly = pyqtSignal()
    finished_fft = pyqtSignal()
    finished_calcfft = pyqtSignal()
    did_nothing = pyqtSignal()
    def __init__(self, master, boxes, logger,  progress_bar = None,  thr_exit = None, max_threads = 1, 
                                startpath = expanduser("~")):
        if master is None:
            super().__init__()
        else:
            super().__init__(master)
        self.master = master
        self.parlist = []
        self.pardict = {}
        self.maxthreads = max_threads
        self.log = logger
        self.logmessage.connect(self.log.put_logmessage)
        self.progbar = progress_bar
        self.thread_exit = thr_exit
        self.temp_path = startpath
        self.temp_name = ""
        # self.threadpool = QThreadPool(self)
        # self.threadpool = QThreadPool.globalInstance()
        self.threadpool = CustomThreadpool(MAX_THREADS = max_threads)
        # self.threadpool.setMaxThreadCount(max_threads)
        self.data,  self.header,  self.logvals,  self.logvalnames = None,  None,  None,  None
        self.processing_history = []
        self.data2D = None
        self.original_data2D = None
        self.corrected_data2D = None
        self.raw_profile= None
        self.component_profiles = []
        self.component_rawprofiles = []
        self.summed_rawprofile = None
        self.timedata = None
        self.summed_adjusted_rawprofile = None
        self.name_as_segments = []
        self.fft_plots =[]
        self.energies = []
        self.tdb_profile = None
        self.energy_profile = None
        self.fitted_peak_energy = None
        self.fitting_params_energy = None
        self.fitting_textstring_energy = None
        self.fitted_peak_channels = None
        self.fitting_params_channels = None
        self.fitting_textstring_channels = None
        self.plotax,  self.plotax1D = None,  None
        self.curvature,  self.curvature_fit,  self.curvature_params = None,  None,  None
        self.segment_plot = None
        self.histogram_plot = None
        self.curvature_corrected = False
        self.fft_applied = False
    def assign_boxes(self,  boxes):
        self.boxes = boxes
        for db in boxes:
            names, values = db.returnValues()
            self.parlist += names
            for nam in names:
                self.pardict[nam] = values[nam]
            db.values_changed.connect(self.make_params_visible)
        self.make_params_visible()
    def get_external_params(self, newdict):
        templist = [str(x) for x in self.pardict.keys()]
        for k in self.parlist:
            try:
                val = newdict[k]
            except:
                continue
            else:
                if k in templist:
                    self.pardict[k] = val
    @pyqtSlot()
    def make_params_visible(self):
        for db in self.boxes:
            names, values = db.returnValues()
            for n in names:
                if not n in self.parlist:
                    self.parlist.append(n)
                self.pardict[n] = values[n]
        for k in self.parlist:
            val = self.pardict[k]
            if len(val) == 1:
                val = val[0]
            if 'cuts' in k:
                self.cuts = val
            elif 'bpp' in k:
                self.bpp = val
            elif 'tdb' in k:
                self.tdb_fact = val
            elif 'cray' in k:
                self.cray = val
            elif 'ffts' in k:
                self.ffts = val
            elif 'poly' in k:
                self.poly = val
            elif 'eline' in k:
                self.eline = val
            elif 'bkg_perc' in k:
                self.bkg_perc = val
            elif 'segsize' in k:
                self.segsize = val
            elif 'eVpair1' in k:
                self.eVpair1 = val
            elif 'eVpair2' in k:
                self.eVpair2 = val
            elif 'redfac' in k:
                self.redfac = val
    def thread_start(self, target_function,  args = []):
        # 1 - create Worker and Thread inside the Form
        self.obj = Worker()  # no parent!
        self.thread = QThread()  # no parent!
        # 2 - Connect Worker`s Signals to Form method slots to post data.
        # self.obj.intReady.connect(self.onIntReady)
        # 3 - Move the Worker object to the Thread object
        self.obj.moveToThread(self.thread)
        # 4 - Connect Worker Signals to the Thread slots
        self.obj.finished.connect(self.thread.quit)
        # 5 - Connect Thread started signal to Worker operational slot method
        self.thread.started.connect(self.obj.run_repeated)
        # custom config
        self.obj.assignFunction(target_function)
        self.obj.assignArgs(args)
        self.obj.set_waittime(0.1)
        #
        if self.progbar is not None:
            self.progbar.setRange(0, len(args))
            self.obj.intReady.connect(self.progbar.setValue)
        # * - Thread finished signal will close the app if you want!
        self.thread.finished.connect(self.thread_exit)
        # 6 - Start the thread
        self.thread.start()
    @pyqtSlot()
    def save_last_params(self, lastfunction = None):
        try:
            source = open(os.path.join(expanduser("~"),'.ADLERcore.txt'), 'w')
        except:
            return None
        else:
            source.write('Lastdir: '+str(self.temp_path) + '\n')
            source.write('Lastfile: '+str(self.temp_name) + '\n')
            for kk in self.parlist:
                source.write(" ".join([str(u) for u in [kk, self.pardict[kk] ]]) + '\n')
            if not lastfunction == None:
                source.write('Last function called: ' + str(lastfunction) + '\n')
            source.close()
    def load_last_params(self):
        try:
            source = open(os.path.join(expanduser("~"),'.ADLERcore.txt'), 'r')
        except:
            return None
        else:
            for line in source:
                toks = line.split()
                if len(toks) > 1:
                    if toks[0] == 'Lastdir:':
                        try:
                            self.temp_path = toks[1]
                        except:
                            pass
            source.close()
    def logger(self, message):
        now = time.gmtime()
        timestamp = ("AdlerCore "
                     +"-".join([str(tx) for tx in [now.tm_mday, now.tm_mon, now.tm_year]])
                     + ',' + ":".join([str(ty) for ty in [now.tm_hour, now.tm_min, now.tm_sec]]) + '| ')
        self.logmessage.emit(timestamp + message)
    def read_tdb_profile(self,  fname = 'BKG_OVERRIDE.txt'):
        try:
            source = open(fname, 'r')
        except:
            self.logger("Did not find an override file BKG_OVERRIDE.txt - using built-in background from June 2020.")
            source = open(resource_path('Structured_bkg_per_second.dat'), 'r')
        else:
            self.logger("Loading background per channel per second from BKG_OVERRIDE.txt")
        arr = []
        for line in source:
            arr.append([float(x) for x in line.split()])
        source.close()
        self.tdb_profile = np.array(arr)
    def load_files(self,  flist):
        if len(flist) > 0:
            self.current_flist = flist
            self.temp_path = os.path.split(flist[0])[0]
            self.name_as_segments = simplify_number_range(flist)
            self.temp_name = "_".join(['Merged']+self.name_as_segments)
            # the_object = LoadFileList(flist,  self.bpp, self.cray,  self.threadpool,  self.progbar)
            datalist = []
            headlist = []
            varloglist = []
            names = []
            for fname in flist:
                data, tadded_header, tadded_log,  fname = simple_read(fname, self.bpp,  self.cray)
                datalist.append(data)
                headlist.append(tadded_header)
                varloglist.append(tadded_log)
                names.append(fname)
            datashape = data.shape
            self.det_width = datashape[1]
            self.det_length = datashape[0]
            # the_object = LoadFileList(flist,  self.bpp, self.cray,  None,  self.progbar,  mthreads = self.maxthreads)
            # the_object.runit()
            # the_object.isitover()
            self.data,  self.header,  self.logvals,  self.logvalnames = None,  None,  None,  None
            self.shortnames = None
            self.energies = []
            self.timedata = (1, np.zeros(1))
            # self.logger("LoadFileList -> have all files been processed? " + str(the_object.finished))
            # self.logger("LoadFileList -> did all threads terminate? " + str(the_object.noproblems))
            self.data,  self.header,  self.logvals,  self.shortnames,  self.timedata,  self.energies = postprocess([datalist, headlist, varloglist, names])
            self.logvalnames = self.logvals.keys()
            self.data = self.data[:, :, self.cuts[0]:self.cuts[1]]
            self.det_width = self.data.shape[2]
            # self.data,  self.header,  self.logvals = load_filelist(flist,  self.bpp,  self.cray,  self.poly)
            # self.plotax = [(1,self.det_length), self.cuts]
            self.plotax = [(1,self.det_length), self.cuts]
            scalefac = 1.0/self.det_width * self.tdb_fact
            if (self.tdb_profile is not None) and self.tdb_fact > 0.0 :
                temp = self.tdb_profile[:, 1].copy()
                temp *= scalefac
                if not len(temp) == self.data.shape[1]:
                    self.bkg_profile = None
                else:
                    for nd in np.arange(len(self.data)):
                        temp2 = temp*self.timedata[1][nd]
                        for n in np.arange(self.det_width):
                            self.data[nd, :, n] -= temp2
            else:
                self.bkg_profile = None
            # print("STATS: ", self.timedata[0], self.timedata[1].sum(),  self.data.sum(),  self.data.std())
    def adjust_offsets_automatically(self,  tolerance = 10.0):
        profiles,  offsets = [],  [0.0]
        shifted = []
        for ds in self.data:
            profiles.append(make_profile(ds, xaxis = np.linspace(1, self.det_length, int(self.det_length/self.redfac)), 
                                 tpoolin = self.threadpool, pbarin = self.progbar,  maxthreads = self.maxthreads))
        temp = np.zeros(profiles[0].shape)
        temp[:, 0] = profiles[0][:, 0]
        temp2 = temp.copy()
        shifted.append(profiles[0].copy())
        for prof in profiles[1:]:
            p = prof.copy()
            pfit, pcov, infodict, errmsg, success = leastsq(profile_offsets, [0.0, 1.0, 0.0],
                                                                                   args = (p, profiles[0], self.threadpool, self.progbar, self.maxthreads), 
                                                                                   full_output = 1)
            offsets.append(pfit[0])
            p[:, 0] -= pfit[0]
            shifted.append(p)
        self.component_profiles = shifted
        self.component_rawprofiles = profiles
        for p in profiles:
            temp[:, 1] += p[:, 1]
        for s in shifted:
            # the_object = MergeCurves(s, temp2,  tpool = self.threadpool, pbar = self.progbar)
            the_object = MergeCurves(s, temp2,  tpool = None, pbar = self.progbar,  mthreads = self.maxthreads)
            the_object.runit()
            temp2 = the_object.postprocess()
            # temp2 = merge2curves(s, temp2)
        self.summed_rawprofile = temp
        self.summed_adjusted_rawprofile = temp2
        return offsets,  shifted
    @pyqtSlot(object)
    def apply_offsets(self,  offsets):
        newdata = [self.data[0].copy()]
        for n, of in enumerate(offsets):
            if n>0:
                newdata.append(apply_offset_to_2D_data(self.data[n], self.cuts[0], self.cuts[1], of))
        self.data = np.array(newdata)
        self.processing_history.append(['OffsetsOnLoad']+offsets)
        self.finished_offsets.emit()
    @pyqtSlot(object)
    def justload_manyfiles(self, flist):
        self.load_files(flist)
        self.original_data2D = self.data.sum(0)
        self.data2D = self.original_data2D
        self.finished_2D.emit()
        self.save_last_params()
    @pyqtSlot(object)
    def preprocess_manyfiles(self, flist):
        self.load_files(flist)
        off,  shift = self.adjust_offsets_automatically()
        self.offsets = off
        self.finished_preprocess.emit()
        self.save_last_params()
    @pyqtSlot(object)
    def process_manyfiles(self, flist):
        self.load_files(flist)
        off,  shift = self.adjust_offsets_automatically()
        self.apply_offsets(off)
        self.original_data2D = self.data.sum(0)
        self.data2D = self.original_data2D
        self.finished_process.emit()
        self.finished_2D.emit()
        self.save_last_params()
    @pyqtSlot()
    def generate_mock_dataset(self):
        self.current_flist = ['Mock_File_001']
        # self.temp_path = os.path.split(flist[0])[0]
        self.name_as_segments = ['Mock_File_001']
        self.temp_name = "_".join(['Merged']+self.name_as_segments)
        # self.data = np.random.random((2048, 2048))*self.bpp
        self.data = rand_mt.normal(loc = 0.2*self.bpp,  scale = 0.1*self.bpp,  size = (2048, 2048))
        self.header = '# Fake dataset for ADLER'
        self.logvals = {'Time' : np.arange(100), 
                               'XPOS' : np.arange(100) + np.random.random(100), 
                               'Current1' : 1e-10*(10.0 +np.random.random(100))}
        self.logvalnames = self.logvals.keys()
        self.shortnames = 'MockAdler001'
        self.timedata = (1,  np.array([999.0]))
        self.original_data2D = self.data
        self.data2D = self.original_data2D
        self.plotax = [(1,2048), self.cuts]
        self.processing_history.append(['FakeDataGenerated'])
        self.finished_process.emit()
        self.finished_2D.emit()
    @pyqtSlot()
    def finalise_manyfiles(self):
        self.apply_offsets(self.offsets)
        self.original_data2D = self.data.sum(0)
        self.data2D = self.original_data2D
        self.finished_2D.emit()
    def curve_profile(self):  
        lcut,  hcut = self.cuts
        if self.data2D is not None:
            curve = curvature_profile(self.data2D, blocksize = self.segsize, percentile = self.bkg_perc,
                                                olimits = self.eline)
            pfit, pcov, infodict, errmsg, success = leastsq(fit_polynomial, [curve[:,1].mean(), 0.0, 0.0], args = (curve,), full_output = 1)
            curvefit = polynomial(pfit, curve[:,0])
            curvefit = np.column_stack([curve[:,0], curvefit])
            self.curvature = curve
            self.curvature_fit = curvefit
            self.curvature_params = pfit
            self.curvatureresult.emit([curve,  curvefit,  pfit])
        else:
            self.did_nothing.emit()
    def calculate_fft(self):
        if self.data2D is None:
            self.logger("No data has been loaded for curvature correction")
            self.did_nothing.emit()
            return None
        tempdata = rfft(self.data2D)
        fft_min, fft_max, fft_mean, fft_sum = tempdata.min(0), tempdata.max(0), tempdata.mean(0), np.abs(tempdata).sum(0)
        x_freqs = fftfreq(self.data2D.shape[1],1.0)
        plots = []
        for nx in [fft_min, fft_max, fft_mean, fft_sum]:
            tempval = nx / nx.sum()
            plots.append(np.column_stack([x_freqs,tempval]))
        self.fft_plots = plots
        self.finished_calcfft.emit()
        return "Done"
    @pyqtSlot()
    def correct_fft(self):
        self.fft_applied = False
        fftmin,  fftmax = self.ffts
        if fftmin is None or fftmax is None:
            self.did_nothing.emit()
            self.finished_fft.emit()
            return None
        if fftmin < -1e5 or fftmax < -1e5:
            self.logger("The FFT filter limits need to be set first!")
            self.finished_fft.emit() 
            self.did_nothing.emit()
            return None
        if self.data2D is None:
            self.logger("No data is available to be processed by the FFT filter")
            self.finished_fft.emit()
            self.did_nothing.emit()
            return None
        else:
            data = self.data2D
        tempdata = rfft(data) 
        x_freqs = fftfreq(data.shape[1],1.0)
        m1 = np.argmax(x_freqs > fftmin)
        m2 = np.argmax(x_freqs > fftmax)
        filler = interp1d(np.concatenate([x_freqs[m1-3:m1], x_freqs[m2:m2+3]]),
                                np.column_stack([tempdata[:, m1-3:m1], tempdata[:,m2:m2+3]]),
                               kind = "slinear")
        # filler = scint.interp1d([x_freqs[m1-3:m1].mean(), x_freqs[m2:m2+3].mean()], 
        #             np.column_stack([tempdata[:, m1-3:m1].mean(1), tempdata[:,m2:m2+3].mean(1)]))
        tempdata[:,m1:m2] = filler(x_freqs[m1:m2])
        print("Now the general smoothing")
        for i in range(tempdata.shape[0]):
            tempdata[i,-3:] = tempdata[i,-3:].mean()
        result = irfft(tempdata)
        self.corrected_data2D = result
        self.data2D = self.corrected_data2D
        self.processing_history.append(['FFTFilterApplied',  fftmin,  fftmax])
        self.fft_applied = True
        self.finished_fft.emit()
        return "Done"
    @pyqtSlot()
    def apply_poly(self):
        self.curvature_corrected = False
        data = self.data2D
        if data is None:
            self.logger("No data available to which curvature correction could be applied.")
            self.finished_poly.emit()
            self.did_nothing.emit()
            return None
        lcut,  hcut = self.cuts
        pmin,  pmax = self.eline
        poly = self.poly
        perc = self.bkg_perc
        if poly is not None:
            if poly[0] < -99.0 and poly[1] < -99.0 and poly[2] < -99.0:
                self.logger("Please specify some realistic parameters for the curvature correction.")
                self.finished_poly.emit()
                self.did_nothing.emit()
                return None
            elif poly[0] > 5.0 and poly[1] >5.0 and poly[2] > 5.0:
                curve = curvature_profile(data[pmin:pmax], blocksize = 1, percentile = perc, override = None, olimits = (0, pmax-pmin))
                curve[:,0] -= 0.5
                newcurve = interp1d(curve[:,0], curve[:,1], fill_value = "extrapolate")(np.arange(lcut,hcut))
                data = SphericalCorrection(data, poly, locut = lcut, hicut = hcut, direct_offsets = newcurve)
            else:
                data = SphericalCorrection(data, poly, locut = lcut, hicut = hcut)
            self.data2D = data
            self.processing_history.append(['CurvatureCorrectionApplied',  poly[0],  poly[1],  poly[2]])
            self.curvature_corrected = True
            self.finished_poly.emit()
            return "Done"
    def make_stripe(self):
        self.segment_plot = None
        if self.data2D is None:
            self.logger("No data has been loaded for curvature correction")
            self.did_nothing.emit()
            return None
        if self.eline[0] < 0 or self.eline[1] < self.eline[0]:
            self.logger("The elastic line limits are ill-defined. Not processing further.")
            self.did_nothing.emit()
            return None
        self.segment_plot = make_stripe(self.data2D, self.cuts, self.eline)
        self.segmentresult.emit([self.segment_plot,  [(self.eline[0],  self.eline[1]), (self.cuts[0]+1, self.cuts[1])]])
        return self.segment_plot,  [(self.eline[0],  self.eline[1]), (self.cuts[0]+1, self.cuts[1])]
    def make_histogram(self):
        self.histogram_plot = None
        if self.data2D is None:
            self.logger("No data has been loaded for histogram generation.")
            self.did_nothing.emit()
            return None
        if self.eline[0] < 0 or self.eline[1] < self.eline[0]:
            self.logger("The elastic line limits are ill-defined. Not processing further.")
            self.did_nothing.emit()
            return None
        self.histogram_plot = make_histogram(self.data2D, self.cuts, self.eline)
        self.historesult.emit([self.histogram_plot,  [(self.eline[0],  self.eline[1]), (self.cuts[0]+1, self.cuts[1])]])
        vals,  bins = self.histogram_plot
        xvals = (bins[1:] + bins[:-1])*0.5
        temphist = np.column_stack([xvals,  vals])
        WriteProfile(os.path.join(self.temp_path, self.temp_name+"_HISTOGRAM.txt"), temphist)
        return self.histogram_plot,  [(self.eline[0],  self.eline[1]), (self.cuts[0]+1, self.cuts[1])]
    @pyqtSlot()
    def autoprocess_file(self,  isfinal = True):
        self.process_file(guess = True,  final = isfinal)
    @pyqtSlot()
    def process_file(self,  guess = False,  final = True):  
        if self.data2D is None:
            self.did_nothing.emit()
            return None,  None, None, None, None
        self.raw_profile = make_profile(self.data2D,  self.redfac,  xaxis = np.linspace(1, self.det_length, int(self.det_length/self.redfac)), 
                                                  tpoolin = self.threadpool, pbarin = self.progbar, maxthreads = self.maxthreads)
        self.plotax1D = [(1,self.det_length), (self.cuts[0]+1, self.cuts[1])]
        WriteProfile(os.path.join(self.temp_path, self.temp_name+"_1D.txt"), self.raw_profile,
                        header = self.header.split('\n'), params = [self.parlist,  self.pardict], 
                        varlog = [self.logvalnames,  self.logvals])
        x1, eV1 = self.eVpair1
        x2, eV2 = self.eVpair2
        if self.eline[0] < 0 and self.eline[1] < 0:
            return self.raw_profile.copy(), None, None, None, None
        else:
            bkg = self.raw_profile[np.where(self.raw_profile[:, 1] < np.percentile(self.raw_profile[:, 1], self.bkg_perc))]
            if guess:
                if (eV1 < 0) or (eV2 < 0):
                    self.logger("Guessing the elastic line position based on the intensity.")
                    fit, peak, chi2 = elastic_line(self.raw_profile, bkg)
                elif len(self.energies) > 0:
                    self.logger("Guessing the elastic line position based on the energy calibration.")
                    enval = self.energies.mean()
                    slope = (x2-x1)/(eV2 - eV1)
                    xpos = x1 + (enval - eV1)*slope
                    pwidth = 10 * self.redfac
                    uplim_lim1 = self.det_length -1 - pwidth
                    uplim_lim2 = self.det_length -1
                    self.logger("Based on the photon energy " + str(enval) + " the elastic line is at " + str(round(xpos, 3)) + " channels." )
                    if xpos > 0 and xpos < self.det_length:
                        lim1 = int(xpos - 15)
                        lim2 = int(xpos + 7)
                        if lim1 < 0:
                            lim1 = 0
                        if lim1 > uplim_lim1:
                            lim1 = uplim_lim1
                        if lim2 > uplim_lim2:
                            lim2 = uplim_lim2
                        if lim2 < pwidth:
                            lim2 = pwidth
                        if lim2 - lim1 < pwidth:
                            lim2 = lim1 +pwidth
                        if lim1 > lim2:
                            lim1 = lim2 - pwidth
                        fit, peak, chi2 = elastic_line(self.raw_profile, bkg, olimits = (lim1,  lim2))
                    else:
                        self.logger("Reverting to the guess of the elastic line position based on the intensity.")
                        fit, peak, chi2 = elastic_line(self.raw_profile, bkg)
                else:
                    fit, peak, chi2 = elastic_line(self.raw_profile, bkg)
            else:
                fit, peak, chi2 = elastic_line(self.raw_profile, bkg, olimits = (self.eline[0],  self.eline[1]))
            if fit is not None:
                # chi2 = chisquared(peak, self.raw_profile)
                self.logger('heigth, FWHM, centre, baseline')
                self.logger(str(fit[0]))
                self.logger(str(fit[1]))
                peak_area = fit[0][0]*abs(fit[0][1])/gauss_denum*(2*np.pi)**0.5
                peak_area_error = peak_area*((fit[1][0]/fit[0][0])**2 + (fit[1][1]/fit[0][1])**2)**0.5
                fitstring = ' '.join(['FWHM =', str(abs(round(fit[0][1],3))), ' +/- ', str(abs(round(fit[1][1],3))),'\n',
                            'centre =', str(round(fit[0][2],3)), ' +/- ', str(abs(round(fit[1][2],3))),'\n',
                            'area =', str(round(peak_area,3)), ' +/- ', str(abs(round(peak_area_error,3)))])# , '\n', 
                            # '$\chi^{2}$=', str(round(chi2, 3))])
            else:
                fitstring = "The fitting failed. Not enough data points?"
            self.fitted_peak_channels = peak
            self.fitting_params_channels = fit
            self.fitting_textstring_channels = fitstring
            tempthing = self.raw_profile.copy()
            if final:
                self.fittingresult.emit([tempthing,  bkg, peak,  fit,  fitstring])
                if fit is not None:
                    self.params_fitting.emit({ 'FIT_centre':np.array([round(fit[0][2],3), abs(round(fit[1][2],3))]), 
                                                       'FIT_fwhm':np.array([abs(round(fit[0][1],3)), abs(round(fit[1][2],3))]), 
                                                       'FIT_area':np.array([round(peak_area,3), abs(round(peak_area_error,3))])})
            return tempthing,  bkg, peak,  fit,  fitstring
    @pyqtSlot()
    def auto_eV_profile(self):
        self.eV_profile(guess = True)
    @pyqtSlot()
    def eV_profile(self,  guess = False): 
        if self.data2D is None:
            self.logger("There is no file to be processed.")
            self.did_nothing.emit()
            return None, None,  None, None,  None
        x1, eV1 = self.eVpair1
        x2, eV2 = self.eVpair2
        lcut, hcut = self.cuts
        emin,  emax = self.eline
        if (eV1 < 0) or (eV2 < 0):
            self.logger("Invalid energy range, skipping the calculation")
            self.did_nothing.emit()
            return None, None,  None, None,  None
        else:
            profi, back, peak, fit,  fitstring = self.process_file(guess,  final = False)
            xaxis = profi[:,0].copy()
            xstep = (xaxis[1:] - xaxis[:-1]).mean()
            slope = (eV2-eV1)/(x2-x1)
            xaxis = np.concatenate([xaxis[:1]-xstep, xaxis, xaxis[-1:]+xstep])
            new_xaxis = (xaxis - x1)*slope + eV1
            recalc = interp1d(xaxis, new_xaxis)
            profi[:,0] = recalc(profi[:,0])
            self.energy_profile = profi.copy()
            if emin < 0 and emax < 0:
                WriteProfile(os.path.join(self.temp_path, self.temp_name+"_1D_absEnergy.txt"), self.energy_profile,
                        header = self.header.split('\n'), params = [self.parlist,  self.pardict], 
                        varlog = [self.logvalnames,  self.logvals])
                return profi, back, peak, fit
            back[:,0] = recalc(back[:,0])
            peak[:,0] = recalc(peak[:,0])
            new_centre = recalc(fit[0][2])
            self.logger('heigth, FWHM, centre, baseline')
            self.logger(str(fit[0]))
            self.logger(str(fit[1]))
            peak_area = fit[0][0]*abs(slope*fit[0][1])/gauss_denum*(2*np.pi)**0.5
            peak_area_error = peak_area*((fit[1][0]/fit[0][0])**2 + (fit[1][1]/fit[0][1])**2)**0.5
            profi[:,0] -= new_centre
            back[:,0] -= new_centre
            peak[:,0] -= new_centre
            profi[:,0] *= -1
            back[:,0] *= -1
            peak[:,0] *= -1
            fitstring = ' '.join(['FWHM (meV) =', str(abs(round(1000.0*slope*fit[0][1],3))), ' +/- ', str(abs(round(1000.0*slope*fit[1][1],3))),'\n',
                        'centre (meV) =', str(round(0.0,3)), ' +/- ', str(abs(round(1000.0*fit[1][2],3))),'\n',
                        'area =', str(round(peak_area,3)), ' +/- ', str(abs(round(peak_area_error,3))) ])
                        # , '\n', 
                        # '$\chi^{2}$=', str(round(chi2, 3))])
            self.energy_profile = profi                
            WriteProfile(os.path.join(self.temp_path, self.temp_name+"_1D_deltaE.txt"), self.energy_profile,
                        header = self.header.split('\n'), params = [self.parlist,  self.pardict], 
                        varlog = [self.logvalnames,  self.logvals])
            self.fitted_peak_energy = peak
            self.fitting_params_energy = fit
            self.fitting_textstring_energy = fitstring
            self.energyresult.emit([profi, back, peak, fit,  fitstring])
            self.params_energy.emit({  'ENERGY_centre':np.array([0.0, abs(round(1000.0*fit[1][2],3))]), 
                                                    'ENERGY_fwhm': np.array([abs(round(1000.0*slope*fit[0][1],3)), abs(round(1000.0*slope*fit[1][1],3))]), 
                                                    'ENERGY_area':np.array([round(peak_area,3), abs(round(peak_area_error,3))])})
            return profi, back, peak, fit,  fitstring

def precise_merge(curves, outname = 'best_merge.txt'):
    datalist, energylist, unitlist = [],[],[]
    for fname in curves:
        a,b,c = read_1D_curve(fname)
        datalist.append(a)
        energylist.append(b)
        unitlist.append(c)
    # energies = np.array(energylist).mean()
    # I ignore the units and energies. Let's assume I put the correct files in the list.
    minx, maxx, stepx = 1e5,-1e5,-1.0
    for dat in datalist:
        minx = min(minx, dat[:,0].min())
        maxx = max(maxx, dat[:,0].max())
        step = (dat[1:,0] - dat[:-1,0]).mean()
        stepx = max(stepx, step)
    newx = np.arange(minx, maxx + 0.1*stepx, stepx)
    target = np.zeros((len(newx),2))
    target[:,0] = newx
    for dat in datalist:
        target = merge2curves(dat,target)
    WriteEnergyProfile(outname, target, [])
    return target

def single_n2_profile(params, xarr, fixed_lorentz = 0.12, yarr = None, extraweight = False):
    num_peaks = int((len(params) - 3)/2)
    profile = np.zeros(xarr.shape)
    dists = []
    scalefac = 1.0
    for n in range(num_peaks):
        profile += my_voigt(xarr, abs(params[2*n]), abs(params[2*n+1]), abs(params[-3]), fixed_lorentz)
        dists.append(params[2*n+1])
    dists = np.sort(dists)
    dists = dists[1:] - dists[:-1]
    profile += params[-2]*xarr + params[-1]
    if yarr is not None:
        profile -= yarr
        if extraweight:
            if len(dists) > 0:
                scalefac = 1.0 + max(0.0, 0.15 - dists.min())
    return profile * scalefac
    
def single_neon_profile(params, xarr, fixed_lorentz = 0.254, yarr = None, extraweight = False):
    num_peaks = int((len(params) - 5)/2)
    profile = np.zeros(xarr.shape)
    scalefac = 1.0
    for n in range(num_peaks):
        profile += my_voigt(xarr, abs(params[2*n]), abs(params[2*n+1]), abs(params[-5]), fixed_lorentz)
    profile += abs(params[-4])*np.arctan((xarr-params[-3])/params[-2]) + params[-1]
    if yarr is not None:
        profile -= yarr
        if extraweight:
            scalefac = 1.0 + max(0.0, 870 - params[-3])
    return profile * scalefac
    
def single_edge_profile(params, xarr, fixed_lorentz = 0.254, yarr = None):
    profile = np.zeros(xarr.shape)
    profile += params[-4]*np.arctan((xarr-params[-3])/params[-2]) + params[-1]
    if yarr is not None:
        profile -= yarr
    return profile

def fit_edge_profile(num_arr,  nlines = 0):
    mid_y = (num_arr[:, 1].max()+num_arr[:, 1].min())/2.0
    crit = np.abs(num_arr[:, 1] - mid_y)
    ref = crit.min()
    mid_x = num_arr[:, 0][np.where(crit==ref)].mean()
    step = np.abs(num_arr[1:, 0] - num_arr[:-1, 0]).mean()
    params = [num_arr[-1, 1], mid_x, 10*step, num_arr[0, 1]]
    pfit, pcov, infodict, errmsg, success = leastsq(single_edge_profile, params, args = (num_arr[:, 0], 0.254, num_arr[:, 1]), full_output =1)
    final_profile = single_edge_profile(pfit, num_arr[:, 0], 0.254)
    GOF = ((final_profile - num_arr[:, 1])**2).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
    print('GOF: ',  GOF, ((final_profile - num_arr[:, 1])**2).sum()**0.5, np.abs(num_arr[:, 1]).sum())
    return final_profile, pfit[1],  pfit[2],  GOF

def fit_n2_gas(inp_num_arr, fixed_gamma = 0.06):
    peak_list = np.array([400.76, 400.996, 401.225, 401.45, 401.66, 401.88, 402.1, 402.29, 402.48])
    params = []
    # fixed_gamma = 0.012
    seq = np.argsort(inp_num_arr[:, 0])
    num_arr = inp_num_arr[seq]
    temp_prof = interp1d(num_arr[:, 0],  num_arr[:, 1])
    for peak in peak_list:
        if peak >= num_arr[:, 0].min() and peak <= num_arr[:, 0].max():
            params += [temp_prof(peak), peak]
    params += [0.01, 0.0, 0.0]
    pfit, pcov, infodict, errmsg, success = leastsq(single_n2_profile, params, args = (num_arr[:, 0], fixed_gamma, num_arr[:, 1],  True), full_output =1)
    final_profile = single_n2_profile(pfit, num_arr[:, 0], fixed_gamma)
    # determine the fit quality
    GOF = ((final_profile - num_arr[:, 1])**2 ).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
    print('GOF: ',  GOF, ((final_profile - num_arr[:, 1])**2 ).sum()**0.5, np.abs(num_arr[:, 1]).sum() )
    resline = np.concatenate([[GOF], pfit])
    results = [resline]
    if 1.0/GOF < 50:
        for xx in range(3):
            newparams = np.array(params)
            plen = len(newparams)
            rands = np.random.rand(plen)
            newparams[0:-3:2] *= rands[0:-3:2]
            newparams[1:-3:2] += 0.1*(rands[1:-3:2] - 0.5)
            pfit, pcov, infodict, errmsg, success = leastsq(single_n2_profile, params, args = (num_arr[:, 0], fixed_gamma, num_arr[:, 1],  True), full_output =1)
            final_profile = single_n2_profile(pfit, num_arr[:, 0], fixed_gamma)
            # determine the fit quality
            GOF = ((final_profile - num_arr[:, 1])**2 ).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
            resline = np.concatenate([[GOF], pfit])
            results.append(resline)
        results = np.array(results)
        results = results[np.argsort(results[:, 0])]
        pfit = results[0, 1:]
        final_profile = single_n2_profile(pfit, num_arr[:, 0], fixed_gamma)
        # determine the fit quality
        GOF = ((final_profile - num_arr[:, 1])**2 ).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
        print('best GOF: ',  GOF, ((final_profile - num_arr[:, 1])**2 ).sum()**0.5, np.abs(num_arr[:, 1]).sum() )
    # final_profile = single_n2_profile(params, num_arr[:, 0], fixed_gamma)
    # gauss_sigma = params[-3]
    # peak_pos = params[1:-3:2]
    gauss_sigma = pfit[-3]
    lorentz_gamma = fixed_gamma
    peak_pos = pfit[1:-3:2]
    return final_profile, gauss_sigma, lorentz_gamma, peak_pos,  GOF

def fit_neon_gas(inp_num_arr,  fixed_gamma = 0.126):
    peak_list = np.array([867.2, 868.8, 869.36, 869.6,  869.95, 870.23, 870.42, 870.60])
    atan2_params = [870.15, 0.125]
    # y = y0 + A*atan((x-xc)/w)
    params = []
    seq = np.argsort(inp_num_arr[:, 0])
    num_arr = inp_num_arr[seq]
    temp_prof = interp1d(num_arr[:, 0],  num_arr[:, 1])
    for peak in peak_list[:1]:
        if peak >= num_arr[:, 0].min() and peak <= num_arr[:, 0].max():
            params += [temp_prof(peak), peak]
        else:
            params += [num_arr[:, 1].max(), num_arr[:, 0].mean()]
    for peak in peak_list[1:]:
        if peak >= num_arr[:, 0].min() and peak <= num_arr[:, 0].max():
            params += [temp_prof(peak), peak]
    if len(params) == 2:
        params[0] = num_arr[:, 1].max()
        params[1] = num_arr[:, 0][np.where(num_arr[:, 1] == params[0])][0]
        params += [0.01, 0.0,  atan2_params[0], atan2_params[1], 0.0]
    else:
        params += [0.01, num_arr[-1, 1],  atan2_params[0], atan2_params[1], num_arr[0, 1]]
    pfit, pcov, infodict, errmsg, success = leastsq(single_neon_profile, params, args = (num_arr[:, 0], fixed_gamma, num_arr[:, 1],  True), full_output =1)
    final_profile = single_neon_profile(pfit, num_arr[:, 0], fixed_gamma)
    # determine the fit quality
    GOF = ((final_profile - num_arr[:, 1])**2).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
    print('GOF: ',  GOF, ((final_profile - num_arr[:, 1])**2 ).sum()**0.5, np.abs(num_arr[:, 1]).sum() )
    #
    resline = np.concatenate([[GOF], pfit])
    results = [resline]
    if 1.0/GOF < 50:
        for xx in range(3):
            newparams = np.array(params)
            plen = len(newparams)
            rands = np.random.rand(plen)
            newparams[0:-5:2] *= rands[0:-5:2]
            newparams[1:-5:2] += 0.1*(rands[1:-5:2] - 0.5)
            pfit, pcov, infodict, errmsg, success = leastsq(single_n2_profile, params, args = (num_arr[:, 0], fixed_gamma, num_arr[:, 1],  True), full_output =1)
            final_profile = single_n2_profile(pfit, num_arr[:, 0], fixed_gamma)
            # determine the fit quality
            GOF = ((final_profile - num_arr[:, 1])**2 ).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
            resline = np.concatenate([[GOF], pfit])
            results.append(resline)
        results = np.array(results)
        results = results[np.argsort(results[:, 0])]
        pfit = results[0, 1:]
        final_profile = single_n2_profile(pfit, num_arr[:, 0], fixed_gamma)
        # determine the fit quality
        GOF = ((final_profile - num_arr[:, 1])**2 ).sum()**0.5/ np.abs(num_arr[:, 1]).sum()
        print('best GOF: ',  GOF, ((final_profile - num_arr[:, 1])**2 ).sum()**0.5, np.abs(num_arr[:, 1]).sum() )
    gauss_sigma = pfit[-5]
    lorentz_gamma = fixed_gamma
    peak_pos = pfit[1:-5:2]
    return final_profile, gauss_sigma, lorentz_gamma, peak_pos,  GOF
    
