
#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

# from PyQt5.Qt import Qt as _Qt
# from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot,  QMutex,  Qt
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout
from PyQt5.QtWidgets import QSizePolicy, QFileDialog, QTableWidget,  \
                                                QPushButton, QTableWidgetItem
import numpy as np
from scipy.optimize import leastsq
from scipy.interpolate import interp1d
# import copy
from PEAXIS import ComputeKernel
from VariablesGUI import VarBox

input_variables = [
{'Name': 'arm_theta',  'Unit':'degree',  'Value':90,  'MinValue':33.0,  'MaxValue':139.0,  'Length': 1,  'Type':'float', 'Key':'arm_theta'}, 
{'Name': 'poly_order',  'Unit':'N/A',  'Value':0,  'MinValue':0,  'MaxValue':9,  'Length': 1,  'Type':'int', 'Key':'poly_order'}, 
]

output_variables = [
{'Name': 'zdir',  'Unit':'degree',  'Value':0.0,  'MinValue':-180.0,  'MaxValue':180,  'Length': 1,  'Type':'float', 'Key':'zdir'}, 
{'Name': 'incl',  'Unit':'degree',  'Value':0.0,  'MinValue':-180.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'incl'}, 
{'Name': 'rot',  'Unit':'degree',  'Value':0.0,  'MinValue':-180.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'rot'}, 
]

def poly(args, xvals, yvals = None):
    newyvals = np.zeros(xvals.shape)
    for n, ar in enumerate(args):
        newyvals += ar * xvals**n
    if yvals is None:
        return newyvals
    else:
        return newyvals - yvals

def read_text(source):
    data = []
    for line in source:
        valid = line.split('#')[0]
        toks = valid.split()
        if len(toks) > 3:
            numbs = []
            for n in range(4):
                try:
                    dd = float(toks[n])
                except:
                    dd = -1.0
                numbs.append(dd)
            data.append(numbs)
    return data

def write_text(data, target):
    for entry in data:
        line = " ".join([str(round(x, 5)) for x in entry]) + '\n'
        target.write(line)

tabnames = ['ArmTheta','ZDIR','INCL','ROT', 'UseIt']
class OffsetList(QObject):
    gotvals = pyqtSignal()
    needanupdate = pyqtSignal()
    def __init__(self, master,  nrows =1,  ncolumns = len(tabnames)):
        super().__init__(master)
        self.master = master
        self.data = []
        self.useit = []
        self.busy = True
        self.table = QTableWidget(nrows, ncolumns, self.master)
        self.table.setSizePolicy(QSizePolicy.MinimumExpanding,QSizePolicy.MinimumExpanding)
        for n in range(self.table.columnCount()):
            self.table.setItem(0, n, QTableWidgetItem(tabnames[n]))
        self.table.cellChanged.connect(self.update_values)
        self.table.cellChanged.connect(self.update_ticks)
        self.needanupdate.connect(self.redraw_table)
    def add_row(self, data):
        self.busy = True
        self.table.blockSignals(True)
        rowitems = []
        lastnit = len(data)
        temp = []
        for nit, dud in enumerate(data):
            d = str(dud)
            rowitems.append(QTableWidgetItem(str(d).strip("()[]'")))
            if nit < 4:
                try:
                    temp.append(float(d.strip("()[]'")))
                except:
                    temp.append(-1.0)
        if len(temp) == 4:
            self.data.append(temp)
        chkBoxItem = QTableWidgetItem()
        chkBoxItem.setFlags(Qt.ItemIsUserCheckable | Qt.ItemIsEnabled)
        chkBoxItem.setCheckState(Qt.Checked)
        rowitems.append(chkBoxItem)
        self.useit.append(True)
        lastrow = self.table.rowCount()
        self.table.insertRow(lastrow)
        for n, ri in enumerate(rowitems):
            self.table.setItem(lastrow, n, ri)
        for nit in range(lastnit, self.table.columnCount()):
            self.table.setItem(lastrow, nit, QTableWidgetItem(str("")))
        self.busy = False
        self.table.blockSignals(False)
        self.needanupdate.emit()
    @pyqtSlot()
    def clear_table(self):
        self.data = []
        self.useit = []
        for nr in range(1, self.table.rowCount())[::-1]:
            self.table.removeRow(nr)
        self.gotvals.emit()
    def return_values(self):
        final = []
        for nr in range(len(self.useit)):
            if self.useit[nr]:
                if self.data[nr][0] > 0:
                    final.append(self.data[nr])
        return final
    @pyqtSlot(int, int)
    def update_values(self,  row =0,  column=0):
        if self.busy:
            return None
        self.busy = True
        for nr in range(1,  self.table.rowCount()):
            for nc in range(0, 4):
                if nc <4:
                    try:
                        self.data[nr-1][nc] = float(self.table.item(nr, nc).text().strip("()[]'"))
                    except:
                        continue
        self.busy = False
        self.needanupdate.emit()
        self.gotvals.emit()
    @pyqtSlot(int, int)
    def update_ticks(self,  row =0,  column=0):
        if self.busy:
            return None
        self.busy = True
        for nr in [row]:
            for nc in [column]:
                if nc ==4:
                    # tempdebug = self.table.item(nr, 4).checkState()
                    # self.useit[nr-1] = (self.table.item(nr, 4).checkState() == Qt.Qt.Checked)
                    self.useit[nr-1] = not self.useit[nr-1]
        self.busy = False
        self.needanupdate.emit()
        self.gotvals.emit()
    @pyqtSlot()
    def redraw_table(self):
        if self.busy:
            return None
        self.busy = True
        self.table.blockSignals(True)
        for nr in range(1,  self.table.rowCount()):
            for nc in range(0, 5):
                if nc <4:
                    self.table.item(nr, nc).setText(str(self.data[nr-1][nc]))
                elif nc ==4:
                    if self.useit[nr-1]:
                        self.table.item(nr, 4).setCheckState(Qt.Checked)
                    else:
                        self.table.item(nr, 4).setCheckState(Qt.Unchecked)
        self.busy = False
        self.table.blockSignals(False)
    def assign_fitparams(self, fitparams):
        nums, width,  widtherr,  area,  areaerr = fitparams
        for en,  realnum in enumerate(nums):
            nr = realnum + 1
            for nc in range(5, 9):
                if nc == 5:
                    self.table.item(nr, nc).setText(str(width[en]))
                if nc == 6:
                    self.table.item(nr, nc).setText(str(widtherr[en]))
                if nc == 7:
                    self.table.item(nr, nc).setText(str(area[en]))
                if nc == 8:
                    self.table.item(nr, nc).setText(str(areaerr[en]))


class OffsetsTab(QObject):
    values_changed = pyqtSignal()
    def __init__(self, parent,  forplotting):
        super().__init__(parent)
        self.CK = ComputeKernel(parent)
        self.figures = []            
        self.offset_zdir = None
        self.offset_incl = None
        self.offset_rot = None
        for n in forplotting:
            canvas, figure, layout = n
            self.figures.append(figure)
        # set up the layout
        self.usual_angles = np.arange(33, 139.01, 1.0)
        self.base = QWidget(parent)
        self.ohlay = QHBoxLayout(self.base)
        self.panel = QWidget(self.base)
        self.plots = QWidget(self.base)
        self.vlay = QVBoxLayout(self.panel)
        self.pvlay = QVBoxLayout(self.plots)
        self.tophalf = QWidget(self.base)
        self.hlay1 = QHBoxLayout(self.tophalf)
        self.bottomhalf = QWidget(self.base)
        self.hlay2 = QHBoxLayout(self.bottomhalf)
        self.buttonzone = QWidget(self.base)
        self.hlay3 = QHBoxLayout(self.buttonzone)
        self.buttonzone2 = QWidget(self.base)
        self.hlay4 = QHBoxLayout(self.buttonzone2)
        self.ohlay.addWidget(self.panel)
        self.ohlay.addWidget(self.plots)
        self.vlay.addWidget(self.buttonzone)
        self.vlay.addWidget(self.tophalf)
        self.vlay.addWidget(self.bottomhalf)
        self.vlay.addWidget(self.buttonzone2)
        for n in forplotting:
            canvas, figure, layout = n
            self.pvlay.addWidget(canvas)
        # now add the components
        # self.box1 = VarBox(self.base, sample_variables,  "Sample (read-only)", prec_override = 5)
        # self.hlay1.addWidget(self.box1.base)
        # self.box2 = VarBox(self.base, instrument_variables,  "Manipulator", prec_override = 5)
        # self.hlay1.addWidget(self.box2.base)
        self.table = OffsetList(self.base)
        self.hlay1.addWidget(self.table.table)
        self.box3 = VarBox(self.base, input_variables,  "Input", prec_override = 5)
        self.hlay2.addWidget(self.box3.base)
        self.box4 = VarBox(self.base, output_variables,  "Output", prec_override = 5)
        self.hlay2.addWidget(self.box4.base)
        b1 = QPushButton("Add from file",  self.buttonzone)
        b2 = QPushButton("Save to file",  self.buttonzone)
        b3 = QPushButton("Clear table",  self.buttonzone)
        b1.clicked.connect(self.load_from_file)
        b2.clicked.connect(self.save_to_file)
        b3.clicked.connect(self.clear_table)
        bb1 = QPushButton("Plot positions",  self.buttonzone2)
        bb2 = QPushButton("Fit offsets",  self.buttonzone2)
        bb1.clicked.connect(self.calculate)
        bb2.clicked.connect(self.calculate_and_fit)
        for bb in [b1, b2, b3]:
            self.hlay3.addWidget(bb)
        for bb in [bb1, bb2]:
            self.hlay4.addWidget(bb)
        # self.hlay2.addWidget(canvas)
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        self.input_vals = self.box3.returnValues()
        # and link the signals to the slots to automatise the operation
        # self.box1.values_changed.connect(self.on_parchange)
        # self.box2.values_changed.connect(self.on_parchange)
        self.table.gotvals.connect(self.on_parchange)
        for dud in range(12):
            self.table.add_row([-1, 0, 0, 0])
        self.box3.values_changed.connect(self.on_parchange)
        # self.box1.disableChanges()
        self.box4.disableChanges()
        self.CK.hkl_result.connect(self.output)
        self.CK.hkl_limits.connect(self.added_output)
        self.on_parchange()
    @pyqtSlot()
    def load_from_file(self):
        result, ftype = QFileDialog.getOpenFileNames(self.base, 'Load files with specular reflection position', '.',
               'Text files (*.txt);;All files (*.*)')
        if len(result) > 0:
            data = []
            for fname in result:
                try:
                    source = open(fname, 'r')
                except:
                    continue
                else:
                    data += read_text(source)
                    source.close()
            for line in data:
                self.table.add_row(line)
            self.on_parchange()
    def save_to_file(self):
        result, ftype = QFileDialog.getSaveFileName(self.base, 'Save the offsets table to a file:', '.',
               'Text file (*.txt);;All files (*.*)')
        if result is not None:
            target = open(result, 'w')
            write_text(self.table.return_values(), target)
            target.close()
    def clear_table(self):
        self.table.clear_table()
        for dud in range(12):
            self.table.add_row([-1, 0, 0, 0])
    @pyqtSlot(object)
    def updateSample(self, input):
        vlist,  vdict = input[0],  input[1]
        self.box1.enableChanges()
        self.box1.takeValues(vlist, vdict)
        self.box1.disableChanges()
    def calculate(self):
        poly_ord = self.input_vals[1]['poly_order']
        zdir,  incl,  rot = self.CK.specular_pos(self.usual_angles)
        arrdata = np.array(self.fitting_data)
        havedata = len(arrdata) > 0
#        if havedata:
#            spec1,  spec2, spec3 = self.CK.specular_pos(arrdata[:, 0])
#            off1, off2, off3 = arrdata[:, 1] -spec1, arrdata[:, 2] - spec2, arrdata[:, 3] - spec3
#            pfit1, pcov, infodict, errmsg, success = leastsq(poly, np.zeros(poly_ord+1),
#                                                                               args=(arrdata[:, 0], off1), full_output=1)
#            pfit2, pcov, infodict, errmsg, success = leastsq(poly, np.zeros(poly_ord+1),
#                                                                               args=(arrdata[:, 0], off2), full_output=1)
#            pfit3, pcov, infodict, errmsg, success = leastsq(poly, np.zeros(poly_ord+1),
#                                                                               args=(arrdata[:, 0], off3), full_output=1)
#            self.offset_zdir = interp1d(self.usual_angles, poly(pfit1, self.usual_angles))
#            self.offset_incl = interp1d(self.usual_angles, poly(pfit2, self.usual_angles))
#            self.offset_rot = interp1d(self.usual_angles, poly(pfit3, self.usual_angles))
        # need to fork it, so we can decide if we are plotting the raw values, or offsets...
        for nfig in range(3):
            tempfig = self.figures[nfig]
            if nfig==0:
                ideal = np.column_stack([self.usual_angles, zdir])
                if havedata:
                    real = np.column_stack([arrdata[:, 0], arrdata[:, 1]])
                temptitle ='ZDIR'
            if nfig==1:
                ideal = np.column_stack([self.usual_angles, incl])
                if havedata:
                    real = np.column_stack([arrdata[:, 0], arrdata[:, 2]])
                temptitle='INCL'
            if nfig==2:
                ideal = np.column_stack([self.usual_angles, rot])
                if havedata:
                    real = np.column_stack([arrdata[:, 0], arrdata[:, 3]])
                temptitle='ROT'
            tempfig.clear()
            ax = tempfig.add_subplot(111)
            ax.grid(True)
            ax.set_title(temptitle)
            ax.plot(ideal[:, 0],  ideal[:, 1],  '-',  label = 'ideal')
            if havedata:
                ax.plot(real[:, 0], real[:, 1], 's', label = 'measured')
            ax.legend(loc=0)
            tempfig.canvas.draw()
    def calculate_and_fit(self):
        poly_ord = self.input_vals[1]['poly_order']
        zdir,  incl,  rot = self.CK.specular_pos(self.usual_angles)
        arrdata = np.array(self.fitting_data)
        havedata = len(arrdata) > 0
        if not havedata:
            return None
        else:
            spec1,  spec2, spec3 = self.CK.specular_pos(arrdata[:, 0])
            off1, off2, off3 = arrdata[:, 1] -spec1, arrdata[:, 2] - spec2, arrdata[:, 3] - spec3
            try:
                pfit1, pcov, infodict, errmsg, success = leastsq(poly, np.zeros(poly_ord+1),
                                                                               args=(arrdata[:, 0], off1), full_output=1)
            except:
                self.offset_zdir = None
            else:
                self.offset_zdir = interp1d(self.usual_angles, poly(pfit1, self.usual_angles))
            try:
                pfit2, pcov, infodict, errmsg, success = leastsq(poly, np.zeros(poly_ord+1),
                                                                               args=(arrdata[:, 0], off2), full_output=1)
            except:
                self.offset_incl = None
            else:
                self.offset_incl = interp1d(self.usual_angles, poly(pfit2, self.usual_angles))
            try:
                pfit3, pcov, infodict, errmsg, success = leastsq(poly, np.zeros(poly_ord+1),
                                                                               args=(arrdata[:, 0], off3), full_output=1)
            except:
                self.offset_rot = None
            else:
                self.offset_rot = interp1d(self.usual_angles, poly(pfit3, self.usual_angles))
        # need to fork it, so we can decide if we are plotting the raw values, or offsets...
        for nfig in range(3):
            tempfig = self.figures[nfig]
            if nfig==0:
                if self.offset_zdir is not None:
                    ideal = np.column_stack([self.usual_angles, self.offset_zdir(self.usual_angles)])
                else:
                    ideal = None
                if havedata:
                    real = np.column_stack([arrdata[:, 0], off1])
                temptitle ='ZDIR'
            if nfig==1:
                if self.offset_incl is not None:
                    ideal = np.column_stack([self.usual_angles, self.offset_incl(self.usual_angles)])
                else:
                    ideal = None
                if havedata:
                    real = np.column_stack([arrdata[:, 0], off2])
                temptitle='INCL'
            if nfig==2:
                if self.offset_rot is not None:
                    ideal = np.column_stack([self.usual_angles, self.offset_rot(self.usual_angles)])
                else:
                    ideal = None
                if havedata:
                    real = np.column_stack([arrdata[:, 0], off3])
                temptitle='ROT'
            tempfig.clear()
            ax = tempfig.add_subplot(111)
            ax.grid(True)
            ax.set_title(temptitle)
            if ideal is not None:
                ax.plot(ideal[:, 0],  ideal[:, 1],  '-',  label = 'fit')
            if havedata:
                ax.plot(real[:, 0], real[:, 1], 's', label = 'offset')
            ax.legend(loc=0)
            tempfig.canvas.draw()
        self.show_offsets_at_angle()
    def show_offsets_at_angle(self):
        angle = self.input_vals[1]['arm_theta']
        zdir, incl, rot = 0, 0, 0
        if self.offset_zdir is not None:
            zdir = self.offset_zdir(angle)
        if self.offset_incl is not None:
            incl = self.offset_incl(angle)
        if self.offset_rot is not None:
            rot = self.offset_rot(angle)
        vdict = {'zdir':[zdir],'incl':[incl],'rot':[rot]}
        self.box4.enableChanges()
        self.box4.receiveValues(vdict)
        self.box4.disableChanges()
    def on_parchange(self):
        self.fitting_data = self.table.return_values()
        self.input_vals = self.box3.returnValues()
        self.show_offsets_at_angle()
        # self.calculate()
    @pyqtSlot(object)
    def added_output(self, results):
        hkl_min,  hkl_max = results
        newvals = 0.5*np.abs(hkl_max-hkl_min)
        vnames = ['HKL_var']
        vvals = {
                    'HKL_var': newvals
        }
        self.box4.enableChanges()
        self.box4.takeValues(vnames,  vvals)
        self.box4.disableChanges()
    @pyqtSlot(object)
    def output(self, results):
        # self.sample_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        (Qmod,  Q_perp,  Q_par),  hkl,  VD = results
        vnames = ['Q', 'Q_perp',  'Q_par',  'HKL']
        vvals = {'Q' : [Qmod], 
                    'Q_perp' : [Q_perp], 
                    'Q_par':[Q_par], 
                    'HKL': hkl
        }
        print(Qmod,  hkl)
        self.box4.enableChanges()
        self.box4.takeValues(vnames,  vvals)
        self.box4.disableChanges()
        # now some added plotting
        # VD = self.CK.plot_vectors()
        plot_vectors_projected(VD,  projection = (1, 0),  fig = self.figures[0])
        plot_vectors_projected(VD,  projection = (1, 2),  fig = self.figures[1])
        plot_vectors_projected(VD,  projection = (0, 2),  fig = self.figures[2])

def plot_vectors_projected(vdict,  projection = (1, 0), outFile = "", fig = None):
    if fig == None:
        fig = mpl.figure(figsize = [7.0, 6.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    ax = fig.add_subplot(111)
    ax.grid(True)
    ax.set_xlim([1.0, -1.0])
    ax.set_ylim([-1.0, 1.0])
    axlabels = ['X', 'Y', 'Z']
    title = axlabels[projection[1]] + ' vs. ' + axlabels[projection[0]] 
    # ax.tick_params(bottom = 'off', top = 'off', left = 'off', right = 'off',
    #                labelbottom = 'off', labeltop = 'off', labelleft = 'off', labelright = 'off')
    ax.set_title(title +" scattering geometry")
    ax.set_ylabel(axlabels[projection[1]])
    # ax.text(0.12, 0.70, "This is not the same coverage calculator as before.")
    names = vdict.keys()
    for nv in names:
        start,  end = vdict[nv]
        spoint = (start[projection[0]],  start[projection[1]])
        epoint = (end[projection[0]],  end[projection[1]])
        ax.plot([spoint[0], epoint[0]], [spoint[1],  epoint[1]],  label = nv)
        # ax.annotate(nv, epoint, xytext = spoint,
        #         arrowprops={'arrowstyle':'-|>', 'mutation_scale':10, 'linewidth':2.0})
                #, 'shrinkA':5, 'shrinkB':7, 'mutation_scale':20, 'mutation_aspect':1.3})
                #, 'width':1.5, 'frac':0.2, 'headwidth':5})
    ax.legend(loc=0)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        mpl.savefig(outFile)
        mpl.close()
