#!/usr/bin/env python3

#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

__doc__ = """ This will be the PEAXIS sample tracker.
Given the manipulator coordinates and some calibration data,
it will be able to calculate and show where the sample is in respect to the beam.
It is planned to combine it with sample photos."""

import numpy as np
import scipy.optimize as scopt
# import scipy.interpolate as scint
# import math
import sys
import os
# import time
import copy
# import ctypes
from os.path import expanduser

# from PyQt5.Qt import Qt as _Qt
from PyQt5.QtGui import QColor, QVector3D, QQuaternion, QFont
from PyQt5.QtCore import QSize, QMetaObject, QLocale
from PyQt5.QtWidgets import QFrame, QTabWidget, QSizePolicy, QWidget, QHBoxLayout, QVBoxLayout, QFormLayout, \
                                            QFileDialog, QPushButton, QLineEdit, QApplication, QMainWindow
# from PyQt5.Qt3DInput import QInputAspect
# from PyQt5.Qt3DExtras import QPhongMaterial, QCylinderMesh, \
#     QCuboidMesh, QPlaneMesh, QSphereMesh, Qt3DWindow, QOrbitCameraController
# from PyQt5.Qt3DRender import QPointLight, QGeometry, QGeometryRenderer
# from PyQt5.Qt3DCore import QEntity, QTransform

mpl_scale = 1.0
mpl_figure_scale = 1.0
font_scale = 1.0

try:
    source = open(os.path.join(expanduser("~"),'.SampleSimulatorConfig.txt'), 'r')
except:
    pass
else:
    for line in source:
        toks = line.split()
        if len(toks) > 1:
            if toks[0] == 'Matplotlib_scale:':
                try:
                    mpl_scale = float(toks[1])
                except:
                    pass
            if toks[0] == 'Matplotlib_figure_scale:':
                try:
                    mpl_figure_scale = float(toks[1])
                except:
                    pass
            if toks[0] == 'Font_scale:':
                try:
                    font_scale = float(toks[1])
                except:
                    pass
    source.close()

import matplotlib
matplotlib.use("Qt5Agg")
matplotlib.rcParams['legend.fontsize'] = int(14*mpl_scale)
matplotlib.rcParams['legend.borderpad'] = 0.3*mpl_scale
matplotlib.rcParams['legend.labelspacing'] = 0.2*mpl_scale
matplotlib.rcParams['xtick.labelsize'] = int(16*mpl_scale)
matplotlib.rcParams['ytick.labelsize'] = int(16*mpl_scale)
matplotlib.rcParams['axes.titlesize'] = int(16*mpl_scale)
matplotlib.rcParams['axes.labelsize'] = int(16*mpl_scale)
matplotlib.rcParams['axes.grid'] = True
matplotlib.rcParams['font.size'] = int(14*mpl_scale)
import matplotlib.pyplot as mpl
from matplotlib.backends import qt_compat
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar2QTAgg
from matplotlib.patches import Circle, Rectangle

GlobFont = QFont('Sans Serif', int(12*font_scale))

from geom_tools import *
from SampleInputTab import SampleInputTab
from SampleTab import SampleTab
from ReverseSampleTab import ReverseSampleTab
from OrientationTab import OrientationTab
from TotalCoverageTab import TotalCoverageTab
from OffsetsTab import OffsetsTab
from PEAXIS import tilt_values

# z is vertical, x is along the beam, y is in-plane perpendicular to the beam, facing right

# print(tilt_values(0.0))

holder_dimensions = np.array([18.0,15.0, 1.0]) # width, length, thickness
holder_offset = np.array([0.0, -1.0, 0.0]) # along x, y, z

holder_vertices = np.array([
            [ holder_dimensions[1]/2.0,  holder_dimensions[2]/2.0,  holder_dimensions[0]/2.0],
            [ holder_dimensions[1]/2.0,  holder_dimensions[2]/2.0, -holder_dimensions[0]/2.0],
            [-holder_dimensions[1]/2.0,  holder_dimensions[2]/2.0,  holder_dimensions[0]/2.0],
            [-holder_dimensions[1]/2.0,  holder_dimensions[2]/2.0, -holder_dimensions[0]/2.0],
            [ holder_dimensions[1]/2.0, -holder_dimensions[2]/2.0,  holder_dimensions[0]/2.0],
            [ holder_dimensions[1]/2.0, -holder_dimensions[2]/2.0, -holder_dimensions[0]/2.0],
            [-holder_dimensions[1]/2.0, -holder_dimensions[2]/2.0,  holder_dimensions[0]/2.0],
            [-holder_dimensions[1]/2.0, -holder_dimensions[2]/2.0, -holder_dimensions[0]/2.0],
            ]) + holder_offset

holder_normal = np.array([0.0,-1.0,0.0])

rotation_centre = np.array([0.0,0.0,0.0])

manip_axisx = np.array([1.0,0.0,0.0])
manip_axisy = np.array([0.0,1.0,0.0])
manip_axisz = np.array([0.0,0.0,1.0])

def holder_position(xyz = np.array([0.0,0.0,0.0]), angles=np.array([0.0,0.0,0.0]), arm_theta = 90.0):
    """
    Returns the vertices of the holder plate after applying all the transformations.
    Takes in the manipulator translations (xyz) and rotations (angles).
    Angles are (zdir, rot, incl).
    """
    points = holder_vertices.copy()
    xaxis = manip_axisx.copy().reshape((1,3))
    yaxis = manip_axisy.copy().reshape((1,3))
    zaxis = manip_axisz.copy().reshape((1,3))
    normvec = holder_normal.copy().reshape((1,3))
    rotlist = [points, xaxis, yaxis, zaxis, normvec]
    ang_offsets = tilt_values(abs(90-arm_theta))
    # print(ang_offsets)
    # now we have temporary objects to work on
    # first we tilt the chamber according to the arm theta
    Cham_rot = arb_rotation(rotlist[1][0], -ang_offsets)
    for n, i in enumerate(rotlist):
        rotlist[n] = rotate(Cham_rot, i)
    # then we rotate everything according to zdir
    Z_rot = arb_rotation(rotlist[3][0], 180.0-angles[0])
    for n, i in enumerate(rotlist):
        rotlist[n] = rotate(Z_rot, i)
    # then we rotate everything according to rot
    Y_rot = arb_rotation(rotlist[2][0], angles[1])
    for n, i in enumerate(rotlist):
        rotlist[n] = rotate(Y_rot, i)
    # then we rotate everything according to incl
    X_rot = arb_rotation(rotlist[1][0], angles[2])
    for n, i in enumerate(rotlist):
        rotlist[n] = rotate(X_rot, i)
    # the last step: the translations
    points = rotlist[0]
    points[:,0] += xyz[0]
    points[:,1] += xyz[1]
    points[:,2] += xyz[2]
    face = points[:4]
    holderx = face[2]-face[3]
    holdery = face[1]-face[3]
    holderx = holder_dimensions[0] * holderx / length(holderx)
    holdery = holder_dimensions[1] * holdery / length(holdery)
    return points, (face, normvec), (holderx, holdery)

# print(holder_vertices)
# print(holder_position(angles = np.array([90.0, 0.0,10.0])))

def which_way(txtinput):
    temp = txtinput.lower().strip()
    if temp[0] == 't':
        return [0,1,2,3] # top, bottom, left, right
    elif temp[0] == 'b': 
        return [1,0,3,2] # bottom, top, right, left
    elif temp[0] == 'l':
        return [2,3,1,0]
    elif temp[0] == 'r':
        return [3,2,0,1]
    else:
        return []
        
def which_way(txtinput):
    temp = txtinput.lower().strip()
    return temp[0]

def read_image(filename):
    temp = mpl.imread(filename)
    # rawname = ".".join(filename.split(".")[:-1]) + '.raw'
    # temp.tofile(rawname)
    return temp

def parse_database(filelist):
    database = []
    for i in filelist:
        source = open(i, 'r')
        dict = {}
        for line in source:
            toks = line.split(':')
            leftside = toks[0].strip(' |?/\\][]\n(){}')
            rightside = toks[1].strip(' |?/\\][]\n(){}')
            print(leftside)
            print(rightside)
            try:
                data = np.array([float(x) for x in rightside.split(',')])
            except:
                dict[leftside] = rightside
            else:
                dict[leftside] = data
        print(dict)
        database.append(dict)
    return database

def spot_on_sample(data_dict):
    distortion = data_dict['Matrix']
    beam = data_dict['BeamMark']
    edges = data_dict['EdgeMark']
    trans = data_dict['XYZ']
    try:
        armtheta = data_dict['ArmTheta']
    except:
        armtheta = 90.0
    rots = data_dict['ZRI']
    wway = data_dict['WhichWay']
    # here we sort out the PIXEL COORDINATES
    ymin = edges[:2].min()
    ymax = edges[:2].max()
    xmin = edges[2:].min()
    xmax = edges[2:].max()
    beamx = (beam[0] - xmin)/(xmax-xmin)
    beamy = (beam[1] - ymin)/(ymax-ymin)
    # now we make sure that the coordinates are relative to the orientation facing the top
    hdir = which_way(wway)
    if hdir == 't':
        beampos = np.array([beamx, beamy, 0.0])
    elif hdir == 'b':
        beampos = np.array([1.0-beamx, 1.0-beamy, 0.0])
    elif hdir == 'l':
        beampos = np.array([1.0-beamy, beamx, 0.0])
    elif hdir == 'r':
        beampos = np.array([beamy, 1.0-beamx, 0.0])
    else:
        beampos = np.array([0.0,0.0,0.0])
    # let's get the spatial coordinates of the beam spot, using the manipulator parameters
    vertices, normal, basis = holder_position(trans, rots, armtheta)
    beamspot = vertices[3] + beampos[0]*basis[0] + beampos[1]*basis[1]
    return beamspot, normal, vertices

def beam_spot_dist(args, spots):
    line_origin = np.array([-1000.0, args[0], args[1]])
    line_vector = args[2:5]
    result = []
    for s in spots:
       result.append(line_point_distance(line_origin, line_vector, s))
    return np.concatenate(result)
    
def beam_spot_dist_fixed(args, spots):
    line_origin = np.array([-1000.0, args[0], args[1]])
    line_vector = np.array([1.0,0.0,0.0])
    result = []
    for s in spots:
       result.append(line_point_distance(line_origin, line_vector, s))
    return np.concatenate(result)

def find_beam(spots, change_direction = False):
    start_dir = np.array([1.0,0.0,0.0])
    ys, zs = [], []
    for s in spots:
        ys.append(s[1])
        zs.append(s[2])
    ys = np.array(ys)
    zs = np.array(zs)
    ymean, ysdt = ys.mean(0), ys.std(0)
    zmean, zsdt = zs.mean(0), zs.std(0)
    start_args = np.concatenate([np.array([ymean, zmean]), start_dir])
    if (len(spots) == 1) and not change_direction:
        return np.array([ymean, zmean]), np.array([-1.0, -1.0])
    if not change_direction:
        pfit, pcov, infodict, errmsg, success = scopt.leastsq(beam_spot_dist_fixed, start_args[:2], args=(spots), full_output=1)
        if pcov is not None:
            s_sq = (beam_spot_dist_fixed(pfit, spots)**2).sum()/(len(np.array(spots).ravel())-len(pfit))
            pcov = pcov * s_sq
    else:
        pfit, pcov, infodict, errmsg, success = scopt.leastsq(beam_spot_dist, start_args, args=(spots), full_output=1)
        if pcov is not None:
            s_sq = (beam_spot_dist(pfit, spots)**2).sum()/(len(np.array(spots).ravel())-len(pfit))
            pcov = pcov * s_sq
    error = [] 
    for i in range(len(pfit)):
        try:
            error.append(np.absolute(pcov[i][i])**0.5)
        except:
            error.append( 0.00 )
    return pfit, error

def vector_components(args, vecs, start, target):
    newone = start + args[0]*vecs[0] + args[1]*vecs[1]
    return newone-target

def find_beam_on_sample(xyz, zri, armtheta, beam_ori, beam_dir):
    vertices, face, basis = holder_position(xyz, zri, armtheta)
    face_verts, face_normal = face[0].copy(), face[1].copy()
    plane_normal = normalise(np.cross(basis[0],basis[1]))
    beam_dir = normalise(np.array(beam_dir))
    point = line_plane_intersection(beam_ori, beam_dir, face_verts[3], plane_normal)
    try:
        plen = len(point)
    except:
        point = None
        onsample = ('Surface parallel to the beam', None)
        return point, onsample, vertices
    pfit, pcov, infodict, errmsg, success = scopt.leastsq(vector_components, 0.5*np.ones(2), args=(basis, face_verts[3], point), full_output=1)
    if pcov is not None:
        s_sq = (vector_components(pfit, basis, face_verts[3], point)**2).sum()/(len(point.ravel())-len(pfit))
        pcov = pcov * s_sq
    error = [] 
    for i in range(len(pfit)):
        try:
            error.append(np.absolute(pcov[i][i])**0.5)
        except:
            error.append( 0.00 )
    if np.dot(face_normal, beam_dir) > 0.0:
        onsample = ('Sample facing away from the beam', None)
    elif pfit[0] > 0.0 and pfit[0] < 1.0 and pfit[1] > 0.0 and pfit[1] < 1.0:
        onsample = (pfit*np.array([holder_dimensions[0], holder_dimensions[1]]),
                    error*np.array([holder_dimensions[0], holder_dimensions[1]]))
    else:
        onsample = ('Beam not on sample', None)
    return point, onsample, vertices
    
# Here some plotting routines
def plot_beam_on_sample(pic, ax, outFile = "", fig = None, text = '', marker = (0,0)):
    if fig == None:
        fig = mpl.figure(figsize = [12.0, 8.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    labels = ['Counts','Counts']
    symbolcount = 0
    handles = []
    ptlabels = []
    # print(pic.shape, pic.min(), pic.max())
    try:
        dims = len(pic.shape)
    except:
        nopicture = True
    else:
        nopicture = False
    axes = fig.add_subplot(111)
    axlabels = ['Holder width [mm]', 'Holder height [mm]']
    if nopicture:
        axes.plot([0.0,18.0,18.0,0.0,0.0], [0.0, 0.0, 15.0, 15.0, 0.0], 'b-')
    else:
        xxx = axes.imshow(pic, extent = [0.0, 15.0,
                                        0.0, 18.0], interpolation = 'none',
                                        cmap = mpl.get_cmap('OrRd'), aspect = 'auto',
                                        vmin = np.percentile(pic, 20), vmax = np.percentile(pic, 90)
                                        # vmin = 1e-3, vmax = 1.0
                                        )
    axes.grid(True)
    axes.set_xlabel(axlabels[0])
    axes.set_ylabel(axlabels[1])
    box = axes.get_position()
    axes.set_position([box.x0, box.y0 + box.height * 0.2,
             box.width, box.height * 0.8])
    tpos_x = axes.get_xlim()[0]
    if marker is not None:
        beamc1 = Circle(marker, 0.1, color = 'red')
        axes.add_artist(beamc1)
        beamc2 = Circle(marker, 0.08, color = 'white')
        axes.add_artist(beamc2)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        fig.canvas.draw()
        mpl.savefig(outFile, bbox_inches = 'tight')
        mpl.close()
        
def plot_beam_in_space(pic, projection='x', outFile = "", fig = None, text = '', marker = None):
    if fig == None:
        fig = mpl.figure(figsize = [12.0, 8.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    labels = ['Counts','Counts']
    symbolcount = 0
    handles = []
    ptlabels = []
    axes_x = fig.add_subplot(221)
    axes_y = fig.add_subplot(222)
    axes_z = fig.add_subplot(223)
    projections = ['x', 'y', 'z']
    titles = ['View along beam', 'Side view', 'Top view']
    for n, axes in enumerate([axes_x, axes_y, axes_z]):
        projection = projections[n]
        if projection == 'x':
            points = np.column_stack([pic[:,1], pic[:,2]])
            if marker is not None:
                mark = [marker[1],marker[2]]
            axlabels = ['y [mm]', 'z [mm]']
        if projection == 'y':
            points = np.column_stack([pic[:,0], pic[:,2]])
            if marker is not None:
                mark = [marker[0],marker[2]]
            axlabels = ['x [mm]', 'z [mm]']
        if projection == 'z':
            points = np.column_stack([pic[:,1], pic[:,0]])
            if marker is not None:
                mark = [marker[1],marker[0]]
            axlabels = ['y [mm]', 'x [mm]']
        # print(pic.shape, pic.min(), pic.max())
        face1 = points[[0,2,3,1,0]]
        face2 = points[[4,6,7,5,4]]
        axes.plot(face2[:,0], face2[:,1], 'b-')
        axes.plot(face1[:,0], face1[:,1], 'r-')
        axes.grid(True)
        axes.set_xlabel(axlabels[0])
        axes.set_ylabel(axlabels[1])
        axes.set_title(titles[n])
        box = axes.get_position()
        axes.set_position([box.x0, box.y0 + box.height * 0.2,
                 box.width, box.height * 0.8])
        tpos_x = axes.get_xlim()[0]
        if marker is not None:
            beamc1 = Circle(mark, 0.1, color = 'red')
            axes.add_artist(beamc1)
            beamc2 = Circle(mark, 0.08, color = 'white')
            axes.add_artist(beamc2)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        fig.canvas.draw()
        mpl.savefig(outFile, bbox_inches = 'tight')
        mpl.close()



def plot2D(pic, outFile = "", fig = None, text = '', mark = [-1,-1], line_tblr = [-1,-1,-1,-1]):
    if fig == None:
        fig = mpl.figure(figsize = [12.0, 8.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    labels = ['Counts','Counts']
    symbolcount = 0
    handles = []
    ptlabels = []
    tshape = pic.shape
    print(pic.shape, pic.min(), pic.max())
    axes = fig.add_subplot(111)
    axlabels = ['Pixels (horizontal)', 'Pixels (vertical)']
    topval = np.nan_to_num(pic).max()
    if topval == 0.0:
        topval = 1.0
    xxx = axes.imshow(pic)
    axes.set_xlabel(axlabels[0])
    axes.set_ylabel(axlabels[1])
    #
    if mark[0] > 0 and mark[1] > 0 and mark[0] < tshape[0] and mark[1] < tshape[1]:
        beamc1 = Circle(mark, 7.0, color = 'red')
        axes.add_artist(beamc1)
        beamc2 = Circle(mark, 5.0, color = 'white')
        axes.add_artist(beamc2)
    if line_tblr[0] > 0 and line_tblr[0] < tshape[0]:
        border1 = Rectangle([0, line_tblr[0]], tshape[1], 7, color = 'red')
        axes.add_artist(border1)
        border1a = Rectangle([0, line_tblr[0]], tshape[1], 5, color = 'yellow')
        axes.add_artist(border1a)
    if line_tblr[1] > 0 and line_tblr[1] < tshape[0]:
        border2 = Rectangle([0, line_tblr[1]], tshape[1], 7, color = 'red')
        axes.add_artist(border2)
        border2a = Rectangle([0, line_tblr[1]], tshape[1], 5, color = 'yellow')
        axes.add_artist(border2a)
    if line_tblr[2] > 0 and line_tblr[2] < tshape[1]:
        border3 = Rectangle([line_tblr[2], 0], 7, tshape[0], color = 'red')
        axes.add_artist(border3)
        border3a = Rectangle([line_tblr[2], 0], 5, tshape[0], color = 'yellow')
        axes.add_artist(border3a)
    if line_tblr[3] > 0 and line_tblr[3] < tshape[1]:
        border4 = Rectangle([line_tblr[3], 0], 7, tshape[0], color = 'red')
        axes.add_artist(border4)
        border4a = Rectangle([line_tblr[3], 0], 5, tshape[0], color = 'yellow')
        axes.add_artist(border4a)
    #
    box = axes.get_position()
    axes.set_position([box.x0, box.y0 + box.height * 0.2,
             box.width, box.height * 0.8])
    tpos_x = axes.get_xlim()[0]
    ty1, ty2 = axes.get_ylim()
    tpos_y = ty2 + 0.05 * (ty2-ty1)
    axtextf = mpl.axes([0.20, 0.11, 0.10, 0.01], frameon = False) # , axisbg = '0.9')
    axtextf.set_yticks([])
    axtextf.set_xticks([])
    axtextf.set_title(text)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        fig.canvas.draw()
        mpl.savefig(outFile, bbox_inches = 'tight')
        mpl.close()

# for a challenge, a 3D scene in Qt

# now the GUI

class QHLine(QFrame):
    def __init__(self):
        super().__init__()
        self.setFrameShape(QFrame.HLine)
        self.setFrameShadow(QFrame.Sunken)

params = {}
params['Single'] = {}
params['Profile'] = {}

single_vars = [
        ["Input File", ["Single", "IFname"], "The beam spot position database.", ''],
        ]

profile_vars = [
        ["Beam origin", ["Profile", "BeamOri"], "Geometrical definition of the beam: a point on the line", '-1000.0, 7.0, -16.0'],
        ["Beam direction", ["Profile", "BeamDir"], "Geometrical definition of the beam: a vector along the line", '1.0,0.0,0.0'],
        ["Arm theta angle", ["Profile", "ArmTheta"], "The position of the detector arm [degrees]", '90.0'],
        ["X, Y, Z", ["Profile", "XYZ"], "Manipulator translations", '0.0,0.0,-20.0'],
        ["ZDir, Rot, Incl", ["Profile", "ZRI"], "Manipulator rotations", '90.0,0.0,0.0'],
        ["Beam position in space", ["Profile", "BeamSpace"], "3D coordinates of the beam spot", '0.0,0.0,0.0'],
        ["Beam position on holder", ["Profile", "BeamFlat"], "2D coordinates of the beam spot", '0.0,0.0'],
        ]

for k in single_vars:
    params['Single'][k[1][1]] = ""
    
for k in profile_vars:
    params['Profile'][k[1][1]] = ""

class InterfaceTabbed:
    def __init__(self, master, size, param):
        self.master = master
        # self.master.resize(size[0], size[1]) 
        self.params = param
        self.input_fields = {}
        self.input_keys = []
        self.profile_keys = []
        self.database = []
        self.samplephoto = None
        self.current_array = None
        self.current_data = None
        self.added_header = None
        self.current_file = ""
        self.temp_name = ""
        self.tabbar = QTabWidget(self.master)
        self.tabbar.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Expanding)
        # self.tabbar.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Expanding)
        self.tabbar.setMinimumSize(QSize(size[0], size[1]))
        forplots = []
        for i in range(6):
            forplots.append(self.MakeCanvas(self.master))
        intab = SampleInputTab(self.master)
        self.tabbar.addTab(intab.base,  "Sample Definition")
        stab = SampleTab(self.master, forplots[0:3] )
        self.tabbar.addTab(stab.base,  "From motors to HKL")
        # revtab = ReverseSampleTab(self.master, [])
        # self.tabbar.addTab(revtab.base,  "From HKL to motors")
        tilttab = OrientationTab(self.master, [])
        self.tabbar.addTab(tilttab.base,  "Flexible orientation finder")
        intab.new_sample_params.connect(stab.updateSample)
        # intab.new_sample_params.connect(revtab.updateSample)
        intab.new_sample_params.connect(tilttab.updateSample)
        offtab = OffsetsTab(self.master, forplots[3:6] )
        self.tabbar.addTab(offtab.base,  "Offsets")
        tottab = TotalCoverageTab(self.master, [self.MakeCanvas(self.master)])
        self.tabbar.addTab(tottab.base, "Total Coverage")
        # self.SingleTab()
        self.temp_path = expanduser("~")
        # self.master.setMinimumSize(QSize(size[0], size[1]))
        self.master.resize(size[0], size[1])
        self.master.setCentralWidget(self.tabbar)
        self.master.setFont(GlobFont)
        # self.master.resized.connect(self.on_resize)
        self.load_last_params()
        intab.on_parchange()
        # self.set_up_3D()
    def set_up_3D(self):
        self.view = Qt3DWindow()
        self.view.defaultFrameGraph().setClearColor(QColor(0x4d4d4f))
        container = QWidget.createWindowContainer(self.view)
        screenSize = self.view.screen().size()
        container.setMinimumSize(200, 100)
        container.setMaximumSize(screenSize)
        self.tabbar.addTab(container, '3D View')
        self.rootEntity = QEntity()
        material = QPhongMaterial(self.rootEntity)
        cuboid = QCuboidMesh()
        cuboid.setXExtent(holder_dimensions[0])
        cuboid.setYExtent(holder_dimensions[2])
        cuboid.setZExtent(holder_dimensions[1])
        self.cuboidTransform = QTransform()
        cuboidMaterial = QPhongMaterial()
        cuboidMaterial.setDiffuse(QColor(0x665499))
        #
        m_cuboidEntity = QEntity(self.rootEntity)
        m_cuboidEntity.addComponent(cuboid)
        m_cuboidEntity.addComponent(cuboidMaterial)
        m_cuboidEntity.addComponent(self.cuboidTransform)   
        # coordinate system
        self.axes = []
        for n, x in enumerate(['x', 'y', 'z']):
            tempcyl = QCylinderMesh()
            tempcyl.setRadius(0.1)
            tempcyl.setLength(50)
            tempcyl.setRings(50)
            tempcyl.setSlices(20)
            tempTransform = QTransform()
            tempTransform.setScale(1.0)
            tempMaterial = QPhongMaterial()
            if n == 0:
                tempTransform.setRotation(QQuaternion.fromAxisAndAngle(QVector3D(1.0, 0.0, 0.0), 90.0))
                tempTransform.setTranslation(QVector3D(0.0, 0.0, 25.0))
                tempMaterial.setDiffuse(QColor(0xBB0000))
            elif n == 1:
                tempTransform.setRotation(QQuaternion.fromAxisAndAngle(QVector3D(0.0, 1.0, 0.0), 90.0))
                tempTransform.setTranslation(QVector3D(0, 25.0, 0.0))
                tempMaterial.setDiffuse(QColor(0x00BB00))
            else:
                tempTransform.setRotation(QQuaternion.fromAxisAndAngle(QVector3D(0.0, 0.0, 1.0), 90.0))
                tempTransform.setTranslation(QVector3D(25.0, 0, 0.0))
                tempMaterial.setDiffuse(QColor(0x0000BB))
            m_cylinderEntity = QEntity(self.rootEntity)
            m_cylinderEntity.addComponent(tempcyl)
            m_cylinderEntity.addComponent(tempMaterial)
            m_cylinderEntity.addComponent(tempTransform)
            self.axes.append(m_cylinderEntity)
        # origin
        sphereMesh = QSphereMesh()
        sphereMesh.setRings(20)
        sphereMesh.setSlices(20)
        sphereMesh.setRadius(1.5)
        sphereTransform = QTransform()
        sphereTransform.setScale(1.0)
        sphereTransform.setTranslation(QVector3D(0.0, 0.0, 0.0))
        sphereMaterial = QPhongMaterial()
        sphereMaterial.setDiffuse(QColor(0xEEEEEE))
        m_sphereEntity = QEntity(self.rootEntity)
        m_sphereEntity.addComponent(sphereMesh)
        m_sphereEntity.addComponent(sphereMaterial)
        m_sphereEntity.addComponent(sphereTransform)
        # beam
        tempcyl = QCylinderMesh()
        tempcyl.setRadius(0.1)
        tempcyl.setLength(500)
        tempcyl.setRings(100)
        tempcyl.setSlices(10)
        tempTransform = QTransform()
        tempTransform.setScale(1.0)
        tempMaterial = QPhongMaterial()
        tempTransform.setRotation(QQuaternion.fromAxisAndAngle(QVector3D(0.0, 0.0, 1.0), 90.0))
        tempTransform.setTranslation(QVector3D(0.0, 0.0, 0.0))
        tempMaterial.setDiffuse(QColor(0xBBEEEE))
        m_cylinderEntity = QEntity(self.rootEntity)
        m_cylinderEntity.addComponent(tempcyl)
        m_cylinderEntity.addComponent(tempMaterial)
        m_cylinderEntity.addComponent(tempTransform)
        self.beamTransform = tempTransform
        # Camera
        self.camera = self.view.camera()
        self.camera.lens().setPerspectiveProjection(45.0, 16.0/9.0, 0.1, 1000.0)
        self.camera.setPosition(QVector3D(-60.0, 20.0, 40.0))
        self.camera.setViewCenter(QVector3D(0,0,0))
        self.camera.setUpVector(QVector3D(0.0, 0.0, 1.0))   
        # add light
        lightEntity = QEntity(self.rootEntity)
        light = QPointLight(lightEntity)
        light.setColor(_Qt.white)
        light.setIntensity(1)
        lightEntity.addComponent(light)
        lightTransform = QTransform(lightEntity)
        lightTransform.setTranslation(self.camera.position())
        lightEntity.addComponent(lightTransform)
        # For camera controls
        camController = QOrbitCameraController(self.rootEntity)
        camController.setCamera(self.camera)
        self.sampleplate = m_cuboidEntity
        self.view.setRootEntity(self.rootEntity)
        self.view.show()
    def createScene(self, xyz = np.array([0.0,0.0,0.0]), angles=np.array([0.0,0.0,0.0]), arm_theta = 90.0,
                beam_source = [-1000.0,0.0,0.0], beam_focus = [0.0,0.0,0.0]):
        # Root entity

        # cuboidTransform.setScale(4.0)
        self.cuboidTransform.setTranslation(QVector3D(xyz[0], xyz[1], xyz[2]))
        xaxis = manip_axisx.copy().reshape((1,3))
        yaxis = manip_axisy.copy().reshape((1,3))
        zaxis = manip_axisz.copy().reshape((1,3))
        normvec = holder_normal.copy().reshape((1,3))
        rotlist = [xaxis, yaxis, zaxis]
        ang_offsets = tilt_values(abs(90-arm_theta))
        Cham_rot = arb_rotation(rotlist[0][0], ang_offsets)
        for n, i in enumerate(rotlist):
            rotlist[n] = rotate(Cham_rot, i)
        Z_rot = arb_rotation(rotlist[2][0], angles[0]-180.0)
        for n, i in enumerate(rotlist):
            rotlist[n] = rotate(Z_rot, i)
        Y_rot = arb_rotation(rotlist[1][0], -angles[1])
        for n, i in enumerate(rotlist):
            rotlist[n] = rotate(Y_rot, i)
        X_rot = arb_rotation(rotlist[0][0], -angles[2])
        for n, i in enumerate(rotlist):
            rotlist[n] = rotate(X_rot, i)
        quat = self.cuboidTransform.fromAxes(QVector3D(rotlist[0][0][0], rotlist[0][0][1], rotlist[0][0][2]),
                                            QVector3D(rotlist[1][0][0], rotlist[1][0][1],rotlist[1][0][2] ),
                                            QVector3D(rotlist[2][0][0], rotlist[2][0][1], rotlist[2][0][2]))
        self.cuboidTransform.setRotation(quat)
        self.camera.setViewCenter(QVector3D(xyz[0], xyz[1], xyz[2]))
        self.beamTransform.setTranslation(QVector3D(0.0, beam_source[1], beam_source[2]))

        # return rootEntity
    def on_resize(self):
        self.master.resize(self.master.sizeHint())
    def save_last_params(self, lastfunction = None):
        try:
            source = open(os.path.join(expanduser("~"),'.SampleSimulatorConfig.txt'), 'w')
        except:
            return None
        else:
            source.write('Lastdir: '+str(self.temp_path) + '\n')
            source.write('Lastfile: '+str(self.temp_name) + '\n')
            for kk in self.input_keys:
                source.write(" ".join([str(u) for u in [kk[0], kk[1], self.params[kk[0]][kk[1]]]]) + '\n')
            if not lastfunction == None:
                source.write('Last function called: ' + str(lastfunction) + '\n')
            source.write('Matplotlib_scale: ' + str(mpl_scale) + '\n')
            source.write('Matplotlib_figure_scale: ' + str(mpl_figure_scale) + '\n')
            source.write('Font_scale: ' + str(font_scale) + '\n')
            source.close()
    def load_last_params(self):
        try:
            source = open(os.path.join(expanduser("~"),'.SampleSimulatorConfig.txt'), 'r')
        except:
            return None
        else:
            for line in source:
                toks = line.split()
                if len(toks) > 1:
                    if toks[0] == 'Lastdir:':
                        try:
                            self.temp_path = toks[1]
                        except:
                            pass
            source.close()    
    def read_inputs(self, newparams, newfields = None):
        if newfields is None:
            newfields = self.input_fields
        temp_params = copy.deepcopy(newparams)
        for kk in self.input_keys:
            try:
                field = newfields[kk[0]][kk[1]]
            except:
                continue
            val = field.text()
            if type(val) == type([]):
                if len(val) == 1:
                    val = val[0]
            try:
                temp_params[kk[0]][kk[1]] = str(val)
            except:
                continue
        return temp_params
    def validate_params(self, newparams, newfields = None):
        temp_params = copy.deepcopy(newparams)
        marker, edges = [],[]
        xyz, zri = [], []
        for i in str(temp_params["Profile"]["BeamOri"]).split(','):
            try:
                marker.append(float(i))
            except:
                marker.append(-1)
        for i in str(temp_params["Profile"]["BeamDir"]).split(','):
            try:
                edges.append(float(i))
            except:
                edges.append(-1)
        for i in str(temp_params["Profile"]["ArmTheta"]).split(',')[:1]:
            try:
                arm_angle = float(i)
            except:
                arm_angle = 90.0
        for i in str(temp_params["Profile"]["XYZ"]).split(','):
            try:
                xyz.append(float(i))
            except:
                xyz.append(0.0)
        for i in str(temp_params["Profile"]["ZRI"]).split(','):
            try:
                zri.append(float(i))
            except:
                zri.append(0.0)
        if not (len(marker) == 3):
            marker = [-1,-1]
        if not (len(edges) == 3):
            edges = [-1,-1]
        if not (len(xyz) == 3):
            xyz = [0,0,0]
        if not (len(zri) == 3):
            zri = [0,0,0]
        # prepare output
        temp_params["Profile"]["BeamOri"] = marker
        temp_params["Profile"]["BeamDir"] = edges
        temp_params["Profile"]["ArmTheta"] = arm_angle
        temp_params["Profile"]["XYZ"] = xyz
        temp_params["Profile"]["ZRI"] = zri
        return temp_params
    def update_params(self, temp_params, newfields = None):
        if newfields is None:
            newfields = self.input_fields
        for kk in self.input_keys:
            val = temp_params[kk[0]][kk[1]]
            field = newfields[kk[0]][kk[1]]
            field.setText(str(val).strip('[])(\'\'""'))
        self.params = temp_params
        # print(self.canvasref.size())
    def load_picture_button(self):      
        result, ftype = QFileDialog.getOpenFileName(self.master, 'Load an image of your sample', self.temp_path,
           'All files (*.*)')
        # print(result)
        if result == "":
            return None
        else:
            self.samplephoto = read_image(result)
            plot2D(self.samplephoto, fig = self.single_figure, text = "")
    def SingleTab(self):
        base = QWidget(self.master)
        base.setSizePolicy(QSizePolicy.Maximum, QSizePolicy.Expanding)
        base_layout = QVBoxLayout()
        form = QWidget(base)
        layout = QFormLayout(form)
        form2 = QWidget(base)
        form3 = QWidget(base)
        second_layout = QHBoxLayout(form2)
        third_layout = QFormLayout(form3)
        form3.setSizePolicy(QSizePolicy.Minimum,QSizePolicy.Minimum)
        # structure of vars: label, dictionary keys, tooltip
        col1 = "background-color:rgb(30,180,20)"
        col2 = "background-color:rgb(0,210,210)"
        col3 = "background-color:rgb(50,150,250)"
        #
        load_button = self.MakeButton(base, 'Choose Input Files', self.load_file_button, 'Pick the database files for beam positions.')
        trig_button = self.MakeButton(base, 'Find fixed beam', self.beam_button, 'Find the beam origin assuming ideal 1,0,0 direction.')
        redo_button = self.MakeButton(base, 'Find full beam', self.full_beam_button, 'Find the beam origin and direction.')
        photo_button = self.MakeButton(base, 'Load Photo', self.load_picture_button, 'Load a photo of your sample holder with the sample on it.')
        # prof_button = self.MakeButton(base, 'Plot Curvature', self.curve_profile, 'Plot position of the peak as a function of horizontal position.')
        param_button = self.MakeButton(base, 'Simulate Sample', self.calc_sample_button, 'Calculate the beam position on sample.')
        proj_button = self.MakeButton(base, 'Simulate 3D', self.calc_space_button, 'Calculate the beam position in space.')
        # eV_button = self.MakeButton(base, 'Energy Transfer Plot', self.eV_profile_button, 'Same as Make Profile, but using energy units on the x axis.')
        # merge_wiz_button = self.MakeButton(base, 'Merge Wizard', self.merge_files_wizard_button, 'Combines files in eV units, using different parameters for each file.')
        # multiplot_button = self.MakeButton(base, 'Overplot results', self.multiplot, 'Plots many merged curves together.')
        # rixsmap_button = self.MakeButton(base, 'RIXS Map!', self.rixsmap, 'Plots many merged curves together as a 2D map.')
        trig_button.setStyleSheet(col1)
        redo_button.setStyleSheet(col1)
        param_button.setStyleSheet(col2)
        proj_button.setStyleSheet(col2)
        # merge_wiz_button.setStyleSheet(col3)
        base_layout.addWidget(load_button)
        for x in single_vars:
            label, keys, tooltip, initv = x[0], x[1], x[2], x[3]
            a = QLineEdit(form)
            layout.addRow(label, a)
            a.setToolTip(tooltip)
            try:
                val = self.params[keys[0]]
                for k in keys[1:]:
                    val = val[k]
            except KeyError:
                val = 0
            if keys[0] in self.input_fields.keys():
                self.input_fields[keys[0]][keys[1]] = a
            else:
                self.input_fields[keys[0]] = {}
                self.input_fields[keys[0]][keys[1]] = a
            self.input_keys.append(keys)
            a.setText(str(initv).strip('[]'))
        for x in profile_vars:
            label, keys, tooltip, initv = x[0], x[1], x[2], x[3]
            a = QLineEdit(form3)
            third_layout.addRow(label, a)
            a.setToolTip(tooltip)
            try:
                val = self.params[keys[0]]
                for k in keys[1:]:
                    val = val[k]
            except KeyError:
                val = 0
            if keys[0] in self.input_fields.keys():
                self.input_fields[keys[0]][keys[1]] = a
            else:
                self.input_fields[keys[0]] = {}
                self.input_fields[keys[0]][keys[1]] = a
            self.input_keys.append(keys)
            self.profile_keys.append((keys[0],keys[1]))
            a.setText(str(initv).strip('[]'))
        third_layout.addRow('', QHLine())
        # third_layout.addRow('Changed parameters?', redo_button)
        third_layout.addRow('', trig_button)
        third_layout.addRow('', redo_button)
        third_layout.addRow('', photo_button)
        # third_layout.addRow('', QHLine())
        # third_layout.addRow('', QHLine())
        third_layout.addRow('', param_button)
        third_layout.addRow('', proj_button)
        # third_layout.addRow('Got multiple files to merge?', merge_wiz_button)
        # third_layout.addRow('', QHLine())
        # third_layout.addRow('Post-processing: stacked plot', multiplot_button)
        # third_layout.addRow('Post-processing: RIXS Map', rixsmap_button)
        base_layout.addWidget(form)
        base_layout.addWidget(form2)
        canvas, figure, clayout = self.MakeCanvas(base)
        second_layout.addWidget(canvas)
        second_layout.addWidget(form3)
        # base_layout.addLayout(clayout)
        self.single_figure = figure
        base.setLayout(base_layout)
        self.canvasref = canvas
        self.canvasref.setMinimumSize(QSize(693, 838))
        self.tabbar.addTab(base, 'SingleFile')
        field = self.input_fields["Profile"]["BeamSpace"] 
        field.setReadOnly(True)
        field = self.input_fields["Profile"]["BeamFlat"] 
        field.setReadOnly(True)
    def MakeCanvas(self, parent):
        mdpi, winch, hinch = 75, 9.0*mpl_figure_scale, 7.0*mpl_figure_scale
        canvas = QWidget(parent)
        layout = QVBoxLayout(canvas)
        figure = mpl.figure(figsize = [winch, hinch], dpi=mdpi )#, frameon = False)
        figAgg = FigureCanvasQTAgg(figure)
        figAgg.setParent(canvas)
        figAgg.setSizePolicy(QSizePolicy.Expanding,QSizePolicy.Expanding)
        figAgg.updateGeometry()
        toolbar = NavigationToolbar2QTAgg(figAgg, canvas)
        toolbar.update()
        layout.addWidget(figAgg)
        layout.addWidget(toolbar)
        return canvas, figure, layout
    def MakeButton(self, parent, text, function, tooltip = ""):
        button = QPushButton(text, parent)
        if tooltip:
            button.setToolTip(tooltip)
        button.clicked.connect(function)
        button.setSizePolicy(QSizePolicy.Preferred,QSizePolicy.Preferred)
        button.setMinimumSize(QSize(120, 50))
        button.setMaximumSize(QSize(500, 250))
        button.setFont(GlobFont)
        return button
    def load_datheader(self, fname):
        header = []
        try:
            source = open(fname, 'r')
        except:
            source = None        
        if not (source is None):
            for line in source:
                if len(line.split()) > 0:
                    if '#' in line.split()[0][0]:
                        header.append(line)
            source.close()
        return header
    def load_file_button(self):
        temppar = self.read_inputs(self.params, newfields = None)
        temppar = self.validate_params(temppar, newfields = self.input_fields)
        self.update_params(temppar)            
        result, ftype = QFileDialog.getOpenFileNames(self.master, 'Load multiple database files.', self.temp_path,
           'All files (*.*)')
        # print(result)
        if result == "":
            return None
        else:
            self.current_file = result
            try:
                self.database = parse_database(result)
            except:
                print("Failed parsing files.")
            self.save_last_params(lastfunction = 'load_file')
    def beam_button(self):
        temppar = self.read_inputs(self.params, newfields = None)
        temppar = self.validate_params(temppar, newfields = self.input_fields)
        self.update_params(temppar)       
        beam_ori = temppar["Profile"]["BeamOri"] 
        beam_dir = temppar["Profile"]["BeamDir"]
        arm_angle = temppar["Profile"]["ArmTheta"]
        xyz = temppar["Profile"]["XYZ"] 
        zri = temppar["Profile"]["ZRI"]
        if not self.database:
            return None
        else:
            spots = []
            for dud in self.database:
               a,b,c = spot_on_sample(dud)
               spots.append(a)
            fit, error = find_beam(spots)
            toks = [str(round(x,3)) for x in [fit[0], error[0], fit[1], error[1]]]
            val = "-1000.0, " + toks[0] + '+/-' + toks[1]+', '+  toks[2] + '+/-' + toks[3]
            val = "-1000.0, " + toks[0] + ', '+  toks[2]
            field = self.input_fields['Profile']['BeamOri']
            field.setText(str(val).strip('[])(\'\'""'))
            self.save_last_params(lastfunction = 'beam_button')
    def full_beam_button(self):
        temppar = self.read_inputs(self.params, newfields = None)
        temppar = self.validate_params(temppar, newfields = self.input_fields)
        self.update_params(temppar)       
        beam_ori = temppar["Profile"]["BeamOri"] 
        beam_dir = temppar["Profile"]["BeamDir"]
        arm_angle = temppar["Profile"]["ArmTheta"]
        xyz = temppar["Profile"]["XYZ"] 
        zri = temppar["Profile"]["ZRI"]
        if not self.database:
            return None
        else:
            spots = []
            for dud in self.database:
               a,b,c = spot_on_sample(dud)
               spots.append(a)
            fit, error = find_beam(spots, change_direction=True)
            toks = [str(round(x,3)) for x in [fit[0], error[0], fit[1], error[1],
                                              fit[2], error[2], fit[3], error[3], fit[4], error[4]]]
            val1 = "-1000.0, " + toks[0] + '+/-' + toks[1]+', '+  toks[2] + '+/-' + toks[3]
            val2 = toks[4] + '+/-' + toks[5] +', '+  toks[6] + '+/-' + toks[7] +', '+  toks[8] + '+/-' + toks[9]
            val1 = "-1000.0, " + toks[0] + ", " + toks[2]
            val2 = toks[4] + ', '+  toks[6] +', '+  toks[8]
            field = self.input_fields["Profile"]["BeamOri"] 
            field.setText(str(val1).strip('[])(\'\'""'))
            field = self.input_fields["Profile"]["BeamDir"] 
            field.setText(str(val2).strip('[])(\'\'""'))
            self.save_last_params(lastfunction = 'beam_button')
    def calc_sample_button(self):
        temppar = self.read_inputs(self.params, newfields = None)
        temppar = self.validate_params(temppar, newfields = self.input_fields)
        self.update_params(temppar)       
        beam_ori = temppar["Profile"]["BeamOri"].copy()
        beam_dir = temppar["Profile"]["BeamDir"].copy()
        arm_angle = temppar["Profile"]["ArmTheta"]
        xyz = temppar["Profile"]["XYZ"].copy()
        zri = temppar["Profile"]["ZRI"].copy()
        inspace, onsample, vertices = find_beam_on_sample(xyz, zri, arm_angle, beam_ori, beam_dir) 
        try:
            toks1 = [str(round(x,3)) for x in inspace]
            val1 = ", ".join(toks1)
        except:
            val1 = 'None'
        try:
            toks2 = [str(round(x,3)) for x in onsample[0]]
            val2 = ", ".join(toks2)
        except:
            val2 = str(onsample)
        field = self.input_fields["Profile"]["BeamSpace"] 
        field.setReadOnly(False)
        field.setText(str(val1).strip('[])(\'\'""'))
        field.setReadOnly(True)
        field = self.input_fields["Profile"]["BeamFlat"] 
        field.setReadOnly(False)
        field.setText(str(val2).strip('[])(\'\'""'))
        field.setReadOnly(True)
        try:
            mark = np.array([float(x) for x in onsample[0][:2]])
        except:
            mark = None
        if self.samplephoto is not None:
            plot_beam_on_sample(self.samplephoto, None, fig = self.single_figure, marker = mark)
        else:
            plot_beam_on_sample(None, None, fig = self.single_figure, marker = mark)
            
    def calc_space_button(self):
        temppar = self.read_inputs(self.params, newfields = None)
        temppar = self.validate_params(temppar, newfields = self.input_fields)
        self.update_params(temppar)       
        beam_ori = temppar["Profile"]["BeamOri"].copy()
        beam_dir = temppar["Profile"]["BeamDir"].copy()
        beam_end = temppar["Profile"]["BeamOri"].copy()
        beam_end[0] = -beam_end[0]
        arm_angle = temppar["Profile"]["ArmTheta"]
        xyz = temppar["Profile"]["XYZ"].copy()
        zri = temppar["Profile"]["ZRI"].copy()
        inspace, onsample, vertices = find_beam_on_sample(xyz, zri, arm_angle, beam_ori, beam_dir) 
        try:
            toks1 = [str(round(x,3)) for x in inspace]
            val1 = ", ".join(toks1)
        except:
            val1 = 'None'
        try:
            toks2 = [str(round(x,3)) for x in onsample[0]]
            val2 = ", ".join(toks2)
        except:
            val2 = str(onsample)
        field = self.input_fields["Profile"]["BeamSpace"] 
        field.setReadOnly(False)
        field.setText(str(val1).strip('[])(\'\'""'))
        field.setReadOnly(True)
        field = self.input_fields["Profile"]["BeamFlat"] 
        field.setReadOnly(False)
        field.setText(str(val2).strip('[])(\'\'""'))
        field.setReadOnly(True)
        try:
            mark = np.array(inspace)
            if len(mark) < 3:
                mark = None
        except:
            mark = None
        plot_beam_in_space(vertices, projection = 'x', fig = self.single_figure, marker = mark)
        
        # 3D part
        self.createScene(xyz, zri, arm_angle, beam_ori, beam_end)
        self.sampleplate.setEnabled(False)
        self.sampleplate.setEnabled(True)
        self.view.requestUpdate()

    def save_results(self):
        temppar = self.read_inputs(self.params, newfields = None)
        temppar = self.validate_params(temppar, newfields = self.input_fields)
        self.update_params(temppar)        
        result, ftype = QFileDialog.getSaveFileName(self.master, 'Save Parameters', self.temp_path,
           'All files (*.*)')        
        if result == "":
            return None
        else:
            dump = open(result, 'w')
            for kk in temppar.keys():
                subdir = temppar[kk]
                for kl in subdir.keys():
                    dump.write(str(kl) + ' : ' + str(subdir[kl]) + '\n')
            dump.close()

def startGUI(some_args):
    app = QApplication(some_args)
    root = QMainWindow()
    root.setWindowTitle("ORDER: ORientation Definition Extending to Reciprocal space")
    # interface = InterfaceEmbedded(root, 800, params)    
    my_locale = QLocale(QLocale.Language.English, QLocale.Country.Ireland)
    # my_locale = QLocale(QLocale.Language.German, QLocale.Country.Germany)
    QLocale.setDefault(my_locale)
    interface = InterfaceTabbed(root, (1100, 1000), params) 
    QMetaObject.connectSlotsByName(root)
    root.show()
    app.exec_()

if __name__ == '__main__':
    startGUI(sys.argv)

    
