
#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

# from PyQt5.Qt import Qt as _Qt
# from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot,  QMutex
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QPushButton
import numpy as np
# import copy
from PEAXIS import ComputeKernel
from VariablesGUI import VarBox

import matplotlib.pyplot as mpl

range_variables = [
{'Name': 'Ei',  'Unit':'eV',  'Value':np.array([200.0,1.0,  1200.0]),
            'MinValue':np.array([200.0, 1e-3, 200.0]),  'MaxValue':np.array([1200.0, 100.0, 1200.0]), 
            'Length': 3,  'Type':'float', 'Key':'Ei'}, 
# {'Name': 'Ef_min',  'Unit':'eV',  'Value':840.0,  'MinValue':150.0,  'MaxValue':1250.0,  'Length': 1,  'Type':'float', 'Key':'Ef_min'}, 
# {'Name': 'Ef_max',  'Unit':'eV',  'Value':866.0,  'MinValue':150.0,  'MaxValue':1250.0,  'Length': 1,  'Type':'float', 'Key':'Ef_max'},
{'Name': 'arm_theta',  'Unit':'degree',  'Value':np.array([33.0, 1.0, 139.0]),
            'MinValue':np.array([33.0, 1e-1, 33.0]),  'MaxValue':np.array([139.0, 10.0, 139.0]), 
            'Length': 3,  'Type':'float', 'Key':'arm_theta'}, 
]

class TotalCoverageTab(QObject):
    values_changed = pyqtSignal()
    def __init__(self, parent,  forplotting):
        super().__init__(parent)
        self.CK = ComputeKernel(parent)
        self.figures = []
        for n in forplotting:
            canvas, figure, layout = n
            self.figures.append(figure)
        # set up the layout
        self.base = QWidget(parent)
        self.ohlay = QHBoxLayout(self.base)
        self.panel = QWidget(self.base)
        # self.plots = QWidget(self.base)
        self.vlay = QVBoxLayout(self.panel)
        # self.pvlay = QVBoxLayout(self.plots)
        self.tophalf = QWidget(self.base)
        self.hlay1 = QHBoxLayout(self.tophalf)
        self.bottomhalf = QWidget(self.base)
        self.hlay2 = QHBoxLayout(self.bottomhalf)
        self.ohlay.addWidget(self.panel)
        # self.ohlay.addWidget(self.plots)
        self.vlay.addWidget(self.tophalf)
        self.vlay.addWidget(self.bottomhalf)
        for n in forplotting:
            canvas, figure, layout = n
            self.hlay2.addWidget(canvas)
        # now add the components
        self.box1 = VarBox(self.base, range_variables,  "Range limits", prec_override = 5)
        self.hlay1.addWidget(self.box1.base)
        button1 = QPushButton("Total Coverage", self.base)
        button1.clicked.connect(self.CK.total_coverage)
        self.hlay1.addWidget(button1)
        # self.box2 = VarBox(self.base, instrument_variables,  "Manipulator", prec_override = 5)
        # self.hlay1.addWidget(self.box2.base)
        # self.box3 = VarBox(self.base, rixs_variables,  "Scattering", prec_override = 5)
        # self.hlay2.addWidget(self.box3.base)
        # self.box4 = VarBox(self.base, output_variables,  "Output", prec_override = 5)
        # self.hlay2.addWidget(self.box4.base)
        # self.hlay2.addWidget(canvas)
        self.range_vals = self.box1.returnValues()
        # self.instrument_vals = self.box2.returnValues()
        # self.rixs_vals = self.box3.returnValues()
        # and link the signals to the slots to automatise the operation
        self.box1.values_changed.connect(self.on_parchange)
        # self.box2.values_changed.connect(self.on_parchange)
        # self.box3.values_changed.connect(self.on_parchange)
        # self.box1.disableChanges()
        # self.box4.disableChanges()
        self.CK.total_result.connect(self.output)
        self.on_parchange()
    @pyqtSlot(object)
    def updateSample(self, input):
        vlist,  vdict = input[0],  input[1]
        self.box1.enableChanges()
        self.box1.takeValues(vlist, vdict)
        self.box1.disableChanges()
    def calculate(self):
        self.CK.minimal_input(self.range_vals)
        # self.CK.calc_Rmatrix()
        # self.CK.calc_total()
    def on_parchange(self):
        self.range_vals = self.box1.returnValues()
        self.calculate()
    @pyqtSlot(object)
    def output(self, results):
        Eax,  qax, angax, angmap = results
        fig = self.figures[0]
        fig.clear()
        ax = fig.add_subplot(111)
        ax.grid(True)
        ax.grid(True, which = "minor", linestyle = "--", axis='y')
        ax.grid(True, which = "minor", linestyle = "--",  color = 'k', axis='x')
        ax.tick_params(axis="x", labelbottom=True, labeltop=False, which = 'major')
        ax.tick_params(axis="x", labelbottom=False, labeltop=True, which = 'minor')
        # ax.tick_params(axis="y", right = True, labelright = True, which = 'minor')
        exts = [Eax.min(),Eax.max(),qax.min(),qax.max() ]
        # temp = angmap[np.where(np.logical_not(angmap==np.nan))]
        angmin = angax.min()
        angmax = angax.max()
        angmap[np.where(angmap<1e-12)]=np.nan
        im = ax.imshow(angmap.T, interpolation='none',
                       origin='lower',
                       extent=exts, clip_on=True,
                       aspect = 'auto',
                       cmap = mpl.get_cmap('gist_rainbow')
                       # vmin = 0.0, vmax = 1e-3 * vscale # vmax = np.percentile(Z, 60.0)
                       )
        # cb = mpl.colorbar(im, ax = im.axes, ticks = np.linspace(angmin, angmax, 5)) #, format = '%.0e', pad = 0.02)
        cb = mpl.colorbar(im, ax = im.axes, ticks = np.concatenate([[33], np.arange(40, 131, 10), [139]]).astype(np.int)) #, format = '%.0e', pad = 0.02)
        if angmax - angmin > 16 :
            cb.ax.set_xticklabels([str(int(x)) for x in np.linspace(angmin, angmax, 5)])
        cb.set_label("Arm angle [$^{\circ}$]")
        ax.set_yticks(np.arange(0, 2.01, 0.2))
        ax.set_yticks(np.arange(0.1, 2.01, 0.2),  minor = True)
        ax.set_yticklabels([str(round(x, 1)) for x in np.arange(0.1, 2.01, 0.2)],  minor = True)
        ax.set_xticks([277, 392, 525, 452, 510, 572, 637, 705, 775, 849, 928, 1012], minor = True)
        ax.set_xticklabels(["C", "N",  "O",  "Ti",  "V",  "Cr",  "Mn",  "Fe",  "Co",  "Ni",  "Cu",  "Zn"],  minor = True)
        ax.set_xlim([Eax[0],Eax[-1]])
        ax.set_ylim([qax[0],qax[-1]])
        ax.set_ylabel("Q [$\AA{}^{-1}$]")
        ax.set_xlabel("E [eV]")
        ticks = ax.xaxis.get_minor_ticks()
        # a = ax.get_xgridlines()
        for n, b in enumerate(ticks):
            if n < 3:
                b.gridline.set_color('blue')
                b.gridline.set_linewidth(2)
            else:
                b.gridline.set_color('red')
                b.gridline.set_linewidth(2)
        labels = ax.get_xminorticklabels()
        for n, b in enumerate(labels):
            if n < 3:
                b.set_color('blue')
            else:
                b.set_color('red')
        fig.canvas.draw()

def plot_vectors_projected(vdict,  projection = (1, 0), outFile = "", fig = None):
    if fig == None:
        fig = mpl.figure(figsize = [7.0, 6.0], dpi=75, frameon = False)
        trigger = True
    else:
        fig.clear()
        trigger = False
    ax = fig.add_subplot(111)
    ax.grid(True)
    ax.set_xlim([1.0, -1.0])
    ax.set_ylim([-1.0, 1.0])
    axlabels = ['X', 'Y', 'Z']
    title = axlabels[projection[1]] + ' vs. ' + axlabels[projection[0]] 
    # ax.tick_params(bottom = 'off', top = 'off', left = 'off', right = 'off',
    #                labelbottom = 'off', labeltop = 'off', labelleft = 'off', labelright = 'off')
    ax.set_title(title +" scattering geometry")
    ax.set_ylabel(axlabels[projection[1]])
    # ax.text(0.12, 0.70, "This is not the same coverage calculator as before.")
    names = vdict.keys()
    for nv in names:
        start,  end = vdict[nv]
        spoint = (start[projection[0]],  start[projection[1]])
        epoint = (end[projection[0]],  end[projection[1]])
        ax.plot([spoint[0], epoint[0]], [spoint[1],  epoint[1]],  label = nv)
        # ax.annotate(nv, epoint, xytext = spoint,
        #         arrowprops={'arrowstyle':'-|>', 'mutation_scale':10, 'linewidth':2.0})
                #, 'shrinkA':5, 'shrinkB':7, 'mutation_scale':20, 'mutation_aspect':1.3})
                #, 'width':1.5, 'frac':0.2, 'headwidth':5})
    ax.legend(loc=0)
    if not outFile:
        if trigger:
            mpl.show()
        else:
            fig.canvas.draw()
    else:
        mpl.savefig(outFile)
        mpl.close()
