
#    This file is part of ORDER.
#
#    ORDER is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.
# 
# Copyright (C) Maciej Bartkowiak, 2019-2021

# from PyQt5.Qt import Qt as _Qt
# from PyQt5 import QtGui, QtCore, QtWidgets
from PyQt5.QtCore import QThread, QObject, pyqtSignal, pyqtSlot,  QMutex
from PyQt5.QtWidgets import QWidget, QHBoxLayout, QVBoxLayout, QFileDialog, QPushButton
import numpy as np
import os
# import copy
# from PEAXIS import ComputeKernel
from VariablesGUI import VarBox, StrBox

unit_cell = [
{'Name': 'a',  'Unit':'Angstrom',  'Value':4.0,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'a'}, 
{'Name': 'b',  'Unit':'Angstrom',  'Value':4.0,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'b'}, 
{'Name': 'c',  'Unit':'Angstrom',  'Value':4.0,  'MinValue':0.0,  'MaxValue':1e5,  'Length': 1,  'Type':'float', 'Key':'c'}, 
{'Name': 'alpha',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'alpha'}, 
{'Name': 'beta',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'beta'}, 
{'Name': 'gamma',  'Unit':'degree',  'Value':90,  'MinValue':0.0,  'MaxValue':180.0,  'Length': 1,  'Type':'float', 'Key':'gamma'},]

sample_orientation = [
{'Name': 'u',  'Unit':'RLU',  'Value':np.array([0, 1, 0]),
                                      'MinValue':-1e5*np.ones(3),
                                      'MaxValue':1e5*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'u'}, 
{'Name': 'v',  'Unit':'RLU',  'Value':np.array([0, 0,  1]),
                                      'MinValue':-1e5*np.ones(3),
                                      'MaxValue':1e5*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'v'}, 
{'Name': 'misalignment',  'Unit':'degrees',  'Value':np.array([0, 0, 0]),
                                      'MinValue':-360*np.ones(3),
                                      'MaxValue':360*np.ones(3),
                                      'Length': 3,  'Type':'float', 'Key':'misalignment'}, 
]

description = {
'Name':'Generic sample', 
'Formula':'A1 B1 C3',
'Space group':'Pm3m',
'Comments':'Some important information we want to remember.'
}


class SampleInputTab(QObject):
    values_changed = pyqtSignal()
    new_sample_params = pyqtSignal(object)
    def __init__(self, parent,  forplotting= []):
        super().__init__(parent)
        # self.CK = ComputeKernel(parent)
#        self.figures = []
#        for n in forplotting:
#            canvas, figure, layout = n
#            self.figures.append(figure)
        # set up the layout
        self.temp_path = '.'
        self.base = QWidget(parent)
        self.ohlay = QHBoxLayout(self.base)
        self.panel = QWidget(self.base)
        self.plots = QWidget(self.base)
        self.vlay = QVBoxLayout(self.panel)
        self.pvlay = QVBoxLayout(self.plots)
        self.tophalf = QWidget(self.base)
        self.hlay1 = QHBoxLayout(self.tophalf)
        self.bottomhalf = QWidget(self.base)
        self.hlay2 = QHBoxLayout(self.bottomhalf)
        self.buttonbar = QWidget(self.base)
        self.hlay3 = QHBoxLayout(self.buttonbar)
        self.ohlay.addWidget(self.panel)
        self.ohlay.addWidget(self.plots)
        self.vlay.addWidget(self.tophalf)
        self.vlay.addWidget(self.bottomhalf)
        self.vlay.addWidget(self.buttonbar)
#        for n in forplotting:
#            canvas, figure, layout = n
#            self.pvlay.addWidget(canvas)
        # now add the components
        self.box1 = VarBox(self.base, unit_cell,  "Unit cell", prec_override = 5)
        self.hlay1.addWidget(self.box1.base)
        self.box2 = VarBox(self.base, sample_orientation,  "Orientation", prec_override = 5)
        self.hlay1.addWidget(self.box2.base)
        self.box3 = StrBox(self.base, description,  "Description (Optional)")
        self.hlay2.addWidget(self.box3.base)
        # self.hlay2.addWidget(canvas)
        self.sample_cell = self.box1.returnValues()
        self.sample_orientation = self.box2.returnValues()
        self.sample_description = self.box3.returnValues()
        # and link the signals to the slots to automatise the operation
        self.box1.values_changed.connect(self.on_parchange)
        self.box2.values_changed.connect(self.on_parchange)
        self.box3.values_changed.connect(self.on_parchange)
        butt1 = QPushButton("Load Sample Definition", self.buttonbar)
        butt1.clicked.connect(self.loadSample)
        butt2 = QPushButton("Save Sample Definition", self.buttonbar)
        butt2.clicked.connect(self.saveSample)
        self.hlay3.addWidget(butt1)
        self.hlay3.addWidget(butt2)
    def on_parchange(self):
        self.unitcell_vals = self.box1.returnValues()
        self.orientation_vals = self.box2.returnValues()
        self.description_vals = self.box3.returnValues()
        self.summarise()
    def summarise(self):
        output_variables = {
        'a':self.unitcell_vals[1]['a'],
        'b':self.unitcell_vals[1]['b'],
        'c':self.unitcell_vals[1]['c'],
        'alpha':self.unitcell_vals[1]['alpha'],
        'beta':self.unitcell_vals[1]['beta'],
        'gamma':self.unitcell_vals[1]['gamma'],
        'u':self.orientation_vals[1]['u'],
        'v':self.orientation_vals[1]['v'],
        'misalignment':self.orientation_vals[1]['misalignment']
        }
        output_names = [str(x) for x in output_variables.keys()]
        self.new_sample_params.emit([output_names, output_variables])
    def saveSample(self):
        fname = None
        fname, ftype = QFileDialog.getSaveFileName(self.base,
           'Load a sample definition from a text file:', self.temp_path,
           'PEAXIS sample file (*.sam);;All files (*.*)')
        if len(fname) > 0:
            self.temp_path, shortname = os.path.split(fname)
            dump = open(fname, 'w')
            for sdic in [self.unitcell_vals[1], self.orientation_vals[1], self.description_vals[1]]:
                for kk in sdic.keys():
                    kname = str(kk)
                    kval = sdic[kk]
                    try:
                        dummy = kval.mean()
                    except:
                        vstr = str(kval)
                    else:
                        vstr = " ".join([str(x) for x in kval])
                    dump.write(" ".join([kname, ':', vstr, '\n']))
            dump.close()
    def loadSample(self):
        fname = None
        fname, ftype = QFileDialog.getOpenFileName(self.base,
           'Load a sample definition from a text file:', self.temp_path,
           'PEAXIS sample file (*.sam);;All files (*.*)')
        if len(fname) > 0:
            self.temp_path, shortname = os.path.split(fname)
            source = open(fname, 'r')
            vlist, vdict = [], {}
            for line in source:
                toks = line.split('#')
                precom = toks[0]
                toks = precom.split(':')
                if len(toks) < 2:
                    continue
                else:
                    kname = toks[0].strip(" :")
                    kvals = toks[1].strip(" :[](){}\n")
                    vtoks = kvals.split()
                    try:
                        float(vtoks[0])
                    except:
                        nvar = kvals
                    else:
                        nvar = np.array([float(x) for x in vtoks])
                    vlist.append(kname)
                    vdict[kname] = nvar
            source.close()
        else:
            return None
        for bb in [self.box1, self.box2, self.box3]:
            bb.takeValues(vlist, vdict)
        self.on_parchange()
